/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: GNOMEInterface.cpp,v 1.6 2001/03/23 04:12:15 pascal Exp $ */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <gnome.h>

#include "../../lib/snprintf.h"

#include "GNOMEInterface.h"
#include "GtkEntryPeer.h"
#include "GtkOptionMenuPeer.h"
#include "GdkPixmapPeer.h"
#include "../../lib/MobilizerPreferences.h"
#include "../../lib/PointerList.h"
#include "../../lib/String.h"
#include "../../renderer/GUIPeer.h"
#include "../../renderer/TestElementPeer.h"
#include "../../renderer/WBMP.h"
#include "../../renderer/WMLGUIElement.h"
#include "../../renderer/WMLWindow.h"

/** 
 * Constructs a new GNOMEInterface object.
 * @param content the canvas that we are drawing in.
 * @param graphicsContext the graphics context
 * @param bColor the background color.
 * @param fColor the foreground color.
 * @param lColor the link color.
 * @param alColor the active link color.
 * @param foundry the foundry of the font.
 * @param name the name of the font.
 * @param size the size of the font.
 */
GNOMEInterface::GNOMEInterface(GtkWidget *content,
			       GdkGC *graphicsContext, 
                               GdkColor *fColor,
                               GdkColor *bColor,
			       GdkColor *lColor, 
			       GdkColor *alColor,
                               const char *foundry,
                               const char *name,
                               int size) {
  canvas = content;
  gc = graphicsContext;
  foreColor = fColor;
  backColor = bColor;
  linkColor = lColor;
  activeLinkColor = alColor;

  int i;
  for(i = 0; i < 12; i++)
    fonts[i] = NULL;

  setFont(foundry, name, size);

}

/**
 * Delete the object
 */
GNOMEInterface::~GNOMEInterface() {

  destroyFontStyles();

}

/**
 * Sets the base font (foundry, name & size) to draw in.
 * @param foundry the font foundry.
 * @param name the font name.
 * @param size the font size in pixels.
 */
void GNOMEInterface::setFont(const char *foundry, const char *name, int size) {

  strcpy(fontFoundry, foundry);
  strcpy(fontName, name);
  fontBaseSize = size;

  preferredSlant[0] = 'i';
  preferredSlant[1] = '\0';
  strcpy(preferredWeight, "bold");

  if(strcmp(name, "fixed"))
    useFixed = false;
  else
    useFixed = true;

  destroyFontStyles();
  createFontStyles();

}


/**
 * Sets the foreground color.
 * @param c the new color.
 */
void GNOMEInterface::setForegroundColor(GdkColor *c) {

  foreColor = c;

}

/**
 * Sets the background color.
 * @param c the new color.
 */
void GNOMEInterface::setBackgroundColor(GdkColor *c) {

  backColor = c;

}

/**
 * Sets the link color.
 * @param c the new color.
 */
void GNOMEInterface::setLinkColor(GdkColor *c) {

  linkColor = c;

}

/**
 * Sets the active link color.
 * @param c the new color.
 */
void GNOMEInterface::setActiveLinkColor(GdkColor *c) {

  activeLinkColor = c;

}

/**
 * Tells the GNOMEInterface the location of the mobilizer preferences
 * @param mp the pointer to the MobilizerPreferences object
 */
void GNOMEInterface::setPreferences(MobilizerPreferences *mp){

}

/** 
 * Tells the GNOMEInterface that WMLWindow is about to start
 * drawing. After this member has been called, calls to the other
 * members can be expected. When width and height are 0, actual
 * drawing wont take place, so GNOMEInterface::drawText wont be called
 * before GNOMEInterface::stopDrawing.
 * @param width the width of the content
 * @param height the height of the content 
 */
void GNOMEInterface::startDrawing(int width, int height) {
  
}

/** 
 * Tells the WindowDrawingInterface that WMLWindow is about to stop
 * drawing. After this member has been called, only
 * GNOMEInterface::startDrawing can be called.  
 */
void GNOMEInterface::stopDrawing() {

}

/**
 * Returns the height of the string in the given WMLWindow::style in
 * pixels.
 * @param s the string 
 * @param style the WMLWindow::style in which the
 * string will be displayed 
 * @return the height in pixels 
 */
int GNOMEInterface::getStringHeight(String* s, int style) {

  setFontStyle(style);
  return currentFont->ascent + currentFont->descent;

}

/**
 * Returns the width of the string in the given WMLWindow::style in pixels.
 * @param s the string
 * @param style the WMLWindow::style in which the string will be displayed
 * @return the width in pixels
 */
int GNOMEInterface::getStringWidth(String *s, int style) {

  setFontStyle(style);
  return gdk_string_width(currentFont, s->getPtr());

}

/**
 * Returns the amount of pixels below the baseline of the font used
 * with the given WMLWindow::style.
 * @param style the WMLWindow::style from which the pixel below
 * baseline must be determined
 * @return amount of pixels below the baseline 
 */
int GNOMEInterface::getPixelsBelowBaseLine(int style) {

  setFontStyle(style);
  return currentFont->descent;

}

/** 
 * Draws the specified text in the display area assosiated with this
 * object.
 * @param text the to be displayed text
 * @param x the x coordinate of the left-upper point at which to
 * display text
 * @param y the y coordinate of the left-upper point at which to
 * display text
 * @param style the WMLWindow::style that must be used to display text 
 */
void GNOMEInterface::drawText(String* text, int x, int y, int style) {

  setFontStyle(style);
  
  /* Save some old values for later. */
  GdkGCValues values;
  gdk_gc_get_values(gc, &values);

  gdk_gc_set_foreground(gc, foreColor);

  /* Change text color if necessary. */
  if(style & WMLWindow::styleLink) 
    gdk_gc_set_foreground(gc, linkColor);

  if(style & WMLWindow::styleToplink) 
    gdk_gc_set_foreground(gc, activeLinkColor);

  /* Draw the text. */
  gdk_draw_string(GTK_LAYOUT(canvas)->bin_window, 
                  currentFont, 
                  gc, 
                  x, 
                  y + currentFont->ascent, 
                  text->getPtr());

  /* Underline text if necessary. */
  if(style & WMLWindow::styleLink ||
     style & WMLWindow::styleToplink ||
     style & WMLWindow::styleUnderlined)
    gdk_draw_rectangle(GTK_LAYOUT(canvas)->bin_window, 
                       gc, 
                       TRUE,
                       x, 
                       y + currentFont->ascent + 2, 
                       gdk_string_width(currentFont, text->getPtr()), 
                       currentFont->descent - 6 > 0 ? currentFont->descent - 6 : 1);

  /* Restore foreground color. */
  if(style & WMLWindow::styleLink || style & WMLWindow::styleToplink) 
    gdk_gc_set_foreground(gc, &values.foreground);

}

/** 
 * Creates a new input element. This member must return a pointer to a
 * newly allocated platform dependent object that is subclassed from
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the
 * input element.
 * @return the platform dependent object
 */
GUIPeer *GNOMEInterface::createInputElement() {

  return new GtkEntryPeer(canvas);

}

/** 
 * Draws an input element in the display area assosiated with this
 * object. The WMLGUIElement in the argument has members like
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wie the WMLGUIElement
 */
void GNOMEInterface::drawInputElement(WMLGUIElement *wie) {
  GtkEntry *gtk_entry;

  gtk_entry = GTK_ENTRY(wie->getPeer()->getPlatformDependentPointer());
  gtk_layout_move(GTK_LAYOUT(canvas),
		  GTK_WIDGET(gtk_entry), 
		  GTK_LAYOUT(canvas)->xoffset + wie->getXCoordinate(), 
		  GTK_LAYOUT(canvas)->yoffset + wie->getYCoordinate());
  gtk_widget_show(GTK_WIDGET(gtk_entry));

}

/** 
 * Creates a new select element. This member must return a pointer to
 * a newly allocated platform dependent object that is subclassed from
 * the GUIPeer class. The width and height members of this class must
 * be set to the correct values, meaning: the width and height of the
 * select element.
 * @return the platform dependent object
 */
GUIPeer *GNOMEInterface::createSelectElement(PointerList *optionList) {

  return new GtkOptionMenuPeer(canvas, optionList);

}

/** 
 * Draws a select element in the display area assosiated with this
 * object. The WMLGUIElement in the argument has members like
 * WMLGUIElement::getHeight() and WMLGUIElement::getXCoordinate().
 *
 * @param wse the WMLGUIElement 
 */
void GNOMEInterface::drawSelectElement(WMLGUIElement *wse) {
  GtkOptionMenu *gtk_option_menu;

  gtk_option_menu = GTK_OPTION_MENU(wse->getPeer()->getPlatformDependentPointer());
  gtk_layout_move(GTK_LAYOUT(canvas),
		  GTK_WIDGET(gtk_option_menu), 
		  GTK_LAYOUT(canvas)->xoffset + wse->getXCoordinate(), 
		  GTK_LAYOUT(canvas)->yoffset + wse->getYCoordinate());
  gtk_widget_show(GTK_WIDGET(gtk_option_menu));

}

/**
 * Creates a peer for the given WBMP image.
 * @param wbmp the image.
 * @return a peer for the image.
 */
GUIPeer *GNOMEInterface::createImage(WBMP *wbmp) {

  return new GdkPixmapPeer(canvas, foreColor, backColor, wbmp);

}

/** 
 * Draws the specified image in the display area assosiated 
 * with this object.
 *
 * @param image the image to display
 * @param x the x coordinate of the left-upper point at which to 
 * display the image
 * @param y the y coordinate of the left-upper point at which to 
 * display the image
 *
 */
void GNOMEInterface::drawImage(WMLGUIElement *im) {
  GdkPixmap *pixmap = 
    (GdkPixmap *) im->getPeer()->getPlatformDependentPointer();

  gdk_draw_pixmap(GTK_LAYOUT(canvas)->bin_window, 
                  gc, 
                  pixmap, 
                  0, 
                  0, 
                  im->getXCoordinate(), 
                  im->getYCoordinate(), 
                  -1, 
                  -1);

}

/**
 * Given a style, sets the variable 'currentFont' to point to
 * a preloaded font with the desired properties.
 * @param style the WMLWindow::style of the font.
 */
void GNOMEInterface::setFontStyle(int style) {
  int index = 0;

  /* If both 'big' and 'small' styles are requested, drop them both. */
  if(style & (WMLWindow::styleBig | WMLWindow::styleSmall))
    style = style & (~(WMLWindow::styleBig | WMLWindow::styleSmall));

  switch(style) {
  case WMLWindow::styleNoStyle:
    index = 0;
    break;
  case WMLWindow::styleBold:
    index = 1;
    break;
  case WMLWindow::styleItalic:
    index = 2;
    break;
  case WMLWindow::styleBig:
    index = 3;
    break;
  case WMLWindow::styleSmall:
    index = 4;
    break;
  case WMLWindow::styleBold + WMLWindow::styleItalic:
    index = 5;
    break;
  case WMLWindow::styleBold + WMLWindow::styleBig:
    index = 6;
    break;
  case WMLWindow::styleBold + WMLWindow::styleSmall:
    index = 7;
    break;
  case WMLWindow::styleItalic + WMLWindow::styleBig:
    index = 8;
    break;
  case WMLWindow::styleItalic + WMLWindow::styleSmall:
    index = 9;
    break;
  case WMLWindow::styleBold + WMLWindow::styleItalic + WMLWindow::styleBig:
    index = 10;
    break;
  case WMLWindow::styleBold + WMLWindow::styleItalic + WMLWindow::styleSmall:
    index = 11;
    break;
  }

  currentFont = fonts[index];

}

/**
 * Preloads all the possible combinations of font styles.
 */
void GNOMEInterface::createFontStyles() {
  int i = 0;

  fonts[i++] = loadFont(WMLWindow::styleNoStyle);
  fonts[i++] = loadFont(WMLWindow::styleBold);
  fonts[i++] = loadFont(WMLWindow::styleItalic);
  fonts[i++] = loadFont(WMLWindow::styleBig);
  fonts[i++] = loadFont(WMLWindow::styleSmall);
  fonts[i++] = loadFont(WMLWindow::styleBold + WMLWindow::styleItalic);
  fonts[i++] = loadFont(WMLWindow::styleBold + WMLWindow::styleBig);
  fonts[i++] = loadFont(WMLWindow::styleBold + WMLWindow::styleSmall);
  fonts[i++] = loadFont(WMLWindow::styleItalic + WMLWindow::styleBig);
  fonts[i++] = loadFont(WMLWindow::styleItalic + WMLWindow::styleSmall);
  fonts[i++] = loadFont(WMLWindow::styleBold + WMLWindow::styleItalic + 
                        WMLWindow::styleBig);
  fonts[i++] = loadFont(WMLWindow::styleBold + WMLWindow::styleItalic +
                        WMLWindow::styleSmall);

}

/**
 * Releases all the preloaded fonts.
 */
void GNOMEInterface::destroyFontStyles() {
  int i;

  for(i = 0; i < 12; i++)
    if(fonts[i])
      gdk_font_unref(fonts[i]);

}

/**
 * Given a style, this function loads and returns the corresponding
 * font. If the font cannot be found in the requested style, it falls
 * back until it finds a font that can be loaded. Ultimately, it falls
 * back to fixed, and if that font cannot be loaded, it returns NULL.
 * @param style the style.
 * @return a font or NULL if no font could be loaded.
 */
GdkFont *GNOMEInterface::loadFont(int style) {
  char fontString[1024];
  GdkFont *font = NULL;

  /** 
   * The user may select a font that cannot do italics or boldface. In
   * that case, it may be impossible to find a font that corresponds
   * exactly to the given style. So this method progressively tries
   * all the options in descending order of faithfulness: italic ->
   * oblique -> roman. If all these fail and boldface is requested, then it 
   * falls back from bold to medium and tries again. If all of these
   * fail, loads 'fixed'.
   */
  createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
  font = gdk_font_load(fontString);

  if(font == NULL) {

    /* Falling back to oblique. */
    preferredSlant[0] = 'o';
    createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
    font = gdk_font_load(fontString);

    if(font == NULL) {

      /* Falling back to roman. */
      preferredSlant[0] = 'r';
      createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
      font = gdk_font_load(fontString);

      if(font == NULL) {
        
        /* We tried italic, oblique, and roman, but still didn't get a
           font, so perhaps the weight is wrong; try medium with
           proper italics. */
        strcpy(preferredWeight, "medium");
        preferredSlant[0] = 'i';
        createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
        font = gdk_font_load(fontString);
        
        if(font == NULL) {
          
          /* Medium with proper italics failed; fallback to oblique. */
          preferredSlant[0] = 'o';
          createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
          font = gdk_font_load(fontString);
          
          if(font == NULL) {

            /* Medium with oblique failed; fallback to roman. */
            preferredSlant[0] = 'r';
            createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
            font = gdk_font_load(fontString);
            
            if(font == NULL) {

              /* If we cannot get a roman medium version of the font then
                 we settle for any version of the font. */
              preferredSlant[0] = '*';
              strcpy(preferredWeight, "*");
              createLFD(fontString, 1024, style, preferredSlant, preferredWeight);
              font = gdk_font_load(fontString);

              if(font == NULL) {

                /* If even that is not possible, then output an error
                   message and get "fixed". */
                printf("Having problems finding the font %s -- falling back to 'fixed'.\n", fontString);
                useFixed = true;
                font = gdk_font_load("fixed");

              }

            }

          }

        }

      }

    }

  }

  return font;

}

/**
 * Given a style, a slant and a weight, this function creates an 
 * X Logical Font Description to match it.
 * @param str pointer to stringbuffer that will receive LFD.
 * @param len the length of the stringbuffer.
 * @param style the style.  
 * @param slant the slant.
 * @param weight the weight.
 */
void GNOMEInterface::createLFD(char *str, int len, int style, const char *slant, const char *weight) {
  char fontWeight[100];
  char fontSlant[100];
  char fontSize[100];

  if(useFixed) {

    strncpy(str, "fixed", len);
    return;
    /* NOT REACHED. */

  }

  strcpy(fontWeight, "medium");
  strcpy(fontSlant, "r");
  sprintf(fontSize, "%d", fontBaseSize);

  if(style & WMLWindow::styleBold)
    strcpy(fontWeight, weight);

  if(style & WMLWindow::styleItalic)
    strcpy(fontSlant, slant);

  if(style & WMLWindow::styleSmall)
    sprintf(fontSize, "%d", fontBaseSize - (fontBaseSize / 4));
  
  if(style & WMLWindow::styleBig)
    sprintf(fontSize, "%d", fontBaseSize + (fontBaseSize / 4));

  snprintf(str, len, "-%s-%s-%s-%s-*-*-%s-*-*-*-*-*-*-*",
           fontFoundry,
	   fontName,
           fontWeight,
           fontSlant,
           fontSize);

}


