/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: GtkCallbacks.cpp,v 1.10 2001/03/25 21:57:43 pascal Exp $ 
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <gnome.h>

/* This hack gets around a problem in some older versions of glade.h,
   which do not play nicely with C++. */
#ifdef __cplusplus
#undef __cplusplus
#define __plusplus
extern "C" {
#endif
#include <glade/glade.h>
#ifdef __plusplus
#define __cplusplus
} /* extern "C" */
#endif

#include "FontSupport.h"
#include "Mobilizer.h"
#include "GtkCallbacks.h"

/* Following global variables are imported from Mobilizer.cpp. */
extern const gchar *authors[];
extern MobilizerGlobals mg;

/**
 * File selection widget and the string to store the chosen filename.  
 */
GtkWidget *file_selector;

/**
 * Wrapper function to destroy a FontFamily, used to 
 * release the memory for each element in the list using 
 * g_slist_foreach().
 * @param family the FontFamily to destroy.
 * @param dummy not used but required by g_slist_foreach().
 */
void gslist_free_family(void *family, void *dummy) {

  destroyFontFamily((FontFamily *) family);

}

/**
 * Compares two FontFamily objects and considers them equal when their
 * friendly names match. Necessary for g_slist_find_custom(). 
 * @param a FontFamily one
 * @param b FontFamily two
 * @return 0 if they are equal.
 */
gint gslist_compare_families(gconstpointer a, gconstpointer b) {

  return(strcmp(((FontFamily *) a)->friendlyName, 
                ((FontFamily *) b)->friendlyName));

}

/**
 * Processes events coming from the WMLWindow.
 * @param bi the browser interface related to the WMLWindow.
 * @param uwl the updates to process.
 */
void process_updates(BrowserInterface *bi) {

  UpdateWindowList *uwl = bi->ww->getUpdates();

  while(uwl->getFirstAction() != -1) {

    switch(uwl->getFirstAction()) {

    case (UpdateWindowList::actionSetOutputWindowText):
      gnome_appbar_push(GNOME_APPBAR(bi->statusbar), 
                        uwl->getFirstArgument()->getPtr());
      break;

    case (UpdateWindowList::actionClearOutputWindowText):
      gnome_appbar_clear_stack(GNOME_APPBAR(bi->statusbar));
      break;

    case (UpdateWindowList::actionChangeBackButtonLabel):
      break;

    case (UpdateWindowList::actionAddButtonLabel):
      break;

    case (UpdateWindowList::actionClearButtons):
      break;

    case (UpdateWindowList::actionHorizontalScrollBar):
      bi->horizontalScrollBarPolicy = 
        (uwl->getFirstArgument()->equals("on") ? 
         GTK_POLICY_ALWAYS : GTK_POLICY_NEVER);

      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(bi->scrolled_window),
				     bi->horizontalScrollBarPolicy,
				     bi->verticalScrollBarPolicy);

      break;

    case (UpdateWindowList::actionHorizontalScrollBarRange):
      bi->layout_width = uwl->getFirstArgument()->getInt();
      gtk_layout_set_size(GTK_LAYOUT(bi->contentarea), 
                          bi->layout_width, 
                          bi->layout_height);
      break;

    case (UpdateWindowList::actionHorizontalScrollBarPage):
      /* handled automatically by GtkScrolledWindow */
      break;

    case (UpdateWindowList::actionHorizontalScrollBarPos):
      /* handled automatically by GtkScrolledWindow */
      break;

    case (UpdateWindowList::actionVerticalScrollBar):      
      bi->verticalScrollBarPolicy = 
        (uwl->getFirstArgument()->equals("on") ? 
         GTK_POLICY_ALWAYS : GTK_POLICY_NEVER);

      gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(bi->scrolled_window),
      				     bi->horizontalScrollBarPolicy,
      				     bi->verticalScrollBarPolicy);
      break;

    case (UpdateWindowList::actionVerticalScrollBarRange):
      bi->layout_height = uwl->getFirstArgument()->getInt();
      gtk_layout_set_size(GTK_LAYOUT(bi->contentarea), 
                          bi->layout_width, 
                          bi->layout_height);
      break;

    case (UpdateWindowList::actionVerticalScrollBarPage):
      /* handled automatically by GtkScrolledWindow */
      break;

    case (UpdateWindowList::actionVerticalScrollBarPos):
      /* handled automatically by GtkScrolledWindow */
      break;

    case (UpdateWindowList::actionSetHandMousePointer):
      gdk_window_set_cursor(bi->contentarea->window,
                            bi->mouseCursors.hand);
      break;

    case (UpdateWindowList::actionSetNormalMousePointer):
      gdk_window_set_cursor(bi->contentarea->window, 
                            bi->mouseCursors.standard);
      break;

    case (UpdateWindowList::actionBackButton):
      gtk_widget_set_sensitive(bi->historyInterface.back, 
                               uwl->getFirstArgument()->equals("on") ? 
                               TRUE : FALSE);
      break;

    case (UpdateWindowList::actionForwardButton):
      gtk_widget_set_sensitive(bi->historyInterface.forward, 
                               uwl->getFirstArgument()->equals("on") ? 
                               TRUE : FALSE);
      break;

    case (UpdateWindowList::actionURIChanged):
      GtkWidget *gtk_entry = gnome_entry_gtk_entry(GNOME_ENTRY(bi->location));
      gtk_entry_set_text(GTK_ENTRY(gtk_entry), uwl->getFirstArgument()->getPtr());
      break;

    }

    uwl->deleteFirst();

  }
  
}

/**
 * Clears the WML widget.
 */  
void contentarea_clear(BrowserInterface *bi) {

  if(bi->gc) {

    GdkGCValues values;
    gdk_gc_get_values(bi->gc, &values);
    gdk_window_clear_area(GTK_LAYOUT(bi->contentarea)->bin_window,
                          0, 0, 
                          bi->contentarea->allocation.width, 
                          bi->contentarea->allocation.height);
    gdk_gc_set_foreground(bi->gc, &values.foreground);
    
  }

}

/**
 * Draws the WML content in the WML widget.
 */
void contentarea_draw(BrowserInterface *bi) {
  GtkAdjustment *va;

  if(bi->ww) {

    va = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(bi->scrolled_window));
    bi->ww->setVertical((int) va->value);
    va = gtk_scrolled_window_get_hadjustment(GTK_SCROLLED_WINDOW(bi->scrolled_window));
    bi->ww->setHorizontal((int) va->value);
    bi->ww->setWindowDimensions(bi->width, bi->height);

    /* Process updates. */
    process_updates(bi);

    bi->ww->drawContentWindow();

  }

}

/**
 * Clears the WML widget and draws the WML content in it.
 */
void contentarea_refresh(BrowserInterface *bi) {

  contentarea_clear(bi);
  contentarea_draw(bi);

}

/**
 * Creates and shows a file selector and installs a callback
 * to listen for events on the fileselector.
 */
void file_selection_create(BrowserInterface *bi) {

   /* Create the selector. */
  file_selector = gtk_file_selection_new("Please select a file to open.");

  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button),
                     "clicked", GTK_SIGNAL_FUNC(file_selection_complete), bi);

  /* Ensure that the dialog box is destroyed when the user clicks a
     button. */
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->ok_button), "clicked", GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(file_selector));
  
  gtk_signal_connect_object(GTK_OBJECT(GTK_FILE_SELECTION(file_selector)->cancel_button), "clicked", GTK_SIGNAL_FUNC (gtk_widget_destroy), GTK_OBJECT(file_selector));
   
  /* Display that dialog. */
  gtk_widget_show(file_selector);

}

/*************************************************************
 * Callbacks.                                                *
 *************************************************************/

/* C linkage required for Gtk callbacks. */
#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

void file_selection_complete(GtkFileSelection *selector, 
                             BrowserInterface *bi) {

   gchar *selected_filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(file_selector));
   loadSource(bi, filenameToURI(selected_filename));

}

/**
 * User pressed return in 'Location:' bar.
 */
void on_location_activate(GtkWidget *w, BrowserInterface *bi) {

  GtkWidget *gtk_entry = gnome_entry_gtk_entry(GNOME_ENTRY(bi->location));
  loadSource(bi, new URI(gtk_entry_get_text(GTK_ENTRY(gtk_entry))));

}

/**
 * Handler for size-allocate event (widget resized).
 */
gboolean on_contentarea_size_allocate(GtkWidget *w,
                                      GtkAllocation *allocation,
                                      BrowserInterface *bi) {

  bi->width = w->allocation.width;
  bi->height = w->allocation.height;

  return FALSE;

}

/**
 * Handler for draw event.
 */
gboolean on_contentarea_draw(GtkWidget *widget, 
                             GdkRectangle *area,
                             BrowserInterface *bi) {

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state],
                            area);

  contentarea_draw(bi);

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state],
                            NULL);
  
  return FALSE;

}

/**
 * Expose event handler (refreshes the WML widget).
 */
gboolean on_contentarea_expose_event(GtkWidget *widget, 
                                     GdkEventExpose *event, 
                                     BrowserInterface *bi) {

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state],
                            &event->area);

  contentarea_draw(bi);

  gdk_gc_set_clip_rectangle(widget->style->fg_gc[widget->state],
                            NULL);
  
  return FALSE;
}

/**
 * Handler for button presses.
 */
gboolean on_contentarea_button_press_event(GtkWidget *w, 
                                           GdkEvent *ev,
                                           BrowserInterface *bi) {
  
  if(bi->ww->mouseClick((int) ev->button.x, (int) ev->button.y)) {

    /* If the mouse click hit a target (link), then we put
       the scrollbars at 0,0. */
    bi->verticalAdjustment->value = 0;
    bi->horizontalAdjustment->value = 0;
    gtk_adjustment_value_changed(bi->verticalAdjustment);
    gtk_adjustment_value_changed(bi->horizontalAdjustment);

    /* Process updates from WMLWindow. */
    process_updates(bi);
    contentarea_refresh(bi);

  }

  return FALSE;

}

/**
 * Handler for key press events.
 */
gboolean on_contentarea_key_press_event(GtkWidget *w, 
                                        GdkEvent *ev,
                                        BrowserInterface *bi) {
  

  return FALSE;

}

/**
 * Handler for mouse motion events.
 */
void on_contentarea_motion_notify_event(GtkWidget *w, 
                                        GdkEvent *ev,
                                        BrowserInterface *bi) {
  
  bi->ww->mouseMove((int) ev->motion.x, (int) ev->motion.y);

  while(bi->ww->getUpdates()->getFirstAction() != -1)
    process_updates(bi);

}

/**
 * Handler to show about box.
 */
void on_about_activate(GtkWidget *w, BrowserInterface *bi) {

  gtk_widget_show(gnome_about_new("Mobilizer", VERSION,
                                  "Copyright 2001 catchy.net",
                                  (const gchar **) authors,

                                  _("Scintillate, scintillate, asteroid minikin."),
                                    NULL));

}

/**
 * Handler for 'Open' button.
 */
void on_open_clicked(GtkWidget *w, BrowserInterface *bi) {

  file_selection_create(bi);

}

/**
 * Handler for 'Back' button.
 */
void on_back_clicked(GtkWidget *w, BrowserInterface *bi) {

  if(bi->ww->historyGo(-1))
    contentarea_refresh(bi);

}

/**
 * Handler for 'Forward' button.
 */
void on_forward_clicked(GtkWidget *w, BrowserInterface *bi) {
  
  if(bi->ww->historyGo(1))
    contentarea_refresh(bi);

}

/**
 * Handler for 'Home' button.
 */
void on_home_clicked(GtkWidget *w, BrowserInterface *bi) {

  loadSource(bi, new URI("http://www.catchy.net/wap/index.wml"));

}

/**
 * Handler for 'Reload' button.
 */
void on_reload_clicked(GtkWidget *w, BrowserInterface *bi) {

  loadSource(bi, new URI(bi->ww->getURI()->toString()));

}

/**
 * Handler for when the font size text box changes (enables 'Apply').
 */
void on_fontsize_changed() {

  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for when text is inserted into the font size text box.
 */
void on_fontsize_insert_text(GtkEditable *w,
                             gchar *new_text,
                             gint new_text_length,
                             gint *position,
                             gpointer user_data) {

  int i, is_number = 1;

  /* Only allow numbers in the fontsize textwidget. */
  for(i = 0; i < new_text_length; i++)
    if(!isdigit(new_text[i]))
      is_number = 0;

  if(!is_number)
    gtk_signal_emit_stop(GTK_OBJECT(w), 
                         gtk_signal_lookup("insert-text", GTK_TYPE_EDITABLE));

}

/**
 * Handler for when a selection has been made from the font 
 * family popup menu (to enable 'Apply').
 */
void on_fontfamilies_menu_deactivate(GtkMenuItem *menuitem) {

  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for the 'Settings->Preferences' menu item.
 */
void on_preferences_activate(GtkWidget *mi, BrowserInterface *bi) {

  /* Create copies of the current colors. */
  mg.currentColors._background.red = mg.currentColors.background.red;
  mg.currentColors._background.green = mg.currentColors.background.green;
  mg.currentColors._background.blue = mg.currentColors.background.blue;

  mg.currentColors._foreground.red = mg.currentColors.foreground.red;
  mg.currentColors._foreground.green = mg.currentColors.foreground.green;
  mg.currentColors._foreground.blue = mg.currentColors.foreground.blue;

  mg.currentColors._link.red = mg.currentColors.link.red;
  mg.currentColors._link.green = mg.currentColors.link.green;
  mg.currentColors._link.blue = mg.currentColors.link.blue;

  mg.currentColors._activeLink.red = mg.currentColors.activeLink.red;
  mg.currentColors._activeLink.green = mg.currentColors.activeLink.green;
  mg.currentColors._activeLink.blue = mg.currentColors.activeLink.blue;

  /* Load the preferences window. */
  GladeXML *xml = glade_xml_new(mg.gladeInterface, "preferences");

  mg.preferencesInterface.preferences = glade_xml_get_widget(xml, "preferences");

  mg.preferencesInterface.fontSelectionInterface.families = glade_xml_get_widget(xml, "fontfamilies");

  mg.preferencesInterface.fontSelectionInterface.size = glade_xml_get_widget(xml, "fontsize");

  mg.preferencesInterface.colorSelectionInterface.link = glade_xml_get_widget(xml, "linkcolorpicker");
  mg.preferencesInterface.colorSelectionInterface.activeLink = glade_xml_get_widget(xml, "activelinkcolorpicker");
  mg.preferencesInterface.colorSelectionInterface.background = glade_xml_get_widget(xml, "backgroundcolorpicker");
  mg.preferencesInterface.colorSelectionInterface.foreground = glade_xml_get_widget(xml, "foregroundcolorpicker");

  /* Get a list of fonts from the X server. */
  getFontList(&(mg.preferencesInterface.fontSelectionInterface.fontList.elements), &(mg.preferencesInterface.fontSelectionInterface.fontList.size));

  /* Create a menu to hold the font list. */
  mg.preferencesInterface.fontSelectionInterface.familiesMenu = GTK_MENU(gtk_menu_new());

  int i, menuIndex = 0;
  GtkWidget *item;
  FontFamily *fontInfo;

  mg.preferencesInterface.fontSelectionInterface.fontInfoList = NULL;

  /* Combine the name and foundry from each font to create a
     friendlyName and populate the menu with unique friendlyNames. */
  for(i = 0; i < mg.preferencesInterface.fontSelectionInterface.fontList.size; i++) {

    fontInfo = createFontFamily(mg.preferencesInterface.fontSelectionInterface.fontList.elements[i]);

    if(fontInfo == NULL)
      break;

    if(!g_slist_find_custom(mg.preferencesInterface.fontSelectionInterface.fontInfoList, fontInfo, gslist_compare_families)) {
      
      /* This friendlyName is not already in the list, so add it. */
      mg.preferencesInterface.fontSelectionInterface.fontInfoList = g_slist_append(mg.preferencesInterface.fontSelectionInterface.fontInfoList, 
                                    fontInfo);
      item = gtk_menu_item_new_with_label(fontInfo->friendlyName);
      gtk_object_set_data(GTK_OBJECT(item), "fontinfo", fontInfo);
      gtk_menu_append(mg.preferencesInterface.fontSelectionInterface.familiesMenu, item);

      /* Set the menu to point to the current font. */
      if(strcmp(mg.currentFont.family->friendlyName, fontInfo->friendlyName) == 0)
        gtk_menu_set_active(mg.preferencesInterface.fontSelectionInterface.familiesMenu, menuIndex);

      gtk_widget_show(item);
      
      menuIndex++;

    } else {
      
      /* This friendlyName already in the list. */
      destroyFontFamily(fontInfo);
      
    }
    
  }

  /* Add the menu to the OptionMenu. */
  gtk_option_menu_set_menu(GTK_OPTION_MENU(mg.preferencesInterface.fontSelectionInterface.families), GTK_WIDGET(mg.preferencesInterface.fontSelectionInterface.familiesMenu));

  /* Listen for changes to the menu so we can tell whether or not
     the 'Apply' button needs to be activated. */
  gtk_signal_connect(GTK_OBJECT(mg.preferencesInterface.fontSelectionInterface.familiesMenu), "deactivate",
		     GTK_SIGNAL_FUNC(on_fontfamilies_menu_deactivate), NULL);

  /* Show the menu. */
  gtk_widget_show(GTK_WIDGET(mg.preferencesInterface.fontSelectionInterface.familiesMenu));

  /* Show the OptionMenu. */
  gtk_widget_show(mg.preferencesInterface.fontSelectionInterface.families);

  /* Set the current font size into the text box. */
  char buf[4];
  snprintf(buf, 4, "%d", mg.currentFont.size);
  gtk_entry_set_text(GTK_ENTRY(mg.preferencesInterface.fontSelectionInterface.size), buf);

  /* Set the current colors into the color widgets. */
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(mg.preferencesInterface.colorSelectionInterface.link),
                             mg.currentColors.link.red,
                             mg.currentColors.link.green,
                             mg.currentColors.link.blue,
                             0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(mg.preferencesInterface.colorSelectionInterface.activeLink),
                             mg.currentColors.activeLink.red,
                             mg.currentColors.activeLink.green,
                             mg.currentColors.activeLink.blue,
                             0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(mg.preferencesInterface.colorSelectionInterface.foreground),
                             mg.currentColors.foreground.red,
                             mg.currentColors.foreground.green,
                             mg.currentColors.foreground.blue,
                             0);
  gnome_color_picker_set_i16(GNOME_COLOR_PICKER(mg.preferencesInterface.colorSelectionInterface.background),
                             mg.currentColors.background.red,
                             mg.currentColors.background.green,
                             mg.currentColors.background.blue,
                             0);

  /* Connect signals. */
  GtkWidget *w;

  w = glade_xml_get_widget(xml, "preferences");

  gtk_signal_connect(GTK_OBJECT(w), "apply",
		     GTK_SIGNAL_FUNC(on_preferences_apply), bi);

  gtk_signal_connect(GTK_OBJECT(w), "close",
		     GTK_SIGNAL_FUNC(on_preferences_close), bi);
  
  w = glade_xml_get_widget(xml, "fontsize");

  gtk_signal_connect(GTK_OBJECT(w), "changed",
		     GTK_SIGNAL_FUNC(on_fontsize_changed), bi);

  gtk_signal_connect(GTK_OBJECT(w), "insert-text",
		     GTK_SIGNAL_FUNC(on_fontsize_insert_text), bi);

  w = glade_xml_get_widget(xml, "backgroundcolorpicker");

  gtk_signal_connect(GTK_OBJECT(w), "color-set",
		     GTK_SIGNAL_FUNC(on_background_color_set), bi);

  w = glade_xml_get_widget(xml, "foregroundcolorpicker");

  gtk_signal_connect(GTK_OBJECT(w), "color-set",
		     GTK_SIGNAL_FUNC(on_text_color_set), bi);

  w = glade_xml_get_widget(xml, "linkcolorpicker");

  gtk_signal_connect(GTK_OBJECT(w), "color-set",
		     GTK_SIGNAL_FUNC(on_link_color_set), bi);

  w = glade_xml_get_widget(xml, "activelinkcolorpicker");

  gtk_signal_connect(GTK_OBJECT(w), "color-set",
		     GTK_SIGNAL_FUNC(on_activelink_color_set), bi);

  /* Release interface description. */
  gtk_object_unref(GTK_OBJECT(xml));


}

/**
 * Handler for the 'Apply' action in the preferences dialog.
 */
void on_preferences_apply(GnomePropertyBox *box, 
                          int page, 
                          BrowserInterface *bi) {
  FontFamily *fontInfo;
  GtkWidget *w;

  w = gtk_menu_get_active(mg.preferencesInterface.fontSelectionInterface.familiesMenu);

  fontInfo = (FontFamily *) gtk_object_get_data(GTK_OBJECT(w), 
                                                       "fontinfo");

  /* Destroy the old default font and set a new one. */
  destroyFontFamily(mg.currentFont.family);
  mg.currentFont.family = copyFontFamily(fontInfo);

  gnome_config_private_set_string("/mobilizer/defaultFont/lfd", 
                                  fontInfo->LFD);

  /* Get the new font size from the text widget. */
  int newFontSize = atoi(gtk_entry_get_text(GTK_ENTRY(mg.preferencesInterface.fontSelectionInterface.size)));

  /* Make sure that the new font size is bigger than 6 pixels
     for the user's sake. If it's smaller, put the current
     size back into the text widget. */
  if(newFontSize > 6) {

    mg.currentFont.size = newFontSize;
    bi->wdi->setFont(fontInfo->foundry, fontInfo->name, mg.currentFont.size);
    gnome_config_private_set_int("/mobilizer/defaultFont/size", 
                                 mg.currentFont.size);

  } else {

    /* Set the current font size into the text box. */
    char buf[4];
    snprintf(buf, 4, "%d", mg.currentFont.size);
    gtk_entry_set_text(GTK_ENTRY(mg.preferencesInterface.fontSelectionInterface.size), buf);

    gnome_config_private_set_int("/mobilizer/defaultFont/size", 
                                 mg.currentFont.size);

  }

  /* Restore the temporary copies of the colors. */
  mg.currentColors.background.red = mg.currentColors._background.red;
  mg.currentColors.background.green = mg.currentColors._background.green;
  mg.currentColors.background.blue = mg.currentColors._background.blue;

  mg.currentColors.foreground.red = mg.currentColors._foreground.red;
  mg.currentColors.foreground.green = mg.currentColors._foreground.green;
  mg.currentColors.foreground.blue = mg.currentColors._foreground.blue;

  mg.currentColors.link.red = mg.currentColors._link.red;
  mg.currentColors.link.green = mg.currentColors._link.green;
  mg.currentColors.link.blue = mg.currentColors._link.blue;

  mg.currentColors.activeLink.red = mg.currentColors._activeLink.red;
  mg.currentColors.activeLink.green = mg.currentColors._activeLink.green;
  mg.currentColors.activeLink.blue = mg.currentColors._activeLink.blue;

  /* Store the colors. */
  gnome_config_private_set_int("/mobilizer/backgroundColor/red", 
                               mg.currentColors.background.red);
  gnome_config_private_set_int("/mobilizer/backgroundColor/green", 
                               mg.currentColors.background.green);
  gnome_config_private_set_int("/mobilizer/backgroundColor/blue", 
                               mg.currentColors.background.blue);

  gnome_config_private_set_int("/mobilizer/foregroundColor/red", 
                               mg.currentColors.foreground.red);
  gnome_config_private_set_int("/mobilizer/foregroundColor/green", 
                               mg.currentColors.foreground.green);
  gnome_config_private_set_int("/mobilizer/foregroundColor/blue", 
                               mg.currentColors.foreground.blue);

  gnome_config_private_set_int("/mobilizer/linkColor/red", 
                               mg.currentColors.link.red);
  gnome_config_private_set_int("/mobilizer/linkColor/green", 
                               mg.currentColors.link.green);
  gnome_config_private_set_int("/mobilizer/linkColor/blue", 
                               mg.currentColors.link.blue);

  gnome_config_private_set_int("/mobilizer/activeLinkColor/red", 
                               mg.currentColors.activeLink.red);
  gnome_config_private_set_int("/mobilizer/activeLinkColor/green", 
                               mg.currentColors.activeLink.green);
  gnome_config_private_set_int("/mobilizer/activeLinkColor/blue", 
                               mg.currentColors.activeLink.blue);

  if(page == -1) {

    /* page -1 means: all property pages set. Now perform run-once
       code. */

    /* Write configuration file to disk. */
    gnome_config_sync();
    
    /* Some magic that I don't understand. */
    mg.currentColors.background.pixel = 
      (gulong) (mg.currentColors.background.red * 65536 + 
                mg.currentColors.background.green * 256 + 
                mg.currentColors.background.blue);
    mg.currentColors.foreground.pixel = 
      (gulong) (mg.currentColors.foreground.red * 65536 + 
                mg.currentColors.foreground.green * 256 + 
                mg.currentColors.foreground.blue);
    mg.currentColors.link.pixel = 
      (gulong) (mg.currentColors.link.red * 65536 + 
                mg.currentColors.link.green * 256 + 
                mg.currentColors.link.blue);
    mg.currentColors.activeLink.pixel = 
      (gulong) (mg.currentColors.activeLink.red * 65536 + 
                mg.currentColors.activeLink.green * 256 + 
                mg.currentColors.activeLink.blue);
    
    /* Allocate the colors, best match, not writable. */
    gdk_colormap_alloc_color(bi->colorMap, &(mg.currentColors.background), 0, 1);
    gdk_colormap_alloc_color(bi->colorMap, &(mg.currentColors.foreground), 0, 1);
    gdk_colormap_alloc_color(bi->colorMap, &(mg.currentColors.link), 0, 1);
    gdk_colormap_alloc_color(bi->colorMap, &(mg.currentColors.activeLink), 0, 1);
    
    gdk_window_set_background(GTK_LAYOUT(bi->contentarea)->bin_window, 
                              &(mg.currentColors.background));

    /* Reload source. */
    loadSource(bi, new URI(bi->ww->getURI()->toString()));

    /* Release the font list that was returned by the X server. */
    freeFontList(mg.preferencesInterface.fontSelectionInterface.fontList.elements);

  }

}

/**
 * Handler for the 'Close' action in the preferences dialog.
 */
void on_preferences_close() {

  if(mg.preferencesInterface.fontSelectionInterface.fontInfoList) {

    g_slist_foreach(mg.preferencesInterface.fontSelectionInterface.fontInfoList, gslist_free_family, NULL);
    g_slist_free(mg.preferencesInterface.fontSelectionInterface.fontInfoList);

  }

  mg.preferencesInterface.fontSelectionInterface.fontInfoList = NULL;

  gtk_widget_destroy(mg.preferencesInterface.fontSelectionInterface.families);
  gtk_widget_destroy(mg.preferencesInterface.preferences);

}

/**
 * Handler for when the user selects a background color.
 */
void on_background_color_set(GnomeColorPicker *colorpicker,
                               guint r,
                               guint g,
                               guint b,
                               guint a,
                               gpointer user_data) {

  mg.currentColors._background.red = r;
  mg.currentColors._background.green = g;
  mg.currentColors._background.blue = b;  

  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for when the user selects a foreground color.
 */
void on_text_color_set(GnomeColorPicker *colorpicker,
                       guint r,
                       guint g,
                       guint b,
                       guint a,
                       gpointer user_data) {

  mg.currentColors._foreground.red = r;
  mg.currentColors._foreground.green = g;
  mg.currentColors._foreground.blue = b;  

  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for when the user selects a link color.
 */
void on_link_color_set(GnomeColorPicker *colorpicker,
                               guint r,
                               guint g,
                               guint b,
                               guint a,
                               gpointer user_data) {

  mg.currentColors._link.red = r;
  mg.currentColors._link.green = g;
  mg.currentColors._link.blue = b;
  
  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for when the user selects an active link color.
 */
void on_activelink_color_set(GnomeColorPicker *colorpicker,
                               guint r,
                               guint g,
                               guint b,
                               guint a,
                               gpointer user_data) {

  mg.currentColors._activeLink.red = r;
  mg.currentColors._activeLink.green = g;
  mg.currentColors._activeLink.blue = b;
  
  gnome_property_box_changed(GNOME_PROPERTY_BOX(mg.preferencesInterface.preferences));

}

/**
 * Handler for when the user activates the 'File->Exit' menu item.
 */
void on_exit1_activate() {

  gtk_main_quit();

}

#ifdef __cplusplus
} /* extern "C" */
#endif /* __cplusplus */


