/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: HTTP.cpp,v 1.7 2001/03/14 13:35:48 wouter Exp $ 
 */

#include <stdlib.h>
#include <string.h>

#include "Exception.h"
#include "StringHash.h"
#include "OutOfMemoryException.h"
#include "HTTP.h"
#include "ContentSinkResult.h"
#include "URI.h"
#include "String.h"
#include "Socket.h"

/**
 * Creates the HTTP object.
 */
HTTP::HTTP() {
}

/**
 * Destroys the HTTP object.
 */
HTTP::~HTTP() {
}

/**
 * Performs a HTTP GET request on the specified URI. If the monitor 
 * object m is non-NULL, it is used to give an indication of progress.
 * @throws OutOfMemoryException if some memory could not be allocated.
 * @param u the URI.
 * @param m the monitor.
 * @param r the ContentSinkResult to store the results in.
 */
void HTTP::get(URI *u, Monitor *m, ContentSinkResult *r) {
  int bufferIndex, bufferSize, read, error = 0;
  char *buffer;
  
  try {

    Socket socket(u->getHost(), u->getPort() ? atoi(u->getPort()) : 80);
    
    StringHash headers;
    String s(""), header(""), serverReplyMessage(""), errorMessage("<wml><card><p>");
    char *headerKey, *headerValue, resultCode[3], *serverReply;

    s.append("GET ");
    s.append(u->getPath());
    if(u->getQuery()) {

      s.append("?");
      s.append(u->getQuery());

    }
    s.append(" HTTP/1.0");
    s.append("\r\n");
    s.append("Host: ");
    s.append(u->getHost());
    s.append("\r\n");
    s.append("Accept: text/vnd.wap.wml\r\n");
    s.append("Accept: image/vnd.wap.wbmp\r\n");
    s.append("\r\n\r\n");
    socket.writeString(&s);

    s.clear();

    /* First get the server reply. */

    header.append(socket.readLine());

    serverReply = strchr(header.getPtr(), ' ');

    if(serverReply) {

      serverReply++;

      resultCode[0] = *serverReply; serverReply++;
      resultCode[1] = *serverReply; serverReply++;
      resultCode[2] = *serverReply; serverReply++;

      /* Skip spaces. */
      while(*serverReply == ' ' && *serverReply != '\0')
        serverReply++;

      /* Store the server message for later use. */
      serverReplyMessage.append(serverReply);

      switch(resultCode[0]) {
      case '1':
        /* Informational. */
        break;
      case '2':
        /* Successfull. */
        break;
      case '3':
        /* Redirection. */
      case '4':
        /* Client error. */
      case '5':
        /* Server error. */
      default:
        /* And any future codes. */
        errorMessage.append(serverReplyMessage.getPtr());
        error = 1;
        break;
      }

      if(!error) {

        /* Now get the HTTP headers. */

        do {   
    
          header.clear();
          header.append(socket.readLine());

          headerValue = strchr(header.getPtr(), ':');

          /* If the header does not contain a ':' then we assume it is
             broken and silently move on. */

          if(headerValue) {
        
            headerKey = header.getPtr();
        
            String *key = new String(headerKey, headerValue - headerKey);

            /* Skip spaces. */
            headerValue++;
            while(*headerValue == ' ' && *headerValue != '\0')
              headerValue++;

            String *value = new String(headerValue);
            headers.put(key, value);
            delete key;
            delete value;

          }      

        }  while(strcmp(header.getPtr(), "\r\n") != 0); 

        /* Then simply read data until EOF. */
  
        bufferSize = 1024;
        bufferIndex = 0;
        buffer = (char *) malloc(bufferSize);
  
        if(buffer) {
      
          read = 1;
    
          while(read) {
      
            read = socket.readFully(buffer + bufferIndex, bufferSize - bufferIndex);
            bufferIndex += read;
      
            if(read) {
        
              buffer = (char *) realloc(buffer, bufferSize + 1024);
        
              if(buffer == NULL)
                throw OutOfMemoryException();
        
              bufferSize += 1024;
        
            }
      
          }
    
        } else {
    
          throw OutOfMemoryException();
    
        }

        /* Remember the buffer in the ContentSinkResult. The buffer is
           freed by the ContentSinkResult when it is no longer needed. */

        r->setContent((void *) buffer, bufferIndex);

      } else {

        /* Server returned an error or returned a reply we do not
           support. */

        errorMessage.append("</p></card></wml>");
        r->setContent(errorMessage.getPtr());

      }

    } else {

      /* Server did not return a valid response (could not find the
         error code). */

      r->setContent("<wml><card><p>Server returned invalid response.</p></card></wml>");

    }

  } catch(Exception& e) {

    String error("<wml><card><p>An error occurred while fetching ");
    error.append(u->toString());
    error.append(": ");
    error.append(e.getMessage());
    error.append("</p></card></wml>");

    r->setContent(error.getPtr());

  }

}
