/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: URI.cpp,v 1.15 2001/03/23 04:12:16 pascal Exp $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "snprintf.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "URI.h"
#include "URILexer.h"
#include "Garbage.h"

extern int uri_parse(void *parseResult);
extern int uri_debug;

/**
 * Constructs a URI object from an ASCII string representation
 * of the URI.
 * @param uri the ASCII string representation of the URI.
 */
URI::URI(const char *uri) {

#ifdef ENABLE_DEBUG
  uri_debug = 1;
#endif

  uri_string = NULL;
  path = NULL;
  path_without_parameters = NULL;
  authority = NULL;

  /* initialize parseResult, this structure is used by 
     uri_parse() to communicate with us. */
  parseResult.scheme = NULL;
  parseResult.host = NULL;
  parseResult.port = NULL;
  parseResult.userInfo = NULL;
  parseResult.absPath = NULL;
  parseResult.relPath = NULL;
  parseResult.query = NULL;
  parseResult.fragment = NULL;
  parseResult.opaque = NULL;
  parseResult.garbage.count = 0;
  parseResult.garbage.table = NULL;

  validate(uri);

}

/**
 * Destroys the URI object.
 */
URI::~URI() {

  flush();

}

/**
 * Returns the scheme for this URI.
 * For example, http, ftp, telnet ...
 * @return pointer at string describing scheme, or NULL.
 */
const char *URI::getScheme() const {
  return parseResult.scheme;
}

/**
 * Returns the host name/address for this URI.
 * For example, www.wins.uva.nl, 192.168.1.1, ...
 * @return pointer at string describing host, or NULL.
 */
const char *URI::getHost() const {
  return parseResult.host;
}

/**
 * Returns the port number for this URI.
 * For example, 80, 1024, ...
 * @return pointer at string describing port, or NULL.
 */
const char *URI::getPort() const {
  return parseResult.port;
}

/**
 * Returns the user info for this URI.
 * For example, guest, anonymous, ...
 * @return pointer at string describing user info, or NULL.
 */
const char *URI::getUserInfo() const {
  return parseResult.userInfo;
}

/**
 * Returns the absolute path component for this URI.
 * For example, cgi-bin/form.pl, index.html, ... 
 * You probably want to use URI::getPath() instead of this
 * function.
 * @return pointer at string describing absolute path, or NULL.
 */
const char *URI::getAbsolutePath() const {
  return parseResult.absPath;
}

/**
 * Returns the relative path component for this URI.
 * For example, .., cgi-bin, etc.
 * You probably want to use URI::getPath() instead of this
 * function.
 * @return pointer at string describing relative path, or NULL.
 */
const char *URI::getRelativePath() const {
  return parseResult.relPath;
}

/**
 * Returns the query for this URI.
 * For example, q=48&k=l, BookID=48, ...
 * @return pointer at string describing query, or NULL.
 */
const char *URI::getQuery() const {
  return parseResult.query;
}

/**
 * Returns the fragment for this URI.
 * For example, chapter1, a48, ...
 * @return pointer at string describing fragment, or NULL.
 */
const char *URI::getFragment() const {
  return parseResult.fragment;
}

/**
 * Returns the opaque part for this URI.
 * For example, someone@microsoft.com
 * @return pointer at string describing fragment, or NULL.
 */
const char *URI::getOpaquePart() const {
  return parseResult.opaque;
}

/**
 * Constructs the "authority" component for this URI.
 * An authority component consists of user info, host name
 * and port number.
 * @return authority or NULL if not applicable.
 */
const char *URI::getAuthority() {

  if(authority == NULL) {

    if(getUserInfo() || getHost() || getPort()) {

      authority = 
	(char *) malloc((getUserInfo() == NULL ? 0 : strlen(getUserInfo())) +
			(getHost() == NULL ? 0 : strlen(getHost())) +
			(getPort() == NULL ? 0 : strlen(getPort())) + 3);
      
      if(authority) {
	
	authority[0] = '\0';
	
	if(getUserInfo()) {
	  strcat(authority, getUserInfo());
	  strcat(authority, "@");
	}
	
	if(getHost())
	  strcat(authority, getHost());
	
	if(getPort()) {
	  strcat(authority, ":");
	  strcat(authority, getPort());
	}
	
      }

    }

  }

  return authority;

}

/**
 * This function returns the path component of the URI.
 * The path component is constructed from the relative
 * and the absolute path. For example:
 * \code
 * URI u("http://w.a.p/a;x/b;y=1/c;fg#a");
 * char *s = u.getPath(); // s = "/a;x/b;y=1/c;fg"
 * \endcode
 * @return the path component of the URI. The path component may
 * be NULL.
 */
const char *URI::getPath() {

  if(path == NULL) {
    
    if(parseResult.relPath == NULL && parseResult.absPath == NULL)
      return NULL;

    path = (char *) malloc(strlen(toString()) + 1);
    path[0] = '\0';

    if(path) {
      
      if(parseResult.relPath) 
	strcpy(path, parseResult.relPath);
      
      if(parseResult.absPath) {
	int j = 0;
	while(path[j++]);
	
  	/* prefix absolute path with slash. */
	path[j - 1] = '/';
	
	strcpy(&(path[j]), parseResult.absPath);
	
      }
      
    }
    
  }

  return path;

}

/**
 * This function returns the path component of the URI.
 * The path component is constructed from the relative
 * and the absolute path. Any parameters are stripped
 * from the path. For example:
 * \code
 * URI u("http://w.a.p/a;x/b;y=1/c;fg#a");
 * char *s = u.getPath(); // s = "/a/b/c"
 * \endcode
 * @return the path component of the URI. The path component may
 * be NULL.
 */
const char *URI::getPathWithoutParameters() {

  if(path_without_parameters == NULL) {
    
    if(parseResult.relPath == NULL && parseResult.absPath == NULL)
      return NULL;

    path_without_parameters = (char *) malloc(strlen(toString()) + 1);
    path_without_parameters[0] = '\0';

    if(path_without_parameters) {
      
      if(parseResult.relPath) 
	copyAndStripParameters(path_without_parameters, parseResult.relPath);
      
      if(parseResult.absPath) {
	int j = 0;
	while(path_without_parameters[j++]);
	
  	/* prefix absolute path with slash. */
	path_without_parameters[j - 1] = '/';
	
	copyAndStripParameters(&(path_without_parameters[j]), parseResult.absPath);
	
      }
      
    }
    
  }

  return path_without_parameters;

}

/**
 * Sets a new scheme component for this URI. To subsequently determine
 * whether the URI is still valid with the new omponent, use
 * the validate() function.  
 * @param s the new scheme component.  
 */
void URI::setScheme(const char *s) {
  setComponent(&parseResult.scheme, s);
}

/**
 * Sets a new port component for this URI. To subsequently determine
 * whether the URI is still valid with the new component, use
 * the validate() function.  
 * @param s the new host component.  
 */
void URI::setHost(const char *s) {
  setComponent(&parseResult.host, s);
}

/**
 * Sets a new port component for this URI. To subsequently determine
 * whether the URI is still valid with the new component, use
 * the validate() function.  
 * @param s the new port component.  
 */
void URI::setPort(const char *s) {
  setComponent(&parseResult.port, s);
}

/**
 * Sets a new user info component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new user info component.  
 */
void URI::setUserInfo(const char *s) {
  setComponent(&parseResult.userInfo, s);
}

/**
 * Sets a new absolute path component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new absolute path component.  
 */
void URI::setAbsolutePath(const char *s) {
  setComponent(&parseResult.absPath, s);
}

/**
 * Sets a new relative path component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new relative path component.  
 */
void URI::setRelativePath(const char *s) {
  setComponent(&parseResult.relPath, s);
}

/**
 * Sets a new query component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new query component.  
 */
void URI::setQuery(const char *s) {
  setComponent(&parseResult.query, s);
}

/**
 * Sets a new fragment component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new fragment component.  
 */
void URI::setFragment(const char *s) {
  setComponent(&parseResult.fragment, s);
}

/**
 * Sets a new opaque part component for this URI. To subsequently
 * determine whether the URI is still valid with the new component,
 * use the validate() function. 
 * @param s the new opaque part component.  
 */
void URI::setOpaquePart(const char *s) {
  setComponent(&parseResult.opaque, s);
}

/**
 * Sets the specified URI copmonent to the specified value.
 * @param s1 pointer to the pointer describing this URI component.
 * @param s2 the new value for the component.
 */
void URI::setComponent(char **s1, const char *s2) {
  char *s3 = (char *) malloc(strlen(s2) + 1);

  if(s3) {

    if(*s1)
      free(*s1);
    
    strcpy(s3, s2);
    *s1 = s3;

    flushDerivatives();
    
  }

}

/**
 * Returns true if the URI is well-formed, false otherwise. If you
 * have made changes to this URI by using (e.g.) setScheme() and want
 * to know if the URI is still valid, then before you call this
 * function, you must first call the validate() method to re-validate
 * the URI.  
 * @return true if URI is well-formed.  
 */
bool URI::isValid() const {
  return valid;
}

/**
 * Validate this URI. This function is useful when modifications
 * have been made to this URI since construction, e.g. by using 
 * setScheme(), and you want to verify that the new URI is still a
 * valid URI.
 * @return true if this URI is valid, false otherwise.
 */
bool URI::validate() {
  return validate(toString());
}

/**
 * Validate the given string.
 * @param uri string representation of the URI to validate.
 * @return true if this URI is valid, false otherwise.
 */
bool URI::validate(const char *uri) {
  /* initialize data structures. */
  flush();
  valid = false;

#ifdef DEBUG
  uri_debug = 1;
#endif

  /* flex/bison stuff follows. */
  uri_scan_start();
  void *bufferState = uri_scan_string(uri);
  int err = uri_parse(&parseResult);
  uri_delete_buffer(bufferState);
  /* end of flex/bison stuff. */

  GarbageRelease(&parseResult.garbage);

  if(err) {

    /* on parse error, invalidate results. */
    valid = false;
    flush();

  } else {

    valid = true;

  }

  return valid;

}

/**
 * Releases and initializes all data held by the object.
 */
void URI::flush() {

  if(parseResult.scheme)
    free(parseResult.scheme);
  parseResult.scheme = NULL;
  if(parseResult.host)
    free(parseResult.host);
  parseResult.host = NULL;
  if(parseResult.port)
    free(parseResult.port);
  parseResult.port = NULL;
  if(parseResult.userInfo)
    free(parseResult.userInfo);
  parseResult.userInfo = NULL;
  if(parseResult.absPath)
    free(parseResult.absPath);
  parseResult.absPath = NULL;
  if(parseResult.relPath)
    free(parseResult.relPath);
  parseResult.relPath = NULL;
  if(parseResult.query)
    free(parseResult.query);
  parseResult.query = NULL;
  if(parseResult.fragment)
    free(parseResult.fragment);
  parseResult.fragment = NULL;
  if(parseResult.opaque)
    free(parseResult.opaque);
  parseResult.opaque = NULL;

  parseResult.garbage.count = 0;
  parseResult.garbage.table = NULL;

  flushDerivatives();

}

/**
 * Releases and initializes all derivative data (i.e. data
 * generated by the functions getAuthority(), getPath(),
 * getPathWithoutParameters() and toString().
 */
void URI::flushDerivatives() {

  if(uri_string) 
    free(uri_string);
  uri_string = NULL;
  if(path)
    free(path);
  path = NULL;
  if(path_without_parameters)
    free(path_without_parameters);
  path_without_parameters = NULL;
  if(authority)
    free(authority);
  authority = NULL;

}

/**
 * URL encodes a string.
 * @return new escaped string, must be released using free() when done.
 * Returns NULL if memory is exhausted.
 */
char *URI::escape(const char *s) {
  assert(s != NULL);

  int i, special_char_count = 0;;

  for(i = 0; s[i]; i++)
    if(isMustEscapeChar(s[i]))
      special_char_count++;


  int new_len = strlen(s) + (special_char_count * 3) + 1;

  char *new_str = (char *) malloc(new_len);
  char digit_buf[3];

  if(new_str) {

    int j = 0;

    for(i = 0; s[i]; i++) {
      
      if(isMustEscapeChar(s[i])) {

	if(s[i] == ' ') {

	  /* encode space as '+' */
	  new_str[j++] = '+';

	} else {

	  /* encode others as %xx sequence */
	  new_str[j++] = '%';
	  snprintf(digit_buf, 3, "%x", s[i]);
	  new_str[j++] = digit_buf[0];
	  new_str[j++] = digit_buf[1];

	}
	
      } else {

	new_str[j++] = s[i];

      }

    }

    new_str[j] = '\0';

  }

  return new_str;

}

/**
 * URL decodes a string.
 * @return new unescaped string, must be released using free() when done.
 */
char *URI::unescape(const char *s) {
  assert(s != NULL);

  char *new_str = (char *) malloc(strlen(s));
  char digit_buf[3];
  int i, j = 0;

  if(new_str) {

    digit_buf[2] = '\0';

    for(i = 0; s[i]; i++) {

      /* walk the input string. */
      if(s[i] == '+') {

	/* + becomes ' ' */
	new_str[j++] = ' ';

      } else if(s[i] == '%') {

	/* %xx encodes a hexadecimal digit. */
	if(s[i + 1] != '\0' && s[i + 2] != '\0') {

	  digit_buf[0] = s[i + 1];
	  digit_buf[1] = s[i + 2];

	  char c = (char) strtol(digit_buf, NULL, 16);

	  new_str[j++] = c;
	  i += 2;

	}

      } else {
	
	new_str[j++] = s[i];

      }

    }

    new_str[j] = '\0';

  }

  return new_str;
}

/**
 * Returns a string representation of the URI. This method is
 * guaranteed to return a non-NULL value, unless memory is exhausted.
 * @return URI string representation.  
 */
const char *URI::toString() {
  int size = 0;

  if(uri_string == NULL) {

    if(getScheme())
      size += strlen(getScheme()) + 1; /* : */
    if(getHost())
      size += strlen(getHost()) + 2; /* // */
    if(getPort())
      size += strlen(getPort()) + 1; /* : */
    if(getUserInfo())
      size += strlen(getUserInfo()) + 1; /* @ */
    if(getAbsolutePath())
      size += strlen(getAbsolutePath()) + 1; /* / */
    if(getRelativePath())
      size += strlen(getRelativePath());
    if(getQuery())
      size += strlen(getQuery()) + 1; /* ? */
    if(getFragment())
      size += strlen(getFragment()) + 1; /* # */
    if(getOpaquePart())
      size += strlen(getOpaquePart());

    uri_string = (char *) malloc(size + 8);
    
    if(uri_string) {
      
      uri_string[0] = '\0';
      
      bool add_path_slash = false, add_double_slash = false;

      if(getScheme()) {

	strcat(uri_string, getScheme());
	strcat(uri_string, ":");
	add_double_slash = true;

      }
      
      if(getOpaquePart()) {

	strcat(uri_string, getOpaquePart());

      } else {

	if(getUserInfo()) {
	  strcat(uri_string, getUserInfo());
	  strcat(uri_string, "@");
	}
      
	if(getHost()) {
	  if(add_double_slash) {
	    strcat(uri_string, "//");
	    add_double_slash = false;
	  }
	  strcat(uri_string, getHost());
	  add_path_slash = true;
	}
      
	if(getPort()) {
	  strcat(uri_string, ":");
	  strcat(uri_string, getPort());
	}
      
	if(getRelativePath()) {
	  if(add_double_slash) {
	    strcat(uri_string, "//");
	    add_double_slash = false;
	  }
	  if(add_path_slash)
	    strcat(uri_string, "/");
	  strcat(uri_string, getRelativePath());
	}
      
	if(getAbsolutePath()) {
	  if(add_double_slash) {
	    strcat(uri_string, "//");
	    add_double_slash = false;
	  }
	  strcat(uri_string, "/");
	  strcat(uri_string, getAbsolutePath());
	}
      
	if(getQuery()) {
	  strcat(uri_string, "?");
	  strcat(uri_string, getQuery());
	}
      
	if(getFragment()) {
	  strcat(uri_string, "#");
	  strcat(uri_string, getFragment());
	}

      }
      
    }

  }

  return uri_string;
}

/**
 * Resolves two URI's to form a new absolute URI. The absolute
 * URI returned must be released by the caller using delete.
 * See [RFC2396] for more information.
 * @param baseURI the absolute base URI.
 * @param relativeURI the relative URI.
 * @return the new URI, or NULL if an error occurs (e.g. not 
 * enough memory, or one of the supplied URI's is invalid).
 */
URI *URI::resolve(URI *baseURI, URI *relativeURI) {
  
  if(!baseURI->isValid() || !relativeURI->isValid())
    return NULL;

  /* This method closely follows the approach detailed in RFC2396,
     section 5.2. The comments outlining the steps have been lifted
     from that documnt. */

  /* 2) If the path component is empty and the scheme, authority, and
     query components are undefined, then it is a reference to the
     current document and we are done.  Otherwise, the reference URI's
     query and fragment components are defined as found (or not found)
     within the URI reference and not inherited from the base URI. */

  if(relativeURI->getPath() == NULL &&
     relativeURI->getScheme() == NULL &&
     relativeURI->getAuthority() == NULL &&
     relativeURI->getQuery() == NULL) {
    char *frag = (char *) malloc(strlen(baseURI->toString()) + 
				 strlen(relativeURI->toString()) + 2);
    strcpy(frag, baseURI->toString());
    strcat(frag, relativeURI->toString());
    URI *u = new URI(frag);
    free(frag);
    return u;
    /* NOT REACHED */
  }

  /* 3) If the scheme component is defined, indicating that the
     reference starts with a scheme name, then the reference is
     interpreted as an absolute URI and we are done.  Otherwise, the
     reference URI's scheme is inherited from the base URI's scheme
     component. */

  if(relativeURI->getScheme())
    return new URI(relativeURI->toString());
    /* NOT REACHED */

  char *string_buf = (char *) malloc(strlen(baseURI->toString()) +
				     strlen(relativeURI->toString()) + 1);
  char *string_buf2 = (char *) malloc(strlen(baseURI->toString()) +
				     strlen(relativeURI->toString()) + 1);

  if(string_buf && string_buf2) {

    string_buf[0] = '\0';

    /* 4) If the authority component is defined, then the reference is
      a network-path and we skip to step 7.  Otherwise, the reference
      URI's authority is inherited from the base URI's authority
      component, which will also be undefined if the URI scheme does
      not use an authority component. */

    if(!relativeURI->getAuthority()) {
      
      if(relativeURI->getPath() && relativeURI->getPath()[0] == '/') {
	
	/* 5) If the path component begins with a slash character
	  ("/"), then the reference is an absolute-path and we skip to
	  step 7. */

	strcpy(string_buf, relativeURI->getPath());

      } else {

	/* 6) If this step is reached, then we are resolving a
	   relative-path reference.  The relative path needs to be
	   merged with the base URI's path.  Although there are many
	   ways to do this, we will describe a simple method using a
	   separate string buffer. */

	char *p = strrchr(baseURI->getPath(), '/');
      
	if(p) {
	  
	  /* 6a) All but the last segment of the base URI's path
	     component is copied to the buffer.  In other words, any
	     characters after the last (right-most) slash character,
	     if any, are excluded. */

	  memcpy(string_buf, baseURI->getPath(), 
		 (p - baseURI->getPath()) + 1);

	  string_buf[(p - baseURI->getPath()) + 1] = '\0';
	
	  /* 6b) The reference's path component is appended to the
	     buffer string. */

	  if(relativeURI->getPath() != NULL) 
	    strcat(string_buf, relativeURI->getPath());
	
	  /* 6c) All occurrences of "./", where "." is a complete path
	     segment, are removed from the buffer string. */

	  URI::stripSingleDots(string_buf);
	
	  /* 6d) If the buffer string ends with "." as a complete path
	     segment, that "." is removed. */
	  
	  if(strlen(string_buf) > 1 &&
	     string_buf[strlen(string_buf) - 2] == '/' &&
	     string_buf[strlen(string_buf) - 1] == '.')
	    string_buf[strlen(string_buf) - 1] = '\0';

	  /* 6e) All occurrences of "<segment>/../", where <segment> is
	     a complete path segment not equal to "..", are removed
	     from the buffer string.  Removal of these path segments
	     is performed iteratively, removing the leftmost matching
	     pattern on each iteration, until no matching pattern
	     remains. (string_buf + 1 to skip leading slash) */

	  URI::stripDoubleDots(string_buf + 1);
	
	  /* 6f) If the buffer string ends with "<segment>/..", where
	     <segment> is a complete path segment not equal to "..",
	     that "<segment>/.." is removed. */

	  if(strlen(string_buf) > 5 &&
	     string_buf[strlen(string_buf) - 1] == '.' &&
	     string_buf[strlen(string_buf) - 2] == '.' &&
	     string_buf[strlen(string_buf) - 3] == '/' &&
	     string_buf[strlen(string_buf) - 4] != '.' &&
	     string_buf[strlen(string_buf) - 5] != '.') {
	    string_buf[strlen(string_buf) - 3] = '\0';
	    char *rslash = strrchr(string_buf, '/');
	    if(rslash)
	      string_buf[(rslash - string_buf) + 1] = '\0';
	  }
	
	}

      }
    
    }

    string_buf2[0] = '\0';

    /* 7) The resulting URI components, including any inherited from
      the base URI, are recombined to give the absolute form of the
      URI reference. */

    /* (it is possible to do this nicer with setter methods
       and toString() -ph). */

    if(relativeURI->getScheme()) {
      strcat(string_buf2, relativeURI->getScheme());
      strcat(string_buf2, ":");
    } else if(baseURI->getScheme()) {
      strcat(string_buf2, baseURI->getScheme());
      strcat(string_buf2, ":");
    }
    
    if(relativeURI->getAuthority()) {
      strcat(string_buf2, "//");
      strcat(string_buf2, relativeURI->getAuthority());
    } else if(baseURI->getAuthority()) {
      strcat(string_buf2, "//");
      strcat(string_buf2, baseURI->getAuthority());
    }
    
    strcat(string_buf2, string_buf);
    
    if(relativeURI->getQuery()) {
      strcat(string_buf2, "?");
      strcat(string_buf2, relativeURI->getQuery());
    }
    
    if(relativeURI->getFragment()) {
      strcat(string_buf2, "#");
      strcat(string_buf2, relativeURI->getFragment());
    }
    
    URI *u = new URI(string_buf2);
    free(string_buf);
    free(string_buf2);
    return u;
    /* NOT REACHED */

  }

  if(string_buf)
    free(string_buf);
  if(string_buf)
    free(string_buf2);

  return NULL;

}

/**
 * Collapses .. references in a path. That is, when invoked with
 * the string b/c/../d, on exit the string will be b/d.
 * @param s a path without leading slash to collapse.
 */
void URI::stripDoubleDots(char *s) {
  int i;
  int j = 0;
  int mark = 0, parent_mark = 0;
  int dots = 0, seg_len = 0;
  bool nothing_collapsed = true;

  char *s2 = (char *) malloc(strlen(s) + 1);

  if(s2) {

    for(i = 0; s[i]; i++) {
      
      s2[j++] = s[i];
           
      if(nothing_collapsed) {

	/* nothing collapsed yet. */
	if(s[i] == '/' && i > 0) {

	  if(dots == 2 && seg_len == 3) {

	    /* found two dots -> collapse to parent. */
	    j = parent_mark;
	    nothing_collapsed = false;

	  } else {

	    parent_mark = mark;
	    mark = j;

	  }
	  
	  dots = 0;
	  seg_len = 0;
	}

	seg_len++;
	
	if(s[i] == '.')
	  dots++;
	else
	  dots = 0;
	
      }

    }
    
    if(i > 0 && nothing_collapsed && dots == 2 && seg_len == 3) {

      /* found two dots at end of string -> collapse to parent. */
      j = parent_mark;
      nothing_collapsed = false;

    }

    s2[j] = '\0';

    /* try to collapse again. */
    if(nothing_collapsed == false)
      stripDoubleDots(s2);

    strcpy(s, s2);

    free(s2);

  }

}

/**
 * Removes . references from a path.
 * @param s path to remove . references from.
 */
void URI::stripSingleDots(char *s) {
  char *s2 = (char *) malloc(strlen(s) + 1);

  if(s2) {

    int i, j = 0;
    int looking_for_dot = 0;

    for(i = 0; s[i]; i++) {

      s2[j++] = s[i];
      
      if(s[i] == '/') {

	if(looking_for_dot && i > 1 && s[i - 1] == '.' && s[i - 2] == '/')
	  j -= 2;

	looking_for_dot++;

      }

    }

    s2[j] = '\0';
    strcpy(s, s2);
    free(s2);

  }

}

/**
 * Copies source to destination, stripping parameters.
 * @param dest destination path.
 * @param source source path.
 */
void URI::copyAndStripParameters(char *dest, const char *source) {
  bool skipping = false;
  int i, j = 0;

  for(i = 0; source[i]; i++) {
    
    /* skip parameters. */
    if(source[i] == ';')
      skipping = true;
    
    /* these delimit the end of a parameter (i think). */
    if(skipping && 
       (source[i] == '/' ||
	source[i] == '?' ||
	source[i] == '#'))
      skipping = false;
    
    if(skipping)
      continue;
    
    dest[j++] = source[i];	  
    
  }

  dest[j] = '\0';

}

/**
 * Returns true when the character should be
 * escaped.
 * @param b the character.
 */
bool URI::isMustEscapeChar(int b) {
  return(/* Non US ASCII */
	 (b > 0x8f && b < 0xff) ||
	 /* US ASCII coded characters */
	 b < 0x1f || b == 0x7f || 
	 /* Space */
	 b == ' ' ||
	 /* Reserved */
	 b == ';' || b == '/' || b == '?' || b == ':' || 
	 b == '@' || b == '&' || b == '=' || b == '+' ||
	 b == '$' || b == ',' ||
	 /* Unwise */
	 b == '{' || b == '}' || b == '|' || b == '\\' ||
	 b == '^' || b == '[' || b == ']' || b == '`' ||
	 /* Delims */
	 b == '<' || b == '>' || b == '#' || b == '%' || 
	 b == '"');
}

/**
 * Performs a number of tests on the URI class.
 * @return the number of tests failed.
 */
int URI::test() {
  int fail = 0;

  URI u1("http://awacs.dhs.org");
  if(!u1.getScheme() ||
     !u1.getHost() ||
     (strcmp(u1.getScheme(), "http") ||
      strcmp(u1.getHost(), "awacs.dhs.org") ||
      u1.getPort() ||
      u1.getUserInfo() ||
      u1.getAbsolutePath() ||
      u1.getRelativePath() ||
      u1.getQuery() ||
      u1.getFragment()))
    fail = showError(u1, fail);

  URI u2("http://awacs.dhs.org/secret/index.html");
  if(!u2.getScheme() ||
     !u2.getHost() ||
     !u2.getAbsolutePath() ||
     (strcmp(u2.getScheme(), "http") ||
      strcmp(u2.getHost(), "awacs.dhs.org") ||
      strcmp(u2.getAbsolutePath(), "secret/index.html") ||
      u2.getPort() ||
      u2.getUserInfo() ||
      u2.getRelativePath() ||
      u2.getQuery() ||
      u2.getFragment()))
    fail = showError(u2, fail);

  URI u3("../images/pic1.gif");
  if(!u3.getRelativePath() ||
     !u3.getAbsolutePath() ||
     (strcmp(u3.getAbsolutePath(), "images/pic1.gif") ||
      strcmp(u3.getRelativePath(), "..") ||
      u3.getScheme() ||
      u3.getHost() ||
      u3.getPort() ||
      u3.getUserInfo() ||
      u3.getQuery() ||
      u3.getFragment()))
    fail = showError(u3, fail);

  URI u4("/navigatie.php3?Expand=true");
  if(!u4.getQuery() ||
     !u4.getAbsolutePath() ||
     (strcmp(u4.getAbsolutePath(), "navigatie.php3") ||
      strcmp(u4.getQuery(), "Expand=true") ||
      u4.getScheme() ||
      u4.getHost() ||
      u4.getPort() ||
      u4.getUserInfo() ||
      u4.getRelativePath() ||
      u4.getFragment()))
    fail = showError(u4, fail);

  URI u5("ftp://p@awacs.dhs.org/music/balsam.mp3");
  if(!u5.getScheme() ||
     !u5.getHost() ||
     !u5.getUserInfo() ||
     !u5.getAbsolutePath() ||
     (strcmp(u5.getScheme(), "ftp") ||
      strcmp(u5.getHost(), "awacs.dhs.org") ||
      strcmp(u5.getUserInfo(), "p") ||
      strcmp(u5.getAbsolutePath(), "music/balsam.mp3") ||
      u5.getPort() ||
      u5.getRelativePath() ||
      u5.getQuery() ||
      u5.getFragment()))
    fail = showError(u5, fail);

  URI u6("mailto:a.haakmat@chello.nl");
  if(!u6.getScheme() ||
     !u6.getOpaquePart() ||
     (strcmp(u6.getScheme(), "mailto") ||
      strcmp(u6.getOpaquePart(), "a.haakmat@chello.nl") ||
      u6.getHost() ||
      u6.getPort() ||
      u6.getUserInfo() ||
      u6.getRelativePath() ||
      u6.getAbsolutePath() ||
      u6.getQuery() ||
      u6.getFragment()))
    fail = showError(u6, fail);

  fail = testPath("http://w.h.com/home/sub/comp#frag",
		  "/home/sub/comp", fail);

  fail = testPath("../home/sub/comp",
		  "../home/sub/comp", fail);

  fail = testPath("http://w.a.p/a;x/b;y=1/c;fg#a",
		  "/a/b/c", fail);

  fail = testPath("http://w.a.p/a;x/b;y=1/c#b",
		  "/a/b/c", fail);

  fail = testEscape("http://w.h.com/dck?x=a#crd",
		    "http%3a%2f%2fw.h.com%2fdck%3fx%3da%23crd",
		    fail);

  fail = testUnescape("http%3a%2f%2fw.h.com%2fdck%3fx%3d12%23crd",
		      "http://w.h.com/dck?x=12#crd",
		      fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g:h",
		     "g:h",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g",
		     "http://a/b/c/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "./g",
		     "http://a/b/c/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g/",
		     "http://a/b/c/g/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "/g",
		     "http://a/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "//g",
		     "http://g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "?y",
		     "http://a/b/c/?y",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g?y",
		     "http://a/b/c/g?y",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "#s",
		     "http://a/b/c/d;p?q#s",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g#s",
		     "http://a/b/c/g#s",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g?y#s",
		     "http://a/b/c/g?y#s",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     ";x",
		     "http://a/b/c/;x",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g;x",
		     "http://a/b/c/g;x",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g;x?y#s",
		     "http://a/b/c/g;x?y#s",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     ".",
		     "http://a/b/c/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "./",
		     "http://a/b/c/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "..",
		     "http://a/b/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "../",
		     "http://a/b/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "../g",
		     "http://a/b/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "../..",
		     "http://a/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "../../",
		     "http://a/",
		     fail);
  
  fail = testResolve("http://a/b/c/d;p?q",
		     "../../g",
		     "http://a/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "/./g",
		     "http://a/./g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "/../g",
		     "http://a/../g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g.",
		     "http://a/b/c/g.",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     ".g",
		     "http://a/b/c/.g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g..",
		     "http://a/b/c/g..",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "..g",
		     "http://a/b/c/..g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "./../g",
		     "http://a/b/g",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "./g/.",
		     "http://a/b/c/g/",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g/./h",
		     "http://a/b/c/g/h",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g/../h",
		     "http://a/b/c/h",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g;x=1/./y",
		     "http://a/b/c/g;x=1/y",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g;x=1/../y",
		     "http://a/b/c/y",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g?y/./x",
		     "http://a/b/c/g?y/./x",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g?y/../x",
		     "http://a/b/c/g?y/../x",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g#s/./x",
		     "http://a/b/c/g#s/./x",
		     fail);

  fail = testResolve("http://a/b/c/d;p?q",
		     "g#s/../x",
		     "http://a/b/c/g#s/../x",
		     fail);
  
  return fail;
}

int URI::testResolve(char *base, char *relative, char *expect, int fail) {
  URI *u1 = new URI(base);
  URI *u2 = new URI(relative);
  URI *u3 = resolve(u1, u2);

  if(u3) {
    if(strcmp(u3->toString(), expect)) {
      printf("resolve failed: got: %s, expected: %s\n", u3->toString(), expect);
      u2->dump();
      fail++;
    }

  } else {

      printf("resolve failed: got: (null), expected: %s\n", expect);

  }

  delete u1;
  delete u2;
  delete u3;

  return fail;
}

int URI::testPath(char *url, char *expect, int fail) {
  URI u(url);
  if(strcmp(expect, u.getPathWithoutParameters())) {
    printf("path creation failed: got: %s, expected: %s\n", u.getPathWithoutParameters(), expect);
    fail++;
  }

  return fail;
}

int URI::testEscape(char *s_in, char *expect, int fail) {
  char *s_out = URI::escape(s_in);

  if(strcmp(s_out, expect) != 0) {

    fail++;
    printf("incorrect escape: %s -> %s\n", s_in, s_out);

  }

  free(s_out);
  return fail;

}

int URI::testUnescape(char *s_in, char *expect, int fail) {
  char *s_out = URI::unescape(s_in);

  if(strcmp(s_out, expect) != 0) {

    fail++;
    printf("incorrect unescape: %s -> %s\n", s_in, s_out);

  }

  free(s_out);
  return fail;

}

int URI::showError(URI u, int fail) {
  printf("failed to parse URL - incorrect results:\n");
  u.dump();
  return fail+1;
}

/**
 * Dumps a representation of the URI object to stdout.
 */
void URI::dump() {
  printf("URI string: %s\n", toString());
  printf("isvalid:\t\t%s\n", (isValid() ? "true" : "false"));
  printf("scheme:\t\t\t%s\n", getScheme());
  printf("host:\t\t\t%s\n", getHost());
  printf("port:\t\t\t%s\n", getPort());
  printf("userinfo:\t\t%s\n", getUserInfo());
  printf("authority:\t\t%s\n", getAuthority());
  printf("abspath:\t\t%s\n", getAbsolutePath());
  printf("relpath:\t\t%s\n", getRelativePath());
  printf("path:\t\t\t%s\n", getPath());
  printf("query:\t\t\t%s\n", getQuery());
  printf("fragment:\t\t%s\n", getFragment());
  printf("opaque:\t\t\t%s\n", getOpaquePart());
}
