/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: URI.h,v 1.8 2001/02/24 20:23:01 pascal Exp $ 
 */

#ifndef __URI_H__
#define __URI_H__

#include "Garbage.h"

/**
 * This structure is used by the URI class to communicate 
 * with the C-based parser and scanner code.
 */
typedef struct {
  char *scheme;
  char *host;
  char *port;
  char *userInfo;
  char *absPath;
  char *relPath;
  char *query;
  char *fragment;
  char *opaque;
  GarbageTable garbage;
} URIParseResult;

/**
 * This class provides a high-level interface for querying and 
 * manipulating Uniform Resource Identifiers. The implementation 
 * tracks [RFC2396]. 
 * 
 * Typical usage:
 * \code
 * URI url("http://www.google.com");
 * if(url.isValid()) {
 *   if(url.getScheme() != NULL)
 *     printf("scheme: %s\n", url.getScheme()); // prints "http"
 *   if(url.getHost() != NULL)
 *     printf("host: %s\n", url.getHost()); // prints "www.google.com"
 * }
 * \endcode
 *
 * The constructor of the URI class parses it's string argument into
 * into it's components. If the parse is successful, then the URI is
 * valid. The isValid() method can be used to determine whether the
 * object represents a valid URI. The various components of the URI
 * can be retrieved using the `const' accessor methods. In addition,
 * there are three methods which perform additional processing to
 * return a result that is perhaps more useful than the raw URI
 * components; these are getAuthority(), getPath() and
 * getPathWithoutParameters().
 *
 * A number of setter methods make it possible to assign new values to
 * each URI component. Some of these setter methods may alter the
 * value returned by getAuthority(), getPath() and
 * getPathWithoutParameters(), but they never alter the value returned
 * by the isValid() method. If you create a URI object, then change
 * some of it's components and want to re-validate the result of your
 * modifications, use the validate() method.
 *
 * All accessor methods (getScheme(), getHost(), getPath(), etc.) may
 * return NULL when they are not defined. The method toString(), which
 * returns a string representation of the URI object, is guaranteed to
 * always return at least a zero-length string however.
 *
 * This class' constructor and the validate() method can not be
 * invoked reentrantly and are not thread-safe.  
 */
class URI {
 public:
  URI(const char *uri);
  ~URI();
  
  bool isValid() const;
  const char *getScheme() const;
  const char *getHost() const;
  const char *getPort() const;
  const char *getUserInfo() const;
  const char *getAbsolutePath() const;
  const char *getRelativePath() const;
  const char *getQuery() const;
  const char *getFragment() const;
  const char *getOpaquePart() const;

  const char *getPath();
  const char *getPathWithoutParameters();
  const char *getAuthority();
  const char *toString();

  void setScheme(const char *s);
  void setHost(const char *s);
  void setPort(const char *s);
  void setUserInfo(const char *s);
  void setAbsolutePath(const char *s);
  void setRelativePath(const char *s);
  void setQuery(const char *s);
  void setFragment(const char *s);
  void setOpaquePart(const char *s);

  bool validate();

  static char *escape(const char *s);
  static char *unescape(const char *s);
  static URI *resolve(URI *baseURI, URI *relativeURI);

  static int test();
  void dump();

 private:
  char *uri_string;
  char *path;
  char *path_without_parameters;
  char *authority;
  bool valid;
  URIParseResult parseResult;

  void flush();
  void flushDerivatives();
  void setComponent(char **s1, const char *s2);
  bool validate(const char *uri);

  static void stripDoubleDots(char *s);
  static void stripSingleDots(char *s);
  static void copyAndStripParameters(char *dest, const char *source);
  static bool isMustEscapeChar(int b);
  static int testResolve(char *base, char *relative, char *expect, int fail);
  static int testPath(char *url, char *expect, int fail);
  static int testEscape(char *s_in, char *expect, int fail);
  static int testUnescape(char *s_in, char *expect, int fail);
  static int showError(URI u, int fail);  

};

#endif
