/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: DTDDataStructures.h,v 1.2 2001/01/27 14:36:40 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph (CRAZY BAD #define!)
 */

#ifndef __DTDDATASTRUCTURES_H__
#define __DTDDATASTRUCTURES_H__

/** 
 * This class makes it possible to write down a dtd by 
 * initializing two structs: DTDDataStructures::Elements and 
 * DTDDataStructures::Attributes. DTDDataStructures::Elements 
 * defines all tags and all possible tags that can exist in their 
 * body with the desired order and  DTDDataStructures::Attributes
 * defines all the required, default and normal attributes of the
 * elements in the dtd.
 */
class DTDDataStructures {

public:

  /** this enum is used by the Elements struct */
  enum {

    error = -1,
    ok = -2

  };

  /** this struct is used by the Elements struct */
  struct BodyList {

    char tag[15];
    int success;
    int failed;
  
  };

  /** 
   * This struct defines the body of a xml element. To explain how
   * the struct works an example: let's say the dtd of a xml language
   * has the following element definition:
   * \code
   * <!ELEMENT contact (phonenumber+, address?)>
   * \endcode
   * (which means: one or more <phonenumber> tags followed by zero or
   * more <address> tags)<br>
   * the struct would look something like:
   * \code
   * struct DTDDataStructures::Elements example[] = {
   *
   *   {"contact", {
   *                {"phonenumber",1,DTDDataStructures::error},
   *                {"phonenumber",1,2},
   *                {"address",2,DTDDataStructures::ok}
   *               },
   *   {"address", { ...
   * }
   * \endcode
   * which means:  when looking at the body of the "contact" element:<br>
   * (state 0:) look for the <phonenumber> tag. When this is found go to
   * state 1 otherwise generate an error.<br>
   * <br>
   * (state 1:) look for the <phonenumber> tag. When this is found go to
   * state 1 otherwise go to state 2.<br>
   * <br>
   * (state 2:) look for the <address> tag. When this is found go to
   * state 2 otherwise </contact> must follow.<br>
   * <br>
   * which is exactly (phonenumber+, address?).<br>
   * #PCDATA is annotated with "", for example:
   * \code
   *                 {"",1,DTDDataStructures::ok}
   * \endcode
   */
  struct Elements {

    char tag[15];
    struct BodyList bodyList[17];

  };

  /** this struct is used by the Attributes struct */
  struct RequiredAttributes {

    char attribute[15];

  }; 

  /** this struct is used by the Attributes struct */
  struct DefaultAttributes {

    char attribute[15];
    char value[10];

  }; 

  /** this type is used by the Attributes struct */
  typedef int AttributeType;

  /** this struct is used by the Attributes struct */
  struct AllAttributes {

    char attribute[17];
    AttributeType type;

  }; 

  /** 
   * This struct defines the attributes of a xml element. To explain
   * how the struct works an example: let's say the dtd of a xml
   * language has the following attributes definition:
   * \code
   * <!ATTLIST p
   *   align (left|right)  "left"
   *   mode  (wrap|nowrap) #REQUIRED
   *   width CDATA         #IMPLIED
   *   >
   * \endcode
   * which means:<br>
   *   the align attribute can be 'left' or 'right' when the 
   *     attribute is not given it will be 'left'<br>
   *   the mode attribute is required<br>
   *   the width attributes can be any character data<br>
   * So
   * \code
   *  <p mode='nowrap' width='bla'>
   * \endcode
   * is possible in the xml document.<br>
   * The Attributes struct would look something like this:
   * \code
   * struct DTDDataStructures::Attributes example[] = {
   *   {"p", {{"mode"}},
   *         {{"align","left"}},
   *         {{"align",SomeAttributeType::alignMethod},
   *          {"mode", SomeAttributeType::wrapMethod},
   *          {"width",SomeAttributeType::cdata}}}
   * }
   * \endcode
   * So first the tag name ("p"),<br>
   * then the required attributes ({{"mode"}})<br>
   * then the attributes with default values and their default values
   *  ({{"align","left"}})<br>
   * and finally all attributes with their types.
   */
  struct Attributes {

    char tag[15];
    struct RequiredAttributes requiredAttributes[3];
    struct DefaultAttributes  defaultAttributes[3];
    struct AllAttributes      allAttributes[14];

  };

};

#endif
