/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: WMLParser.cpp,v 1.13 2001/01/27 14:36:40 wouter Exp $
 *
 * Reviewed 3-12-2000 ph (one-line if's; NULLs)
 */

#include <string.h>

#include "../lib/String.h"
#include "XMLAttributeList.h"
#include "XMLBody.h"
#include "XMLCharData.h"
#include "XMLElement.h"
#include "XMLParser.h"
#include "WMLdtd12.h"
#include "WMLParseErrorException.h"
#include "WMLParser.h"

/**
 * Constructs a new WMLParser.
 *
 * @param source the source which must be parsed
 */
WMLParser::WMLParser(String *source) {

  topElement = new XMLElement();

  /* xml parse the source text: */
  XMLParser parser(source, topElement);

  /* get the error message (when parsing was not successfull) */
  errorMessage = new String(parser.getErrorMessage());

  if(errorMessage->getLength() != 0) {

    /* parse error!, what else can you do but... */
    delete topElement;
    topElement = NULL; 
    return;

  }

  /* when xml parsing was succesfull, wml parse: */

  try {

    if(!topElement->getTagName()->equals("wml"))
      generateError("the root element must be 'wml'.", topElement);

    /* check the parse tree with the dtd (recursively) */
    checkElementWithDTD(topElement);

  } catch (WMLParseErrorException) {

    delete topElement;
    topElement = NULL;
    return;

  }

  applyTemplateToCards();

  return;
}

/**
 * Delete the object
 */
WMLParser::~WMLParser() {

  delete errorMessage;
  delete topElement;

}

/**
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool WMLParser::test() {

  /* not tested: all attribute types, replacement of entities
     getParseTreeAsString(), applyTemplateToCards() */

  /* minimal test */
  if(!test("<wml><card></card></wml>",
           true))
    return false;

  /* wrong root element error */
  if(!test("<a><card></card></a>",
           false))
    return false;

  /* no such element error */
  if(!test("<wml><car></car></wml>",
           false))
    return false;

  /* no tasks in anchor error */
  if(!test("<wml><card><p><anchor></anchor></p></card></wml>",
           false))
    return false;

  /* one task in anchor (no error) */
  if(!test("<wml><card><p><anchor><prev/></anchor></p></card></wml>",
           true))
    return false;

  /* two tasks in anchor error */
  if(!test("<wml><card><p><anchor><prev/><refresh/></anchor></p></card></wml>",
           false))
    return false;

  /* empty body error */
  if(!test("<wml></wml>",
           false))
    return false;

  /* chardata in wrong place error */
  if(!test("<wml>f</wml>",
           false))
    return false;

  /* chardata in wrong place error */
  if(!test("<wml><card><p><do type='accept'><noop>f</noop></do></p></card></wml>",
           false))
    return false;

  /* unknown attribute error */
  if(!test("<wml><card unknow='123'></card></wml>",
           false))
    return false;

  /* required attribute missing error */
  if(!test("<wml><card><p><do><noop/></do></p></card></wml>",
           false))
    return false;

  /* wrong attribute type error */
  if(!test("<wml><card><p align='something'></p></card></wml>",
           false))
    return false;

  return true;

}

/**
 * Tests the wmlsource in the argument for parse errors.
 * @param source the wml source
 * @param shouldParse wheter or not the wmlsource should parse
 * @return true iff the source contains no errors and 
 * souldParse == true OR the source contains errors and
 * souldParse == false
 */
bool WMLParser::test(char* source, bool shouldParse) {

  String s(source);
  WMLParser wp(&s);

  if(wp.getErrorMessage() != NULL) 
    return (shouldParse?false:true);

  return (shouldParse?true:false);

}

/** 
 * Applies the template to all the cars in the deck
 */
void WMLParser::applyTemplateToCards() {

  XMLBody *xb = topElement->getBody();
  xb = xb->getElement("template");

  if(xb != NULL) {

    XMLAttributeList *templateAttributes = 
        xb->getElement("template")->getElement()->getAttributes();
    XMLBody *templateBody = 
        xb->getElement("template")->getElement()->getBody();

    xb = xb->getElement("card");
    while(xb != NULL) {

      /* append the template body to the card body:*/
      xb->getElement()->getBody()->append(templateBody);

      /* add the template attributes:*/

      /* when the onenterforward is given in the template: */
      if(templateAttributes->getValue("onenterforward")) {

        /* and not overriden in the card: */
        if(!xb->getElement()->getAttributes()->getValue("onenterforward")) {

          /* add the template attribute 'onenterforward' from the template: */
          xb->getElement()->getAttributes()->add("onenterforward",templateAttributes->getValue("onenterforward"));

        }
      }

      /* etc. */      
      if(templateAttributes->getValue("onenterbackward")) {

        if(!xb->getElement()->getAttributes()->getValue("onenterbackward")) {

          xb->getElement()->getAttributes()->add("onenterbackward",templateAttributes->getValue("onenterbackward"));

        }

      }

      if(templateAttributes->getValue("ontimer")) {

        if(!xb->getElement()->getAttributes()->getValue("ontimer")) {

          xb->getElement()->getAttributes()->add("ontimer",templateAttributes->getValue("ontimer"));

        }
      }

      if(templateAttributes->getValue("id")) {

        if(!xb->getElement()->getAttributes()->getValue("id")) {

          xb->getElement()->getAttributes()->add("id",templateAttributes->getValue("id"));

        }
      }
      
      xb = xb->getNextXMLBody();
      if(xb == NULL) break;
      xb = xb->getElement("card"); /* go to the next card */

    }
  }
}

/**
 * Gives access to the root element of the xml document, from this
 * root the rest of the document can be accesed.
 *
 * @return a pointer to a XMLElement containing the root element of
 * the xml document, do not delete 
 */
XMLElement *WMLParser::getRootElement() {

  return topElement;

}

/**
 * Gives access to a string with the error message. Returns NULL when
 * no error occured
 *
 * @return the string with the error message, do not delete 
 */
String *WMLParser::getErrorMessage() {

  if(errorMessage->getLength() == 0) 
    return NULL;

  return errorMessage;

}

/**
 * The string in the argument is first cleared and then it is filled
 * with the parse tree in textual form. 
 *
 * @param s the string that will be filled with the parse tree 
 */
void WMLParser::getParseTreeAsString(String *s) {  

  s->clear();
  if(errorMessage->getLength() != 0)
    s->append(errorMessage);
  if(topElement != NULL) 
    topElement->getParseTreeAsString(0, s);
  
}

/**
 * Generates a wml parse error. The XMLElement is supplied for extra
 * information (line number and tag name).
 * @param error the error text that must be replaced
 * @param xe the xml element in which the error occured 
 */
void WMLParser::generateError(char *err, XMLElement *xe) {

  errorMessage->append("line ");
  errorMessage->append(xe->getLineNumber());
  errorMessage->append(": wml error: ");
  errorMessage->append("<");
  errorMessage->append(xe->getTagName()->getPtr());
  errorMessage->append("> element: ");
  errorMessage->append((char *) err);
  throw WMLParseErrorException();

}

/**
 * Generates a wml parse error in which is mentioned what was read and
 * what was supposed to be read.
 * 
 * @param body the element in which the parse error took place 
 * @param lineNumber the linenumber on which the parse error took
 * place
 * @param ol the 'otherwise' list, a list in which all possible tags
 * for this possition are listed
 * @param index the current index in the otherwise list
 * @param insteadOf the string that was read and generated the error 
 */
void WMLParser::generateExpectedError(XMLElement *body, int lineNumber, DTDDataStructures::BodyList* ol, int index, String *insteadOf) {

  errorMessage->append("line ");
  errorMessage->append(lineNumber);
  errorMessage->append(": wml error: in <");
  errorMessage->append(body->getTagName()->getPtr());
  errorMessage->append("> body: ");

  if(insteadOf != NULL) {

    errorMessage->append("read element: ");
    errorMessage->append(insteadOf);
    errorMessage->append(", ");

  }

  errorMessage->append("expected: ");
  int oldindex = 0;
  int count = 0;

  if(index == DTDDataStructures::ok || strcmp(ol[index].tag,"0") == 0) {

    errorMessage->append("end of body.");

  } else {

    while(index != DTDDataStructures::error && index != DTDDataStructures::ok) {

      if(count == 0) 
        count = 1;
      else {

        if(count != 1)
          errorMessage->append(", ");
        else
          count = 2;

        if(!strcmp(ol[oldindex].tag, ""))
           errorMessage->append("character data");
        else {

          errorMessage->append("<");
          errorMessage->append(ol[oldindex].tag);
          errorMessage->append(">");

        }

      }
      oldindex = index;
      index = ol[index].failed;
    }
    if(count != 1) errorMessage->append(" or ");
    errorMessage->append("<");
    errorMessage->append(ol[oldindex].tag);
    errorMessage->append(">.");

  }

  throw WMLParseErrorException();

}

/**
 * Generates a wml parse error in which is mentioned that an attribute
 * that is required
 * was not found in the attribute list
 * 
 * @param xe the element in which the parse error took place
 * @param attributeName the attribute name that is required 
 */
void WMLParser::generateRequiredAttributeError(XMLElement *xe, String *attributeName) {

  errorMessage->append("line ");
  errorMessage->append(xe->getLineNumber());
  errorMessage->append(": wml error: in <");
  errorMessage->append(xe->getTagName()->getPtr());
  errorMessage->append("> attributes: ");
  errorMessage->append(attributeName);
  errorMessage->append(" is a required attribute.");

  delete attributeName;

  throw WMLParseErrorException();
}

/**
 * Generates a wml parse error in which is mentioned that an attribute
 * value is not well formed
 * @param xe the element in which the parse error took place
 * @param attributeName the attribute name which had the wrong value
 * @param attributeValues the possible values
 * @param insteadOf the string that was in stead found in the source 
 */
void WMLParser::generateWrongAttributeValueError(XMLElement *xe, String *attributeName, String *attributeValues, String *insteadOf) {

  errorMessage->append("line ");
  errorMessage->append(xe->getLineNumber());
  errorMessage->append(": wml error: in <");
  errorMessage->append(xe->getTagName()->getPtr());
  errorMessage->append("> attributes: ");
  errorMessage->append(attributeName);
  errorMessage->append(" = '");
  errorMessage->append(insteadOf->getPtr());
  errorMessage->append("': should be ");
  errorMessage->append(attributeValues->getPtr());
  throw WMLParseErrorException();

}

/**
 * This function determines whether the elements are formed according
 * to the dtd
 * @param the xml element to check (child elements are recursively
 * checked)
 */
void WMLParser::checkElementWithDTD(XMLElement *xe) {

  checkAttributesWithDTD(xe);

  DTDDataStructures::BodyList* el = getBodyList(xe->getTagName()->getPtr());
  /* this is a list of tags that may occur in the body of the current
     element,  with information on order etc. getEltList returns NULL
     when the current element does not exist in the DTD: */
  if(el == NULL)
    generateError("no such element.", xe);

  int index = 0; /* the position in the ELTLIST */

  XMLBody *xb = xe->getBody();

  if(xe->getTagName()->equals("anchor")) {

    /* anchor elements may only have one task in their body */
    if(checkBodyForDoubleTasks(xb))
      generateError("must have one task in the body of an anchor element.", xe);

  }

  if(xb == NULL || xb->getContentType() == XMLBody::typeNothing) {

    int oldindex = index;

    /* so the body of the element is empty, here we walk throuth the
       ELTLIST while recognizing nothing */
    while(el[index].failed != DTDDataStructures::error && el[index].failed != DTDDataStructures::ok)
      index = el[index].failed;

    /* when the dtd is ok with the fact that the body is empty, we 
       go on, otherwise a parse error: */
    if(el[index].failed == DTDDataStructures::error)
      generateExpectedError(xe, xe->getLineNumber(), el, oldindex, 0);

    return;

  }

  while(true) {

    if(xb->getContentType() == XMLBody::typeCharacterData) {

      if(index == DTDDataStructures::ok) {

        /* according to the dtd the body had to end now */
        String s("character data");
        generateExpectedError(xe, xb->getCharacterData()->getStartLineNumber(),el, DTDDataStructures::ok, &s);

      }
        
      int oldindex = index;

      while(strcmp(el[index].tag,"")) {
        
        if(el[index].failed == DTDDataStructures::error || el[index].failed == DTDDataStructures::ok) {

          String s("character data");
          generateExpectedError(xe, xb->getCharacterData()->getStartLineNumber(),el, oldindex, &s);

        }

        index = el[index].failed;

      }

      /* so it was ok to find character data in the body, now replace
         the entities with their specified characters:*/
      if(xb->getCharacterData()->getMustReplaceEntities()) 
        replaceEntities(xb->getCharacterDataAsString(),xb->getCharacterData()->getStartLineNumber(),xb->getCharacterData()->getEndLineNumber());

    } else { /* content type is an element */

      int oldindex = index;

      while(strcmp(el[index].tag, xb->getElement()->getTagName()->getPtr())) {

        if(index == DTDDataStructures::ok)          
          generateExpectedError(xe, xb->getElement()->getLineNumber(), el, DTDDataStructures::ok, xb->getElement()->getTagName());

        if(el[index].failed == DTDDataStructures::ok || el[index].failed == DTDDataStructures::error) 
          generateExpectedError(xe, xb->getElement()->getLineNumber(), el, oldindex, xb->getElement()->getTagName());

        index = el[index].failed;

      }

      /* so it was ok to find this tag in the body, now check this
         element, it's attributes and its body with the dtd: */
      checkElementWithDTD(xb->getElement());

    }

    /* now we go on with the rest of the body: */
    index = el[index].success;

    if(xb->getNextXMLBody() == NULL) {

      if(index == DTDDataStructures::ok) return;

      int oldindex = index;

      while(el[index].failed != DTDDataStructures::error && el[index].failed != DTDDataStructures::ok)
        index = el[index].failed;

      if(el[index].failed == DTDDataStructures::error) {

        int l;

        /* so their should have been more in the body,
           generate an error with the correct line number: */

        if(xb->getCharacterData() != NULL)
          l = xb->getCharacterData()->getStartLineNumber();
        else
          l = xb->getElement()->getLineNumber();

        generateExpectedError(xe, l,el, oldindex, 0);

      }

      return;

    }

    xb = xb->getNextXMLBody();

  }

}

/**
 * This function determines whether the attibutes are formed according
 * to the dtd
 * @param the xml element to check 
 */
void WMLParser::checkAttributesWithDTD(XMLElement *xe) {

  DTDDataStructures::RequiredAttributes* req = 
    getRequiredAttributes(xe->getTagName()->getPtr());
  DTDDataStructures::DefaultAttributes* def =
    getDefaultAttributes(xe->getTagName()->getPtr()); 
  DTDDataStructures::AllAttributes* rest = 
    getAllAttributes(xe->getTagName()->getPtr());

  if(req == NULL || def == NULL || rest == NULL) 
    return; /* this can only happen when the tag does not
               exist in the dtd, and this is already checked
               by checkElementWithDTD, so this will never happen,
               only here for debugging purposes. */

  XMLAttributeList *oldAttributes = xe->getAttributes();
  XMLAttributeList *newAttributes = new XMLAttributeList();

  int i;

  for(i = 0; strcmp(rest[i].attribute, ""); i++) {

    String *attrName = new String(rest[i].attribute);
    String *attrValue = oldAttributes->getValue(attrName->getPtr());
  
    if(attrValue != 0) {

      /* putting all attributes from the spec in the new list */
      String *s = new String(attrValue);
      oldAttributes->remove(attrName);
      newAttributes->add(attrName->getPtr(), s);
      delete s;

    }

    delete attrName;

  }

  if(oldAttributes->getName() != NULL) {
    /* when there are attributes left, generate an error */
    errorMessage->append("line ");
    errorMessage->append(xe->getLineNumber());
    errorMessage->append(": wml error: in <");
    errorMessage->append(xe->getTagName());
    errorMessage->append("> attributes: attribute \"");
    errorMessage->append(oldAttributes->getName());
    if(newAttributes->getValue(oldAttributes->getName()->getPtr()))
      errorMessage->append("\" used more than once.");
    else
      errorMessage->append("\" not known.");

    delete newAttributes;

    throw WMLParseErrorException();

  }

  xe->changeAttributes(newAttributes); /* setting the new attributes */

  for(i = 0; strcmp(req[i].attribute, ""); i++) {

    /* here we walk through a list of all required attributes 
       an error is generated when these attributes are not found
       here */
     
    String *attrName = new String(req[i].attribute);
    String *attrValue = newAttributes->getValue(attrName->getPtr());

    if(attrValue == NULL) generateRequiredAttributeError(xe, attrName);
      
    delete attrName;

  }

  for(i = 0; strcmp(def[i].attribute, ""); i++) { 

    /* here we walk through a list of all attributes that have 
       default values, when the attributes are not found here we
       must add them with the appropriate values */

    String *attrName = new String(def[i].attribute);
    String *attrValue = newAttributes->getValue(attrName->getPtr());

    if(attrValue == NULL) {

      attrValue = new String(def[i].value);
      newAttributes->add(attrName->getPtr(),attrValue);
      delete attrValue;

    } 
    delete attrName;

  }

  String attrName;

  for(i = 0; strcmp(rest[i].attribute, ""); i++) {

    /* here we walk through a list of all attributes, we must
       check the values of the attributes with their types */

    attrName.clear();
    attrName.append(rest[i].attribute);

    String *attrValue = newAttributes->getValue(attrName.getPtr());

    if(attrValue != NULL) {

      checkAttributeTypeWithDTD(xe, &attrName, attrValue, rest[i].type);

    } 

  }

}

/** 
 * This function determines whether the attribute value is formed
 * according to the dtd
 * @param xe the element from which the attributes must be checked
 * @param attrName the name of the attribute
 * @param attrValue the value of the attribute
 * @param at the type of the attribute 
 */
void WMLParser::checkAttributeTypeWithDTD(XMLElement *xe, String *attrName, String *attrValue, DTDDataStructures::AttributeType attributeType) {

  String s;

  switch (attributeType) {

  case WMLAttributeType::ignored:
    xe->getAttributes()->remove(attrName);
    break;

  case WMLAttributeType::booleans:
    s.append("'true' or 'false'");
    if(!attrValue->equals("true") && !attrValue->equals("false"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::gomethod:
    s.append("'post' or 'get'");
    if(!attrValue->equals("post") && !attrValue->equals("get"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::inputtype:
    s.append("'text' or 'password'");
    if(!attrValue->equals("text") && !attrValue->equals("password"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::wrapmode:
    s.append("'wrap' or 'nowrap'");
    if(!attrValue->equals("wrap") && !attrValue->equals("nowrap"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::imgalign:
    s.append("'top','middle' or 'bottom'");
    if(!attrValue->equals("top") && !attrValue->equals("middle") && !attrValue->equals("bottom"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::palign:
    s.append("'left','center' or 'right'");
    if(!attrValue->equals("left") && !attrValue->equals("center") && !attrValue->equals("right"))
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);
    break;

  case WMLAttributeType::length:
    if(attrValue->getLastCharacter() == '%') {

      String initAttrValue(attrValue);
      initAttrValue.removeLastCharacter();

      if(!initAttrValue.isNumber()) {
        s.append("a number + %");
        generateWrongAttributeValueError(xe, attrName, &s, attrValue);
      }
      
      break;
    }

    if(!attrValue->isNumber()) {

      s.append("a number");
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);

    }
    break;

  case WMLAttributeType::number:
    if(!attrValue->isNumber()) {

      s.append("a number");
      generateWrongAttributeValueError(xe, attrName, &s, attrValue);

    }
    break;

  case WMLAttributeType::cdata:
    String oldValue(xe->getAttributes()->getValue(attrName->getPtr()));
    xe->getAttributes()->remove(attrName);
    replaceEntities(&oldValue, xe->getLineNumber(),xe->getLineNumber());
    xe->getAttributes()->add(attrName->getPtr(),&oldValue);
    break;

  }

}

/**
 * Replaces the entities in the string. For Example &gt; must be 
 * replaced by '>'/ Line numbers are used for possible error messages.
 * @param cdata the string
 * @param startLineNumber the starting line number of the character data
 * @param endLineNumber the ending line number of the character data 
 */
void WMLParser::replaceEntities(String *cdata, int startLineNumber, int endLineNumber) {

  String pcdata;
  char entity[maxReferenceLength];
  char *current;
  char *end = cdata->getPtr() + cdata->getLength();

  for(current = cdata->getPtr(); current < end; current++) {

    /* entities start with '&' */
    if(*current != '&') {

      /* no '&', so continue */
      pcdata.append(*current);

    } else {

      /* a '&', start reading the entity */

      current++;
      int i;

      /* we must read on till we find a ';' or their is no more data */
      for(i = 0; current < end && *current != ';'; i++, current++) {

        if(i == maxReferenceLength) {

          entity[maxReferenceLength -1] = 0;
          errorMessage->append("line ");
          errorMessage->append(startLineNumber);

          if(startLineNumber != endLineNumber) {

            errorMessage->append(" - ");
            errorMessage->append(endLineNumber);

          }

          errorMessage->append(": wml error: did not recognize reference: ");
          errorMessage->append(entity);

          throw WMLParseErrorException();

        }
        entity[i] = *current;

      }

      if(current == end) {

        errorMessage->append("line ");
        errorMessage->append(startLineNumber);

        if(startLineNumber != endLineNumber) {

          errorMessage->append(" - ");
          errorMessage->append(endLineNumber);

        }

        errorMessage->append(": wml error: reference with '&' but without closing ';'.");
        throw WMLParseErrorException();

      }

      entity[i] = 0;

      replaceEntity(entity, &pcdata, startLineNumber, endLineNumber);
      
    } 

  }

  cdata->clear();
  cdata->append(&pcdata);

  return;

}

/**
 * This function appends the correct string to pcdata for the entity
 * that has been read by replaceEntities().
 * @param entity the string that has just been read by replaceEntities()
 * @param pcdata the string to append to
 * @param startLineNumber the starting line number of the character data
 * @param endLineNumber the ending line number of the character data 
 */
void WMLParser::replaceEntity(char *entity, String* pcdata, int startLineNumber, int endLineNumber) {
    
  if(strcmp(entity,"quot") == 0) {

    pcdata->append('"');
    return;

  }
  if(strcmp(entity,"amp") == 0) {

    pcdata->append('&');
    return;
  
  }
  if(strcmp(entity,"apos") == 0) {

    pcdata->append("'");
    return;
  
  }
  if(strcmp(entity,"lt") == 0) {

    pcdata->append('<');
    return;
  
  }
  if(strcmp(entity,"gt") == 0) {

    pcdata->append('>');
    return;
  
  }
  if(strcmp(entity,"nbsp") == 0) {

    pcdata->append((char)160);
    return;
  
  }
  if(strcmp(entity,"shy") == 0) {

    pcdata->append('-');
    return;
  
  }

  /* So the entity was no named character entity, other possibilities are
     decimal numeric character entities (&#32;) and hexadecimal numeric 
     character entities (&#xa3;) */
  if(entity[0] != '#') {

    /* entity does not start with '#', parse error */

    errorMessage->append("line ");
    errorMessage->append(startLineNumber);
    if(startLineNumber != endLineNumber) {

      errorMessage->append(" - ");
      errorMessage->append(endLineNumber);

    }
    errorMessage->append(": wml error: did not recognize reference: ");
    errorMessage->append(entity);
    throw WMLParseErrorException();

  }

  if(String::isNumber(entity[1])) {

    /* decimal numeric character entiy for example: &#65; (='A')*/

    String number(&(entity[1]));

    if(!number.isNumber()) {

      errorMessage->append("line ");
      errorMessage->append(startLineNumber);
      if(startLineNumber != endLineNumber) {

        errorMessage->append(" - ");
        errorMessage->append(endLineNumber);

      }
      errorMessage->append(": wml error: did not recognize reference: ");
      errorMessage->append(entity);
      throw WMLParseErrorException();
      
    }

    int i = number.getInt();

    if(i == 0) return;

    pcdata->append((char) i);

    return;

  }

  /* only hexadecimal numeric character entity still possible */

  if(entity[1] != 'x') {

    /* entity does not start with '#x', parse error */

    errorMessage->append("line ");
    errorMessage->append(startLineNumber);

    if(startLineNumber != endLineNumber) {

      errorMessage->append(" - ");
      errorMessage->append(endLineNumber);

    }

    errorMessage->append(": wml error: did not recognize reference: ");
    errorMessage->append(entity);
    throw WMLParseErrorException();

  }

  String hexNumber(&(entity[2])); 
  
  if(hexNumber.getLength() > 4 || !hexNumber.isHexNumber()) {

    errorMessage->append("line ");
    errorMessage->append(startLineNumber);
    if(startLineNumber != endLineNumber) {

      errorMessage->append(" - ");
      errorMessage->append(endLineNumber);

    }
    errorMessage->append(": wml error: did not recognize reference: ");
    errorMessage->append(entity);
    throw WMLParseErrorException();

  }

  /* hexadecimal numeric character entity for example: &#x41; (='A') */

  int i = hexNumber.getHexInt();

  if(i == 0) return;

  pcdata->append((char) i);

}

/**
 * Returns the pointer to the DTDDataStructures::BodyList list for
 * the given tag
 * @param tag the tag
 * @return the pointer to the list
 */
DTDDataStructures::BodyList *WMLParser::getBodyList(char *tag) {
  int i;
  int size = sizeof(wmleltspec) / sizeof(wmleltspec[0]);

  for(i = 0; i < size; i++) {

    if(strcmp(tag, wmleltspec[i].tag) == 0) 
      return wmleltspec[i].bodyList;

  }

  return NULL;

}

/**
 * Returns a pointer to the the required attribute list of the given
 * tag
 * @param tag the tag
 * @return the pointer to the required attribute list 
 */
DTDDataStructures::RequiredAttributes *WMLParser::getRequiredAttributes(char *tag) {
  int i;
  int size = sizeof(wmlattrspec) / sizeof(wmlattrspec[0]);

  for(i = 0;i < size; i++)
    if(strcmp(tag, wmlattrspec[i].tag) == 0) 
      return wmlattrspec[i].requiredAttributes;

  return NULL;

}

/**
 * Returns a pointer to the the list of attributes with default values
 * for the given tag 
 * @param tag the tag
 * @return the pointer to list of attributes with default values 
 */
DTDDataStructures::DefaultAttributes *WMLParser::getDefaultAttributes(char *tag) {
  int i;
  int size = sizeof(wmlattrspec) / sizeof(wmlattrspec[0]);

  for(i = 0;i < size; i++)
    if(strcmp(tag, wmlattrspec[i].tag) == 0) 
      return wmlattrspec[i].defaultAttributes;

  return NULL;

}

/**
 * Returns a pointer to the the list of all the attributes for the
 * given tag 
 * @param tag the tag 
 * @return the pointer to list of attributes with default values 
 */
DTDDataStructures::AllAttributes* WMLParser::getAllAttributes(char *tag) {
  int i;
  int size = sizeof(wmlattrspec) / sizeof(wmlattrspec[0]);

  for(i = 0;i < size; i++)
    if(strcmp(tag, wmlattrspec[i].tag) == 0) 
      return wmlattrspec[i].allAttributes;

  return NULL;

}

/**
 * Returns information on the amount of task elements in the xmlbody
 * @param  xb  the body in which must be searched for the tasks
 * @return false when there is one element, otherwise true
 */
bool WMLParser::checkBodyForDoubleTasks(XMLBody *xb) {
  int count;

  for(count = 0; xb; xb = xb->getNextXMLBody()) {

    if(xb->getContentType() == XMLBody::typeElement && 
       (xb->getElement()->getTagName()->equals("go") ||
        xb->getElement()->getTagName()->equals("prev") ||
        xb->getElement()->getTagName()->equals("refresh")))
        count++;
  }

  return count != 1;

}
