/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLAttributeList.cpp,v 1.1 2001/02/11 19:11:28 wouter Exp $
 *
 * Reviewed 4-12-2000 ph (NULLs, one-line if's)
 */

#include <stdlib.h>

#include "../lib/String.h"
#include "XMLAttributeList.h"
#include "XMLElement.h"
#include "XMLParser.h"

/**
 * Constructs a new XMLAttributeList.
 */
XMLAttributeList::XMLAttributeList() {
  
  attributeName = NULL;
  attributeValue = NULL;
  nextAttribute = NULL;
  
}

/**
 * Constructs a new XMLAttributeList containing the same arguments as the
 * one in the argument.
 */
XMLAttributeList::XMLAttributeList(XMLAttributeList *xa) {

  if(xa->getName() != NULL)
    attributeName = new String(xa->getName());
  else
    attributeName = NULL;

  if(attributeName && xa->getValue(attributeName->getPtr()))
    attributeValue = new String(xa->getValue(attributeName->getPtr()));
  else
    attributeValue = NULL;

  if(xa->getNext() != NULL)
    nextAttribute = new XMLAttributeList(xa->getNext());
  else
    nextAttribute = NULL;

}

/**
 * Delete the object.
 */
XMLAttributeList::~XMLAttributeList() {

  delete nextAttribute;
  delete attributeName;
  delete attributeValue;

}

/**
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool XMLAttributeList::test() {

  String s("<tag a = \"1\" b = \"2\"/>");
  XMLElement xe;
  XMLParser xp(&s, &xe);
  XMLAttributeList *xa = xe.getAttributes();

  if(!xa->getName()->equals("a")) return false;

  s.clear();
  s.append("a");

  xa->remove(&s);

  if(!xa->getName()->equals("b")) return false;

  s.clear();
  s.append("b");

  String *p = xa->getValue(s.getPtr());

  if(!p->equals("2")) return false;

  xa->remove(&s);

  s.clear();
  s.append("a");

  xa->add(s.getPtr(), &s);

  p = xa->getValue(s.getPtr());

  if(!p->equals("a")) return false;

  s.append('b');

  xa->add(s.getPtr(), &s);

  p = xa->getValue(s.getPtr());

  if(!p->equals("ab")) return false;

  return true;

}

/** 
 * Returns the next pointer of the first xmlattribute element.
 *
 * @return the xmlattribute pointer
 */
XMLAttributeList *XMLAttributeList::getNext() {

  return nextAttribute;

}

/** 
 * This function is part of the xml parser.
 * It parses the attributes of an element.
 * For example the 
 * \code
 *  class = 'main' id='first'>
 * \endcode
 * part of
 * \code
 *  <card class = 'main' id='first'>
 * \endcode
 * @param tokenizer the tokenizer to use
 * @param parser the parser, this parameter is used to 
 * add errors
 *
 */  
XMLAttributeList::XMLAttributeList(XMLTokenizer *tokenizer, XMLParser *parser) {

  attributeName = NULL;
  attributeValue = NULL;
  nextAttribute = NULL;

  /* first read the attribute name */
  if(tokenizer->getNextToken() != XMLTokenizer::tokenName) {

    /* if there was no name it could be the '>' or '/> */
    if(tokenizer->getLastReadToken() == XMLTokenizer::tokenCloseEmptyTag ||
       tokenizer->getLastReadToken() == XMLTokenizer::tokenCloseTag)
      return;

    parser->generateUnexpectedTokenError();

  }

  attributeName = new String(tokenizer->getTokenAsString());

  /* now we can read the '=' */
  if(tokenizer->getNextToken() != XMLTokenizer::tokenEquals)
    parser->generateUnexpectedTokenError(); 

  /* now we can read the attribute value */
  if(tokenizer->getNextToken() != XMLTokenizer::tokenStringConstant)
    parser->generateUnexpectedTokenError();

  attributeValue = new String(tokenizer->getTokenAsString());

  nextAttribute = new XMLAttributeList(tokenizer, parser);
  
  if(nextAttribute->getName() == NULL) {

    /* if there is nothing left of the attributes to parse 
       the next element is empty */
    /* now we can delete it */
    delete nextAttribute;
    nextAttribute = NULL;

  }

}

/**
 * Adds one attribute to the attributelist.
 *
 * @param name the name of the attribute
 * @param value the value of the attribute
 */
void XMLAttributeList::add(char *name, String *value) {

  if(attributeName == NULL) {

    attributeName = new String(name);
    attributeValue = new String(value);
    return;

  }
  if(nextAttribute == NULL)
    nextAttribute = new XMLAttributeList();

  nextAttribute->add(name, value);

}

/** 
 * Removes the attribute with the given attribute name from this
 * instance of XMLAttributeList.
 *
 * @param name the name of the attribute 
 */
void XMLAttributeList::remove(String *name) {

  if(nextAttribute == NULL) {

    if(name->equals(attributeName)) {

      delete attributeName;
      attributeName = NULL;
      delete attributeValue;
      attributeValue = NULL;

    }
    return;

  } else {

    /* so there is a next one */
    if(name->equals(attributeName)) {

      delete attributeName;
      attributeName = new String(nextAttribute->attributeName);

      delete attributeValue;
      attributeValue = new String(nextAttribute->attributeValue);

      nextAttribute->remove(attributeName);
  
    } 
    else
      nextAttribute->remove(name);

    if(nextAttribute->getName() == NULL) {

      /* if the next attribute was the last one it becomes empty */
      /* now we can delete it */
      delete nextAttribute;
      nextAttribute = NULL;
    }

  }

}

/**
 * Gets the attribute value of the attribute in the list with the
 * given attribute name. When the attribute name is not in the list
 * this function returns NULL.
 *
 * @param name the name of the attribute
 * @return a pointer to the String object containing the
 * attribute value, do not delete 
 */
String *XMLAttributeList::getValue(char *name) {

  if(attributeName == NULL) return NULL;

  if(attributeName->equals(name))
    return attributeValue;

  if(nextAttribute == NULL) return NULL;

  return nextAttribute->getValue(name);

}

/**
 * Gets the attribute name of the first attribute. When
 * the attribute list is empty this function returns NULL.
 *
 * @return a pointer to the String object containing 
 * the attribute name, do not delete
 */
String *XMLAttributeList::getName() {
  
  return attributeName;

}

/**
 * The string in the argument is filled with the parse tree in text. 
 *
 * @param s the string that will be filled with the parse tree
 */
void XMLAttributeList::getParseTreeAsString(String *s) {
  
  s->append(attributeName);
  s->append('=');
  s->append(attributeValue);

  if(nextAttribute != NULL) {

    s->append(' ');
    nextAttribute->getParseTreeAsString(s);

  }

}
