/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLElement.cpp,v 1.8 2001/01/27 14:36:41 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph 
 */

#include <stdlib.h>

#include "../lib/String.h"
#include "XMLAttributeList.h"
#include "XMLBody.h"
#include "XMLElement.h"
#include "XMLParser.h"
#include "XMLTokenizer.h"

/** 
 * Constructs a new XMLElement.
 */
XMLElement::XMLElement() {

  elementTagName = NULL;
  elementAttributes = NULL;
  elementBody = NULL;

}

/**
 * Constructs a new XMLElement with the same data as the one in the argument.
 */
XMLElement::XMLElement(XMLElement *xe) {

  elementTagName = new String(xe->getTagName());
  elementAttributes = new XMLAttributeList(xe->getAttributes());
  elementBody = new XMLBody(xe->getBody());
  elementLineNumber = xe->getLineNumber();

}

/**
 * Delete the object
 */
XMLElement::~XMLElement() {

  delete elementTagName;
  delete elementAttributes;
  delete elementBody;

}

/**
 * This function tests the code of this class
 *
 * @return      true when all tests succeed, otherwise false
 */
bool XMLElement::test() {
  String *s = new String("<tag a = \"1\" b = \"2\"/>");
  XMLElement *xe = new XMLElement();
  XMLParser *xp = new XMLParser(s, xe);

  delete s;

  if(!xe->getTagName()->equals("tag")) return false;

  delete xp;

  delete xe;

  return true;

}


/**
 * This function parses a complete element without the starting token.
 *
 * @param tag the tag that has just been read from the tokenizer
 * @param tokenizer the tokenizer to use 
 * @param parser the parser, this parameter is used to add warnings and errors
 */
void XMLElement::parse(String *tag, XMLTokenizer *tokenizer, XMLParser *parser) {
  
  elementTagName = new String(tag); /* store the tag name */
  elementLineNumber = tokenizer->getLineNumber();
  elementBody = NULL; 
  elementAttributes = new XMLAttributeList(tokenizer, parser);
  /* read the attributes */

  elementBody = new XMLBody();

  if(tokenizer->getLastReadToken() == XMLTokenizer::tokenCloseEmptyTag) 
    /* if this was an empty tag there is no body, so return */
    return;

  elementBody->parse(elementTagName, tokenizer, parser); /* read the body */

}

/** 
 * Changes the attributes of the element. The original one is deleted.
 *
 * @param xa the new attributes 
 */
void XMLElement::changeAttributes(XMLAttributeList *xa) {

  delete elementAttributes;
  elementAttributes = xa;

}

/**
 * Gives access to the attributes of this xml element.
 *
 * @return a pointer to a XMLAttributes object containing the
 * attributes, do not delete 
 */
XMLAttributeList *XMLElement::getAttributes() {

  return elementAttributes;

}

/** 
 * Gives access to the body of this xml element.
 *
 * @return a pointer to a XMLBody object containing the body, do not
 * delete 
 */
XMLBody *XMLElement::getBody() {

  return elementBody;

}

/**
 * Gives access to the tag name of this xml element.
 *
 * @return a pointer to a String object containing the name of the
 * tag, do not delete 
 */
String *XMLElement::getTagName() {

  return elementTagName;

}

/** 
 * The string in the argument is filled with the parse tree in text.
 *
 * @param s the string that will be filled with the parse tree
 * @param depth the depth of the tree, used for leading spaces.
 */
void XMLElement::getParseTreeAsString(int depth, String *s) {
  
  int i;

  for(i = depth; i > 0; i--)
    s->append(' '); /* depth is used to append spaces to the 
		       beginning of the line now it is possible 
		       to see who is a child of whom */
  
  s->append("tag: ");
  s->append(elementTagName);
  if(elementAttributes->getName() != NULL) { /* print attributes 
						when they are pressent */

    s->append(", attributes: ");
    elementAttributes->getParseTreeAsString(s);

  }

  s->append("\n");
  if(elementBody != NULL) 
    elementBody->getParseTreeAsString(depth + 1, s);

}

/** 
 * Returns the starting line number of the element.
 *
 * @return the line number
 */
int XMLElement::getLineNumber() {

  return elementLineNumber;

}
