/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 *      - Redistributions of source code must retain the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials
 *      provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this
 *      software must display the following acknowledgement: This
 *      product includes software developed by catchy.net and its
 *      contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors
 *      may be used to endorse or promote products derived from this
 *      software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: XMLTokenizer.cpp,v 1.12 2001/03/09 12:42:19 wouter Exp $ 
 *
 * Reviewed 4-12-2000 ph
 */

#include "CharStream.h"
#include "../lib/String.h"
#include "XMLEndOfFileException.h"
#include "XMLTokenizer.h"

/** 
 * Constructs a new XMLTokenizer.
 *
 * @param source the source which must be tokenized
 */
XMLTokenizer::XMLTokenizer(String *source) {

  characterStream = new CharStream(source);
  tokenAsString = new String();
  lastReadToken = tokenNothing;
  tokenStartLineNumber = tokenNothing;
  tokenStartLineNumber = 1;

}

/**
 * Delete the object
 */
XMLTokenizer::~XMLTokenizer() {

  delete characterStream;
  delete tokenAsString;

}

/** 
 * This function tests the code of this class
 *
 * @return true when all tests succeed, otherwise false
 */
bool XMLTokenizer::test() {
  String source("<?xml version=\"1.0\"?><!p<!DOCTYPE link><!--comment -- comment --><link target=\"some string\" style='big'>this is<?pi hello?>a link</link>");
  XMLTokenizer xt(&source);
  
  if(xt.isEndOfFile() != false) return false;

  if(xt.getNextToken() != tokenXMLOpen) return false;

  if(xt.getNextToken() != tokenName) return false;

  String* s = xt.getTokenAsString();

  if(!s->equals("version")) return false;
  
  if(xt.getNextToken() != tokenEquals) return false;
  
  if(xt.getNextToken() != tokenStringConstant) return false;
  
  if(!s->equals("1.0")) return false;
  
  if(xt.getNextToken() != tokenXMLClose) return false;
  
  if(xt.getNextToken() != tokenUnknownToken) return false;

  if(!s->equals("<!p")) return false;

  if(xt.getNextToken() != tokenDocTypeOpen) return false;
  
  if(xt.getNextToken() != tokenName) return false;
  
  if(!s->equals("link")) return false;
  
  if(xt.getNextToken() != tokenCloseTag) return false;

  if(xt.getNextToken() != tokenComment) return false;
  
  if(xt.getNextToken() != tokenTagOpen) return false;
  
  if(!s->equals("link")) return false;

  if(xt.getNextToken() != tokenName) return false;
  
  if(!s->equals("target")) return false;
  
  if(xt.getNextToken() != tokenEquals) return false;
  
  if(xt.getNextToken() != tokenStringConstant) return false;
  
  if(!s->equals("some string")) return false;
  
  if(xt.getNextToken() != tokenName) return false;
  
  if(!s->equals("style")) return false;
  
  if(xt.getNextToken() != tokenEquals) return false;
  
  if(xt.getNextToken() != tokenStringConstant) return false;
  
  if(!s->equals("big")) return false;
  
  if(xt.getNextToken() != tokenCloseTag) return false;
  
  if(xt.getNextToken() != tokenUnknownToken) return false;
  
  xt.getNextToken();
  
  if(xt.getNextToken() != tokenProcessingInstruction) return false;
  
  xt.getNextToken();
  
  xt.getNextToken();
  
  if(xt.getNextToken() != tokenTagClose) return false;
  
  if(!s->equals("link")) return false;
  
  if(xt.isEndOfFile() != true) return false;

  return true;

}

/** 
 * Gets a new token from the source.
 * @return the XMLTokenizer::xmlToken
 */
XMLTokenizer::xmlToken XMLTokenizer::getNextToken() {

  lastReadToken = readNewToken();
  return lastReadToken;

}

/** 
 * Returns the last read token.
 * 
 * @return the last read XMLTokenizer::xmlToken
 */
XMLTokenizer::xmlToken XMLTokenizer::getLastReadToken() {

  return lastReadToken;

}

/**
 * Returns a pointer to a String object containing the string that was
 * in the last token. There are 4 special possibilities. 
 *
 * 1) When the XMLTokenizer reads ""hello"" getNextToken returns
 * XMLTokenizer::StringConstant. When getString is called before
 * another call to getNextToken it returns "hello" (without the
 * quotation marks).
 *
 * 2) When the XMLTokenizer reads "width" getNextToken returns
 * XMLTokenizer::Name. When getName is called before another call to
 * getNextToken it returns "width" (without the quotation marks).
 *
 * 3) When the XMLTokenizer reads "<image" getNextToken returns
 * XMLTokenizer::TagOpen. When getStringToken is called before another
 * call to getNextToken it returns "image" (without the quotation
 * marks).
 *
 * 4) When the XMLTokenizer reads "</image" getNextToken returns
 * XMLTokenizer::TagClose. When getStringToken is called before
 * another call to getNextToken it returns "image" (without the
 * quotation marks). 
 * In all other cases getTokenAsString returns the complete token.
 * The pointer to the String object is always the same, so for a
 * XMLTokenizer object you have to call this function only once.
 *
 * @return a pointer to a String object containing the string, DO NOT DELETE */
String *XMLTokenizer::getTokenAsString() {

  return tokenAsString;

}

/** 
 * Reads charcter data from the stream
 *
 * @param s this string is filled with the character data without
 * newlines, double spaces, ...
 * @return true if an error occured
 */
bool XMLTokenizer::readCharacterData(String *s) {

  characterStream->pushBack();
  char c = characterStream->getNextChar();
  bool space = false;

  if(c == ' ' || c == '\n' || c == '\r' || c == '\t') 
    space = true;

  c = characterStream->getNextChar();

  while(c == ' ' || c == '\n' || c == '\r' || c == '\t')
    c = characterStream->getNextChar();

  /* why not:

     for(c = characterStream->getNextChar(); isspace(c); space = true);

     -ph

     you mean(?):

     for(c = characterStream->getNextChar(); isspace(c); space = true)
       c = characterStream->getNextChar();
     
     that would be the same as:

     for(c = characterStream->getNextChar(); 
        isspace(c); c = characterStream->getNextChar())
       ;

     space = true;

     in your code space=true when leaving the loop
     in mine this might not be the case
     
     -wk
  */

  if(c == '<') 
    return true;

  while(c != '<') {
    
    if(c == ' ' || c == '\n' || c == '\r' || c == '\t') {

        space = true;

    } else {

      if(space) 
        s->append(' ');

      s->append(c);
      space = false;

    }
    c = characterStream->getNextChar();

  }

  if(space) 
    s->append(' ');

  characterStream->pushBack();
  return false;

}

/** 
 * Returns the line number of where the tokenizer started reading the
 * token.
 *
 * @return the line number 
 */
int XMLTokenizer::getTokenStartLineNumber() {

  return tokenStartLineNumber;

}

/** 
 * Determines whether the tokenizer is at the end of the file, after
 * seeing spaces, etc.
 *
 * @return true iff the tokenizer is at the end of the file 
 */
bool XMLTokenizer::isEndOfFile() {
  
  return characterStream->isEndOfFile();

}

/** 
 * Returns the line number of the line the tokenizer is currently on.
 *
 * @return the line number 
 */
int XMLTokenizer::getLineNumber() {

  return characterStream->getLineNumber();

}

/** 
 * Pushes the token back on the stream. This can olny be done once.  
 */
void XMLTokenizer::pushbackToken() {

  characterStream->recallPosition();

}

/** 
 * Reads the "...]]>" part of "<![CDATA[...]]>"
 *
 * @param s this string is filled with the character data
 */
void XMLTokenizer::readCDataSection(String *s) {
  char c = ' ';

  while(true) {
    
    while(c != ']') {

      c = characterStream->getNextChar();
      s->append(c);

    }

    /* read ']' */
    c = characterStream->getNextChar();
    s->append(c);
    
    if(c == ']') {
      
      /* read ']]' */
      c = characterStream->getNextChar();
      s->append(c);

      while (c == ']') {

        c = characterStream->getNextChar();
        s->append(c);

      }
      
      if(c == '>') {
	
      	/* read ']]>' */
        /* remove last 3 characters */
        s->removeLastCharacter(); /* > */
        s->removeLastCharacter(); /* ] */
        s->removeLastCharacter(); /* ] */
	      return;
	
      }
      
    }
    
  }

}

/**
 * Reads the comment from after "< !--".
 */
void XMLTokenizer::readComment() {
  char c = ' ';

  while(true) {

    while(c != '-') {

      c = characterStream->getNextChar();
      tokenAsString->append(c);

    }
    /* read '-' */

    c = characterStream->getNextChar();
    tokenAsString->append(c);

    if(c == '-') {
      
      /* read '--' */

      while(c == '-') {

        c = characterStream->getNextChar();
        tokenAsString->append(c);

      }

      if(c == '>')
        return; /* read '-->' */

    }

  }
  return;
}

/**
 * Reads the processing instruction till "?>".
 */
void XMLTokenizer::readProcessingInstruction() {
  char c = ' ';

  while(true) {

    while(c != '?') {
    
      c = characterStream->getNextChar();
      tokenAsString->append(c);

    }
    /* read '?' */

    while(c == '?') {

      c = characterStream->getNextChar();
      tokenAsString->append(c);

    }

    if(c == '>') 
      return; /* read '?>' */
  }
}

/**
 * Determines wheter a character could be in a name
 * @return true if c resembles a name character
 */
bool XMLTokenizer::isNameCharacter(char c) {

  if(c == ':') return true;
  if(c == '-') return true;
  if(c < 'A') return false;
  if(c <= 'Z' ) return true;
  if(c < 'a') return false;
  if(c <= 'z') return true;

  return false;

}

/**
 * Reads a new token from the stream, used by getNewToken 
 * @return the XMLTokenizer::xmlToken
 */
XMLTokenizer::xmlToken XMLTokenizer::readNewToken() {

  /* this character is used to get characters from the character stream */
  char c; 
  
  /* clear the tokenAsString, it could still be filled by another call to 
     readNewToken */
  tokenAsString->clear(); 

  bool firstCharSpace = false;
  
  /* get the first character from the stream */
  c = characterStream->getNextChar(); 
  
  while(c == ' ' || c =='\n' || c == '\r' || c == '\t') {
    
    firstCharSpace = true;
    /* remove leading layout characters */
    c = characterStream->getNextChar();

  }

  characterStream->pushBack();
  characterStream->savePosition();
  c = characterStream->getNextChar();

  tokenAsString->append(c);
  tokenStartLineNumber = getLineNumber();

  if(c == '<') { /* read '<' */

    c = characterStream->getNextChar();
    tokenAsString->append(c);

    if(c == '!') { /*  read '<!' */

      c = characterStream->getNextChar();
      tokenAsString->append(c);

      if(c == '-') { /* read '<!-' */

        c = characterStream->getNextChar();
        tokenAsString->append(c);

        if(c == '-') { /* read '<!--' */

          readComment();
          return tokenComment;

        }
        return tokenUnknownToken; /* '<!-X' is not allowed in XML */

      }
      
      if(c != 'D' && c != '[') 
        return tokenUnknownToken;  

      if(c == '[') {
        /* read '<![' */
        
        return match("CDATA[", tokenCDataOpen);

      }

      /* read '<!D' */

      if(match("OCTYPE", tokenDocTypeOpen) == tokenDocTypeOpen) {

        c = characterStream->getNextChar();
        tokenAsString->append(c);
        if(c != ' ' && c !='\n' && c != '\r' && c != '\t')
          return tokenUnknownToken;

        characterStream->pushBack();

        return tokenDocTypeOpen;  /* read '<!DOCTYPE ' */

      } else 
	      return tokenUnknownToken;

    }

    if(c == '?') { /* read '<?' */

      c = characterStream->getNextChar();
      tokenAsString->append(c);
      if(c != 'x') {  

        readProcessingInstruction();
        return tokenProcessingInstruction;
      
      }                                 /* read '<?x' */
      c = characterStream->getNextChar(); 
      tokenAsString->append(c);
      if(c != 'm') {

        readProcessingInstruction();
        return tokenProcessingInstruction;
      
      }                                 /* read '<?xm' */
      c = characterStream->getNextChar();
      tokenAsString->append(c);
      if(c != 'l') {

        readProcessingInstruction();
        return tokenProcessingInstruction;
      
      }                                 /* read '<?xml' */
      c = characterStream->getNextChar();
      tokenAsString->append(c);
      if(c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '>' || c == '?') {

        characterStream->pushBack();
        return tokenXMLOpen;                 /* read '<?xml ' */ 

      }

      readProcessingInstruction();
      return tokenProcessingInstruction;
    }

    if(c == '/') { /* read '</' */

      c = characterStream->getNextChar();
      if(c == ' ' || c == '\n' || c == '\r' || c == '\t' || c == '>')
        return tokenUnknownToken; /* no space allowed here */

      tokenAsString->clear();
      tokenAsString->append(c); /* append character to the tokenAsString */
      c = characterStream->getNextChar();
      while(c != ' ' && c != '\n' && c != '\r' && c != '\t' && c != '>') {

        tokenAsString->append(c); /* append characters to the tokenAsString */
        c = characterStream->getNextChar();

      }
      while(c == ' ' || c == '\n' || c == '\r' || c == '\t')
        c = characterStream->getNextChar();

      if(c != '>') {

        tokenAsString->append(' ');
        tokenAsString->append(c);
        return tokenUnknownToken;

      }

      return tokenTagClose;

    }

    tokenAsString->clear();
    tokenAsString->append(c); /* read '<X', with X=c */
    if(!isNameCharacter(c)) {

      tokenAsString->clear();
      tokenAsString->append('<');
      tokenAsString->append(c);
      return tokenUnknownToken;

    }
    c = characterStream->getNextChar();

    while(isNameCharacter(c)) {

      tokenAsString->append(c); /* append characters to the tokenAsString */
      c = characterStream->getNextChar();

    }
    
    characterStream->pushBack(); /* the last char is not part of the TagOpen */
    return tokenTagOpen;

  }
  
  if(c == '"' || c == '\'') {

    char beginCharacter = c; /* this could be ' or " */
    c = characterStream->getNextChar();
    tokenAsString->clear();

    while(c != beginCharacter) {

      tokenAsString->append(c); /* append characters to the tokenAsString */

      try {

      	c = characterStream->getNextChar();

      } catch(XMLEndOfFileException) {

        return tokenEndlessString;

      }

    }
    return tokenStringConstant;
  }

  if(c == '=') return tokenEquals;

  if(c == '?') { /* read '?' */

    c = characterStream->getNextChar();
    
    if(c == '>') {

      tokenAsString->append(c);/* read '?>' */
      return tokenXMLClose; 

    }

    characterStream->pushBack();
    return tokenUnknownToken;
  
  }

  if(c == '>')
    return tokenCloseTag; /* read '>' */
  
  if(c == '/') {

    c = characterStream->getNextChar();
    tokenAsString->append(c);
    if(c == '>')
      return tokenCloseEmptyTag;
    return tokenUnknownToken; /* read '>' */

  }

  if(!isNameCharacter(c)) return tokenUnknownToken;

  c = characterStream->getNextChar();
  while(isNameCharacter(c)) {

    tokenAsString->append(c); /* append characters to the tokenAsString */
    c = characterStream->getNextChar();

  }

  characterStream->pushBack();

  if(firstCharSpace == false) return tokenUnknownToken;

  return tokenName;

}

/**
 * Matches the current position of the character stream with the 
 * string in the arguments. When it matches returns token, otherwuse
 * UnknownToken.
 * @param s the string to match the character stream
 * @param token the token to return when s matches the character
 *        stream
 * @return token when s matches the character stream otherwise 
 *         UnknownToken
 */
XMLTokenizer::xmlToken XMLTokenizer::match(char *s, xmlToken token) {

  char c;

  while(s[0] != 0) {

    c = characterStream->getNextChar();
    tokenAsString->append(c);
    if(c != s[0]) return tokenUnknownToken;
    s++;

  }

  return token;

}
