/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptFunctionParameters.cpp,v 1.3 2001/01/27 14:36:41 wouter Exp $
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "WMLScriptValueObject.h"
#include "WMLScriptFunctionParameters.h"

/**
 * Create the parameter object.
 */
WMLScriptFunctionParameters::WMLScriptFunctionParameters() {
  parameters = NULL;
  numberOfParameters = 0;
}

/**
 * Destroy the parameter object.
 */
WMLScriptFunctionParameters::~WMLScriptFunctionParameters() {
  int i;

  if(parameters) {

    for(i = 0; i<size(); i++) 
      if(parameters[i]) 
	delete parameters[i];

    free(parameters);

  }

}

/**
 * Returns the number of parameters in the object.
 * @return nubmer of parameters.
 */
int WMLScriptFunctionParameters::size() {
  return numberOfParameters;
}

/**
 * Returns the value at the specified index. 
 * @return the WMLScriptValueObject at the specified index
 * or NULL if the index is out of range.
 */
WMLScriptValueObject *WMLScriptFunctionParameters::valueAt(int i) {
  if(i >= 0 && i < numberOfParameters) 
    return parameters[i];

  return NULL;
}

/**
 * Adds a string parameter to the list of parameters.
 * @param s pointer to the string to add.
 */
void WMLScriptFunctionParameters::add(char *s) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(s);
  add(v);
}

/**
 * Adds an integer parameter to the list of parameters.
 * @param s pointer to the integer to add.
 */
void WMLScriptFunctionParameters::add(int i) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(i);
  add(v);
}

/**
 * Adds an floating point parameter to the list of parameters.
 * @param s pointer to the floating point value to add.
 */
void WMLScriptFunctionParameters::add(double d) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(d);
  add(v);
}

/**
 * Adds a boolean parameter to the list of parameters.
 * @param s pointer to the boolean value to add.
 */
void WMLScriptFunctionParameters::add(bool b) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(b);
  add(v);
}

/**
 * Adds a WMLScriptValueObject parameter to the list of parameters.
 * @param s pointer to the WMLScriptValueObject to add.
 */
void WMLScriptFunctionParameters::add(WMLScriptValueObject *v) {
  numberOfParameters++;
  size_t size = sizeof(WMLScriptValueObject *) * numberOfParameters;
  parameters = (WMLScriptValueObject **) realloc(parameters, size);
  assert(parameters != NULL);    
  parameters[numberOfParameters - 1] = v;
}

/**
 * Inserts a boolean parameter into the list of parameters.
 * @param b the boolean value.
 * @param before insert the parameter before this element
 * in the parameter list.
 */
void WMLScriptFunctionParameters::insert(bool b, int before) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(b);
  insert(v, before);
}

/**
 * Inserts a string parameter into the list of parameters.
 * @param s a pointer to the string value.
 * @param before insert the parameter before this element
 * in the parameter list.
 */
void WMLScriptFunctionParameters::insert(char *s, int before) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(s);
  insert(v, before);
}

/**
 * Inserts an integer parameter into the list of parameters.
 * @param i the integer value.
 * @param before insert the parameter before this element
 * in the parameter list.
 */
void WMLScriptFunctionParameters::insert(int i, int before) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(i);
  insert(v, before);
}

/**
 * Inserts a floating point parameter into the list of parameters.
 * @param d the floating point value.
 * @param before insert the parameter before this element
 * in the parameter list.
 */
void WMLScriptFunctionParameters::insert(double d, int before) {
  WMLScriptValueObject *v = new WMLScriptValueObject();
  v->setValue(d);
  insert(v, before);
}

/**
 * Inserts a WMLScriptValueObject into the list of parameters.
 * @param v pointer to the WMLScriptValueObject.
 * @param before insert the parameter before this element
 * in the parameter list.
 */
void WMLScriptFunctionParameters::insert(WMLScriptValueObject *v, int before) {
  assert(before >= 0 && before <= numberOfParameters);
  numberOfParameters++;
  size_t size = sizeof(WMLScriptValueObject *) * numberOfParameters;
  parameters = (WMLScriptValueObject **) realloc(parameters, size);
  assert(parameters != NULL);    
  memmove(&parameters[before + 1], &parameters[before], sizeof(WMLScriptValueObject *) * (numberOfParameters - 1 - before));
  parameters[before] = v;
}

