/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptInterpreter.cpp,v 1.14 2001/01/27 14:36:41 wouter Exp $
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <assert.h>

#include "../lib/snprintf.h"

#include "WMLScript.h"

extern int errno;

/**
 * Constructs the WMLScriptInterpreter for a given
 * WMLScriptParseContext.
 * @param a WMLScriptParseContext as returned by a WMLScriptParser.
 */
WMLScriptInterpreter::WMLScriptInterpreter(WMLScriptParseContext *pc) {
  parseContext = pc;
  libraryList = NULL;
  numberOfLibraries = 0;
}

/**
 * Destroys the WMLScriptInterpreter.
 */
WMLScriptInterpreter::~WMLScriptInterpreter() {
  if(libraryList)
    free(libraryList);
}

/**
 * Determines whether a library is available to the interpreter.
 * @param name a pointer to the name of the library.
 * @return true if the library is available.
 */
bool WMLScriptInterpreter::isAvailableLibrary(char *name) const {
  int i;
  for(i = 0; i < numberOfLibraries; i++) 
    if(strcmp(libraryList[i]->getName(), name) == 0)
      return true;

  return false;
}

/**
 * Returns a pointer to the requested WMLScriptLibrary.
 * @param name the name of the library.
 * @return a pointer to the WMLScriptLibrary or NULL if it
 * is not available.
 */
WMLScriptLibrary *WMLScriptInterpreter::getLibrary(char *name) const {
  int i;
  for(i = 0; i < numberOfLibraries; i++) 
    if(strcmp(libraryList[i]->getName(), name) == 0)
      return libraryList[i];

  return NULL;
}

/**
 * Links a library to the interpreter. Linking a library makes
 * the library available to any WMLScript code that runs
 * inside the interpreter.
 * @param lib the library to link.
 * @return positive integer if the library was successfully linked,
 * or -EOUTOFMEMORY on memory allocation failure. A return value
 * > 0 signifies the new number of linked libraries.
 */
int WMLScriptInterpreter::linkLibrary(WMLScriptLibrary *lib) {
  libraryList = (WMLScriptLibrary **) realloc(libraryList, sizeof(WMLScriptLibrary *) * (numberOfLibraries + 1));

  if(libraryList == NULL)
    return -EOUTOFMEMORY;

  libraryList[numberOfLibraries] = lib;
  numberOfLibraries++;

  return numberOfLibraries;
}

/**
 * Unlinks a library from the interpreter. Unlinking a library
 * makes the library unavailable to the interpreter.
 * @param lib the library to unlink.
 * @return positive integer or 0 if the library was successfully unlinked, 
 * -ENOTFOUND if the specified library was not previously linked or 
 * -EOUTOFMEMORY if a memory allocation error occurs. A return value
 * >= 0 signifies the new number of linked libraries.
 */
int WMLScriptInterpreter::unlinkLibrary(WMLScriptLibrary *lib) {
  int i = 0;

  while(i < numberOfLibraries && libraryList[i] != lib) { i++; }

  if(i >= numberOfLibraries) {
    return -ENOTFOUND;
    /* NOT REACHED */
  }

  if(i + 1 < numberOfLibraries) 
    memmove(&libraryList[i], &libraryList[i + 1], sizeof(WMLScriptLibrary *) * (numberOfLibraries - (i + 1)));
  
  int size = sizeof(WMLScriptLibrary *) * (numberOfLibraries - 1);

  if(size == 0) {

    free(libraryList);
    libraryList = NULL;

  } else {

    libraryList = (WMLScriptLibrary **) realloc(libraryList, size);

  }

  numberOfLibraries--;

  if(libraryList == NULL && numberOfLibraries > 0) {
    numberOfLibraries = 0;
    return -EOUTOFMEMORY;
    /* NOT REACHED */
  }

  return numberOfLibraries;
}

/**
 * This method allows programs to call WMLScript functions
 * that have been declared with the "extern" keyword.
 * 
 * @throws WMLScriptWrongArityException
 * when the number of parameters does not match the function
 * declaration.
 * @throws WMLScriptAccessException
 * on attempt to call non-extern function.
 * @throws WMLScriptUnknownIdentifierException 
 * when this function is unknown.
 * @throws WMLScriptException 
 * when some other error occurs.
 * @param name pointer to the name of the function.
 * @param parameters the parameters to the function.  
 * @return the function's return value.
 */
WMLScriptValueObject *WMLScriptInterpreter::call(char *name, WMLScriptFunctionParameters *parameters) {
  WMLScriptValueObject *returnValue;
  WMLScriptVarPool *localVarPool = NULL, *oldVarPool = NULL;

  /* get the parse tree for this function */
  WMLScriptParseTree *parseTree = 
    parseContext->getFunctionPool()->getParseTree(name);

  if(parseTree) {

    /* verify that this function has been marked 'extern' */
    if(!(parseContext->getFunctionPool()->getFlags(name) & FUNC_EXTERN)) 
      throw WMLScriptAccessException();

    /* create a copy of the functions' local variable pool. */
    localVarPool = new WMLScriptVarPool(parseContext->getFunctionPool()->getVarPool(name));

    /* verify that the number of parameters equals the number
       of variables in the local variable bool. */
    if(localVarPool->size() != parameters->size()) {
      delete localVarPool;
      throw WMLScriptWrongArityException(name);
      /* NOT REACHED */
    }

    /* copy parameters into functions' variable pool. */
    int i;
    for(i = localVarPool->size() - 1; i >= 0; i--) 
      localVarPool->setVarValue(localVarPool->varAt(i), new WMLScriptValueObject(parameters->valueAt(parameters->size() - i - 1)));

    /* switch to functions' variable pool. */
    oldVarPool = varPool;
    varPool = localVarPool;

    try {

      /* execute and delete return value, because WMLScript explicitly
         defines an empty string return value if none is supplied. */
      delete ex(parseTree->getTopNode());

      returnValue = new WMLScriptValueObject("");

    } catch(WMLScriptReturnException& e) {

      /* retrieve return value from the exception. */
      returnValue = e.getValue();

    } catch(WMLScriptException&) {

      varPool = oldVarPool;
      delete localVarPool;
      throw;
      /* NOT REACHED */

    }

    varPool = oldVarPool;
    delete localVarPool;

    return returnValue;
    /* NOT REACHED */

  } else
    /* ... could not find parse tree for the function */
    throw WMLScriptUndeclaredIdentifierException(name);
  
}

/**
 * This is the main interpreter loop.
 * @param n parse tree node.
 * @throws WMLScriptException
 * may throw almost any exception, either itself or through a
 * function call or through a library call.
 * @return return value.
 */
WMLScriptValueObject *WMLScriptInterpreter::ex(nodeType *n) {
  int err;
  char *s;
  char str_buf[200];
  WMLScriptValueObject *returnValue = NULL;
  WMLScriptValueObject *newVal = NULL;

  if(!n) {
    throw WMLScriptCorruptParseTreeException("ex: NULL node");
    /* NOT REACHED */
  }

  switch(n->type) {

  case typeCon:
    returnValue = new WMLScriptValueObject();
    switch(n->con.value.type) {
    case dataBool:
      returnValue->setValue((bool) (n->con.value.value.boolValue == 0 ? 
				    false : true));
      break;
    case dataInt:
      returnValue->setValue((int) (n->con.value.value.intValue));
      break;
    case dataFloat:
      returnValue->setValue((double) (n->con.value.value.floatValue));
      break;
    case dataString:
      returnValue->setValue((char *) (&n->con.value.value.stringValue));
      break;
    case dataInvalid:
      break;
    }

    s = returnValue->toString();
    free(s);
    
    return returnValue;

  case typeId:
    returnValue = (WMLScriptValueObject *) varPool->getVarValue(n->id.name);
    if(!returnValue) {

      throw WMLScriptUndeclaredIdentifierException(n->id.name,
						   n->id.location.first_line,
						   n->id.location.first_column);
      /* NOT REACHED */

    } else {

      returnValue = new WMLScriptValueObject(returnValue);

    }

    return returnValue;
    /* NOT REACHED */

  case typeOpr:
    switch(n->opr.oper) {

    case STATEMENT:
      return(ex(n->opr.op[0]));
      /* NOT REACHED */

    case STATEMENTLIST:
      if(n->opr.nops) {

	if(n->opr.nops == 2) {

	  delete ex(n->opr.op[0]);
	  return(ex(n->opr.op[1]));
	  /* NOT REACHED */

	} else if(n->opr.nops == 1) {
	  
	  return(ex(n->opr.op[0]));
	  /* NOT REACHED */

	} else {

	  throw WMLScriptCorruptParseTreeException("ex: too many operands in STATEMENTLIST");
	  /* NOT REACHED */

	}

      } else {

	/* empty block */
	returnValue = new WMLScriptValueObject();
	returnValue->setValue("");
	return returnValue;
	/* NOT REACHED */

      }
    case VAR:
      return ex(n->opr.op[0]);

    case VARDECL:
      err = varPool->createVar(n->opr.op[0]->id.name);

      if(err == EOUTOFMEMORY) {

	throw WMLScriptOutOfMemoryException();
	/* NOT REACHED */

      }

      assert(!err);

      /* declaration with initializer */
      if(n->opr.nops == 2) {
	
	if((err = varPool->setVarValue(n->opr.op[0]->id.name, ex(n->opr.op[1])))) {
	  assert(0);
	  /* NOT REACHED */
	  
	}
	
      } else {
	
	/* no initializer, initialize to empty string */
	returnValue = new WMLScriptValueObject();
	returnValue->setValue("");

	if((err = varPool->setVarValue(n->opr.op[0]->id.name, returnValue))) {

	  assert(0);
	  /* NOT REACHED */

	}
	
      }
      
      returnValue = 
	(WMLScriptValueObject *) varPool->getVarValue(n->opr.op[0]->id.name);
      
      if(!returnValue) {
	
	assert(0);
	/* NOT REACHED */
	
      } else {

	returnValue = new WMLScriptValueObject(returnValue);
	
      }
      
      return returnValue;
      /* NOT REACHED */

    case VARDECLLIST:
      delete ex(n->opr.op[0]);
      return(ex(n->opr.op[1]));
      /* NOT REACHED */

    case EXPRESSION:
      if(n->opr.nops == 1) {

	return ex(n->opr.op[0]);
	/* NOT REACHED */

      } else if(n->opr.nops == 2) {

	delete ex(n->opr.op[0]);
	return ex(n->opr.op[1]);
	/* NOT REACHED */

      } else {

	throw WMLScriptCorruptParseTreeException("ex: too many operands in EXPRESSION");
	/* NOT REACHED */

      }

    case EXPRESSIONSTATEMENT:
      return ex(n->opr.op[0]);
      /* NOT REACHED */

    case ASSIGNMENTEXPRESSION:
      newVal = ex(n->opr.op[1]);

      if((err = varPool->setVarValue(n->opr.op[0]->id.name, newVal))) {

	delete newVal;

	if(err == EVARNOTFOUND) {

	  throw WMLScriptUndeclaredIdentifierException(n->id.name,
						       n->id.location.first_line,
						       n->id.location.first_column);
	/* NOT REACHED */
	  
	} else {
	
	  throw WMLScriptRuntimeException(n->id.name,
					n->id.location.first_line,
					  n->id.location.first_column);
	  
	  /* NOT REACHED */
	}

    }

      returnValue = ex(n->opr.op[0]);
      s = returnValue->toString();
      free(s);
      return returnValue;
      /* NOT REACHED */

    case POSTFIXEXPRESSION:
      returnValue = ex(n->opr.op[0]);
      newVal = ex(n->opr.op[1]);

      if((err = varPool->setVarValue(n->opr.op[0]->id.name, newVal))) {

	/* it would be very improbable to get here. errors regarding
	   undeclared variables are most likely caught by the
	   newVal line above. */
	delete newVal;
	
	if(err == EVARNOTFOUND) {

	  throw WMLScriptUndeclaredIdentifierException(n->id.name,
						       n->id.location.first_line,
						       n->id.location.first_column);
	  /* NOT REACHED */

	} else {

	  throw WMLScriptRuntimeException(n->id.name,
					  n->id.location.first_line,
					  n->id.location.first_column);
	  /* NOT REACHED */
	  
	}

      }
      return returnValue;
      /* NOT REACHED */

    case '-':
      return operatorSubtraction(n);
      
    case '+':
      return operatorAddition(n);
   
    case '*':
      return operatorMultiplication(n);

    case '/':
      return operatorDivision(n);

    case DIV:
      return operatorIntegerDivision(n);

    case '%':
      return operatorRemainder(n);

    case EQ:
      return operatorEquals(n);

    case NE:
      return operatorNotEquals(n);

    case AND:
      return operatorLogicalAnd(n);

    case OR:
      return operatorLogicalOr(n);

    case '&':
      return operatorBitwiseAnd(n);

    case '^':
      return operatorBitwiseXor(n);

    case '|':
      return operatorBitwiseOr(n);

    case IF:
      return operatorIf(n);

    case WHILE:
      return operatorWhile(n);

    case '<':
      return operatorLess(n);

    case '>':
      return operatorGreater(n);

    case LE:
      return operatorLessEqual(n);

    case GE:
      return operatorGreaterEqual(n);

    case ISVALID:
      return operatorIsValid(n);

    case TYPEOF:
      return operatorTypeOf(n);

    case UPLUS:
      return operatorUplus(n);

    case UMINUS:
      return operatorUminus(n);

    case UINVERT:
      return operatorUinvert(n);

    case UNOT:
      return operatorUnot(n);

    case SHL:
      return operatorShiftLeft(n);

    case SHR:
      return operatorShiftRight(n);

    case SHRR:
      return operatorShiftRightZeroFill(n);

    case FOR:
      return operatorFor(n);

    case BREAK:
      operatorBreak(n);

    case CONTINUE:
      operatorContinue(n);

    case RETURN:
      operatorReturn(n);

    case LOCALSCRIPTFUNCTIONCALL:
      return operatorLocalScriptFunctionCall(n);

    case LIBRARYSCRIPTFUNCTIONCALL:
      return operatorLibraryScriptFunctionCall(n);

    default:
      snprintf(str_buf, 200, "ex: not implemented: n->opr.oper = %d\n", n->opr.oper);
      throw WMLScriptRuntimeException(str_buf);

    }

  }

  throw WMLScriptRuntimeException("ex: return without a value.");

}

/**
 * Performs a library function call.
 * @param n the parse tree node describing the call.
 * @return the result of the function call.
 */
WMLScriptValueObject *WMLScriptInterpreter::operatorLibraryScriptFunctionCall(nodeType *n) {
  assert(n->opr.nops == 3);
  WMLScriptValueObject *returnValue;

  /* verify that library is available. */
  if(isAvailableLibrary(n->opr.op[0]->id.name)) {

    WMLScriptLibrary *library = getLibrary(n->opr.op[0]->id.name);
    
    int id = library->getFunctionID(n->opr.op[1]->id.name);

    /* verify that the requested library function is available. */
    if(id != -1) {
      
      /* set up the library function arguments and
	 call the library function. */
      WMLScriptFunctionParameters parameters;
      pushArguments2(n->opr.op[2], &parameters);
      
      try {

	returnValue = library->call(id, &parameters);

      } catch(WMLScriptExitedException& e) {

	/* special case: Lang.exit throws exception. */
	throw WMLScriptExitedException(e.getValue(),
				       n->opr.location.first_line,
				       n->opr.location.first_column);
	/* NOT REACHED */

      } catch(WMLScriptAbortedException& e) {

	/* special case: Lang.abort throws exception. */
	throw WMLScriptAbortedException(e.getValue(),
					n->opr.location.first_line,
					n->opr.location.first_column);
	/* NOT REACHED */

      } catch(WMLScriptWrongArityException&) {

	/* wrong number of arguments to library function. */
	char *s = (char *) malloc(strlen(n->opr.op[0]->id.name) + 
				  strlen(n->opr.op[1]->id.name) + 2);
	sprintf(s, "%s.%s", n->opr.op[0]->id.name, n->opr.op[1]->id.name);
	WMLScriptWrongArityException e2(s,
					n->opr.location.first_line,
					n->opr.location.first_column);
	free(s);
	throw e2;

      } catch(WMLScriptException&) {
	
	throw;
	/* NOT REACHED */

      }
      
    } else
      throw WMLScriptUnknownLibraryFunctionException(n->opr.op[1]->id.name,
					   n->opr.location.first_line,
					   n->opr.location.first_column);

  } else
    throw WMLScriptUnknownLibraryException(n->opr.op[0]->id.name,
					   n->opr.location.first_line,
					   n->opr.location.first_column);

  return returnValue;

}

/**
 * Calls a local function.
 * @param parse tree node describing the function call.
 * @return result of the function call.
 */
WMLScriptValueObject *WMLScriptInterpreter::operatorLocalScriptFunctionCall(nodeType *n) {
  assert(n->opr.nops == 2);

  WMLScriptValueObject *returnValue;
  WMLScriptVarPool *localVarPool = NULL, *oldVarPool = varPool;
  
  /* get the parse tree for the function to call. */
  WMLScriptParseTree *parseTree = 
    parseContext->getFunctionPool()->getParseTree(n->opr.op[0]->id.name);

  if(parseTree) {

    try {

      /* create a copy of the functions' variable pool. */
      localVarPool = new WMLScriptVarPool(parseContext->getFunctionPool()->getVarPool(n->opr.op[0]->id.name));
      
      /* store the parameters into the functions' variable pool. */
      pushArguments(localVarPool, n->opr.op[1]);
      
      /* save the variable pool and set it to the functions' variable 
	 pool. */
      varPool = localVarPool;

      /* execute. */
      delete ex(parseTree->getTopNode());
      returnValue = new WMLScriptValueObject("");

    } catch(WMLScriptReturnException& e) {

      /* get return value. */
      returnValue = e.getValue();

    } catch(WMLScriptWrongArityException&) {

      /* wrong number of parameters. */
      char *s = (char *) malloc(strlen(n->opr.op[0]->id.name) + 1);
      strcpy(s, n->opr.op[0]->id.name);
      WMLScriptWrongArityException e2(s, 
				      getLineNumber(n),
				      getColumnNumber(n));
      free(s);

      varPool = oldVarPool;
      delete localVarPool;
      throw e2;
      /* NOT REACHED */

    } catch(WMLScriptException&) {

      varPool = oldVarPool;
      delete localVarPool;
      throw;
      /* NOT REACHED */

    }

    varPool = oldVarPool;
    delete localVarPool;

  } else {

    /* function parse tree not found. */
    throw WMLScriptUndeclaredIdentifierException(n->opr.op[0]->id.name, 
						 n->opr.location.first_line,
						 n->opr.location.first_column);

  }

  return returnValue;

}

int WMLScriptInterpreter::pushArgument(WMLScriptVarPool *vp,
				       nodeType *n, 
				       int i) {
  int r;

  WMLScriptValueObject *varValue = ex(n);

  if((r = vp->setVarValue(vp->varAt(i++), varValue))) {

    if(r == ENULLVAR) 
      /* signals attempt to assign to non-existant variable (wrong arity);
	 it is very improbable to get here; the calling function
	 pushArguments() already performs an arity check. */
      throw WMLScriptWrongArityException(vp->varAt(i++),
					 getLineNumber(n),
					 getColumnNumber(n));

    else 
      throw WMLScriptException("Internal error");

  }

  return i;

}

void WMLScriptInterpreter::pushArguments(WMLScriptVarPool *vp, nodeType *n) {
  int i = 0;

  if(n->opr.oper == ARGUMENTS) {
    
    if(n->opr.nops == 0) 
      
      return;
    
    else if(n->opr.nops == 1) {
      
      n = n->opr.op[0];
      
      if(n->opr.oper == ARGUMENTLIST) {
	
	while(n->opr.nops == 2) {
	  
	  i = pushArgument(vp, n->opr.op[1], i);
	  n = n->opr.op[0];
	  
	}
	
	if(n->opr.nops == 1)
	  
	  i = pushArgument(vp, n->opr.op[0], i);

	else {

	  throw WMLScriptCorruptParseTreeException("pushArguments: malformed node; expected single-operand ARGUMENTLIST");

	}

	if(i != vp->size()) {

	  //dprintf("pushArguments: too few arguments, expected %d, got %d\n", vp->size(), i);
	  throw WMLScriptWrongArityException();

	}

      }
      
    }
    
  } else {

    //dprintf("pushArguments: malformed node; expected ARGUMENTS\n");
    throw WMLScriptCorruptParseTreeException("pushArguments: malformed node");

  }

}

void WMLScriptInterpreter::pushArguments2(nodeType *n, WMLScriptFunctionParameters *parameters) {

  if(n->opr.oper == ARGUMENTS) {
    
    if(n->opr.nops == 0) 
      
      return;
    
    else if(n->opr.nops == 1) {
      
      n = n->opr.op[0];
      
      if(n->opr.oper == ARGUMENTLIST) {
	
	while(n->opr.nops == 2) {

	  parameters->insert(ex(n->opr.op[1]), 0);
	  n = n->opr.op[0];
	  
	}
	
	if(n->opr.nops == 1)
	  
	  parameters->insert(ex(n->opr.op[0]), 0);

	else {

	  throw WMLScriptCorruptParseTreeException("pushArguments2: malformed node; expected single-operand ARGUMENTLIST");

	}

      }
      
    }
    
  } else {

    printf("pushArguments2: malformed node; expected ARGUMENTS\n");
    throw WMLScriptException();

  }

}

void WMLScriptInterpreter::operatorBreak(nodeType *n) {
  throw WMLScriptBreakException(n->opr.location.first_line,
				n->opr.location.first_column);
}

void WMLScriptInterpreter::operatorContinue(nodeType *n) {
  throw WMLScriptContinueException(n->opr.location.first_line,
				   n->opr.location.first_column);
}

void WMLScriptInterpreter::operatorReturn(nodeType *n) {
  WMLScriptValueObject *returnValue;

  if(n->opr.nops == 1) {

    returnValue = ex(n->opr.op[0]);

  } else {
    
    returnValue = new WMLScriptValueObject();
    returnValue->setValue("");

  }

  throw WMLScriptReturnException(returnValue, 
				 n->opr.location.first_line,
				 n->opr.location.first_column);

}

WMLScriptValueObject *WMLScriptInterpreter::operatorFor(nodeType *n) {
  WMLScriptValueObject *condition = NULL, *tempVal = NULL;
  int alwaysTrue = 0;
  
  if(n->opr.nops == 4) {
  
    if(n->opr.op[0]) {

      /* initializer */
      delete ex(n->opr.op[0]);
      
    }
    
    /* condition */
    if(n->opr.op[1]) {
      
      condition = ex(n->opr.op[1]);
      condition->convertTo(dataBool);
      
    } else {
      
      alwaysTrue = 1;

    }

    while(alwaysTrue ||
	  (condition->getType() == dataBool && 
	   condition->getBoolValue() == true)) {
      
      tempVal = NULL;

      /* statement */
      try {
	
	      tempVal = ex(n->opr.op[3]);

      } catch(WMLScriptBreakException&) {
	
        break;

      } catch(WMLScriptContinueException&) {

      }

      delete tempVal;

      /* update expression */
      if(n->opr.op[2]) {

	delete ex(n->opr.op[2]);

      }

      if(!alwaysTrue) {

	delete condition;
	condition = NULL;
	condition = ex(n->opr.op[1]);
	condition->convertTo(dataBool);
	
      }
      
    }

    delete condition;
    
    /* returning NULL is not right -- have to think about it */
    return NULL;

  } else 
    
    throw WMLScriptCorruptParseTreeException("wrong number of arguments in operatorFor");

}

WMLScriptValueObject *WMLScriptInterpreter::operatorWhile(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = NULL;

  val1->convertTo(dataBool);

  while(val1->getType() == dataBool && val1->getBoolValue()) {

    try {
      
      delete val2;
      val2 = NULL;
      val2 = ex(n->opr.op[1]);

    } catch(WMLScriptBreakException&) {

      break;

    } catch(WMLScriptContinueException&) {

    }

    delete val1;
    val1 = NULL;
    val1 = ex(n->opr.op[0]);
    val1->convertTo(dataBool);
    
  }

  delete val2;

  return val1;

}

WMLScriptValueObject *WMLScriptInterpreter::operatorIf(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = NULL;

  val1->convertTo(dataBool);

  if(val1->getType() == dataBool && val1->getBoolValue()) {

    delete val1;
    val1 = NULL;
    returnValue = ex(n->opr.op[1]);
    
  } else if(n->opr.nops == 3) {

    delete val1;
    val1 = NULL;
    returnValue = ex(n->opr.op[2]);

  }

  delete val1;

  return returnValue;

}

WMLScriptValueObject *WMLScriptInterpreter::operatorLogicalOr(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = NULL;
  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  val1->convertTo(dataBool);

  if(val1->getType() == dataBool && val1->getBoolValue()) {

    returnValue->setValue(true);

  } else if(val1->getType() != dataInvalid) {
    /* first operand was false and not invalid -> 
       now evaluate second operand */
    val2 = ex(n->opr.op[1]);
    val2->convertTo(dataBool);

    if(val2->getType() == dataBool && val2->getBoolValue()) {

      returnValue->setValue(true);
      
    } else if(val2->getType() == dataBool && val2->getBoolValue() == false) {

      returnValue->setValue(false);

    }

  }

  delete val1;
  if(val2) 
    delete val2;

  return returnValue;

}

WMLScriptValueObject *WMLScriptInterpreter::operatorLogicalAnd(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = NULL;
  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  val1->convertTo(dataBool);
  
  if(val1->getType() == dataBool && val1->getBoolValue() == false) {

    returnValue->setValue(false);

  } else if(val1->getType() != dataInvalid) {
    /* first operand was true and not invalid -> 
       now evaluate second operand */
    val2 = ex(n->opr.op[1]);
    val2->convertTo(dataBool);
    
    if(val2->getType() == dataBool && val2->getBoolValue()) {
      
      returnValue->setValue(true);
    
    } else if(val2->getType() == dataBool && val2->getBoolValue() == false) {

      returnValue->setValue(false);

    }
    
  }

  delete val1;
  delete val2;
  
  return returnValue;
  
}

WMLScriptValueObject *WMLScriptInterpreter::operatorShiftLeft(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 << *val2);
  delete val1;
  delete val2;
  return returnValue;
}
 
WMLScriptValueObject *WMLScriptInterpreter::operatorShiftRight(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 >> *val2);
  delete val1;
  delete val2;
  return returnValue;
}

/* needs proper semantics!! */
WMLScriptValueObject *WMLScriptInterpreter::operatorShiftRightZeroFill(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
    
  if(val1->getType() == dataInt && val2->getType() == dataInt) 

    returnValue->setValue((int) (val1->getIntValue() >> 
				 val2->getIntValue()));

  char *s = returnValue->toString();
  free(s);

  delete val1;
  delete val2;

  return returnValue;
  
}

WMLScriptValueObject *WMLScriptInterpreter::operatorUminus(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(-*val);
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorUplus(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(+*val);
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorUinvert(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(~*val);
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorUnot(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(!*val);
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorTypeOf(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(typeOf(val));
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorIsValid(nodeType *n) {
  WMLScriptValueObject *val = ex(n->opr.op[0]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(isValid(val));
  delete val;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorLess(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 < *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorLessEqual(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 <= *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorGreater(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 > *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorGreaterEqual(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 >= *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorBitwiseAnd(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 & *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorBitwiseXor(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 ^ *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorBitwiseOr(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 | *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorRemainder(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 % *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorMultiplication(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 * *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorIntegerDivision(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(div(*val1, *val2));
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorDivision(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 / *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorNotEquals(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 != *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorEquals(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 == *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorSubtraction(nodeType *n) {
  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 - *val2);
  delete val1;
  delete val2;
  return returnValue;
}

WMLScriptValueObject *WMLScriptInterpreter::operatorAddition(nodeType *n) {

  WMLScriptValueObject *val1 = ex(n->opr.op[0]);
  WMLScriptValueObject *val2 = ex(n->opr.op[1]);
  WMLScriptValueObject *returnValue = new WMLScriptValueObject(*val1 + *val2);
  delete val1;
  delete val2;
  return returnValue;
}
