/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptInterpreter.h,v 1.6 2001/01/27 14:36:41 wouter Exp $
 */

#ifndef __WMLSCRIPTINTERPRETER_H__
#define __WMLSCRIPTINTERPRETER_H__

/**
 * The WMLScript interrpeter. The interrpreter can execute 
 * parsed WMLScript code. It's functionality can be extended
 * by using the linkLibrary() function. Also, it provides 
 * a means to interface with the host enviroment for debugging
 * facilities and net access.
 *
 * Example:
 *
 * \code
 * WMLScriptValueObject *value;
 *
 * try {
 *
 *   WMLScriptParser parser = WMLScriptParser();
 *   parseContext = parser.parse("extern function answer() { return 42; };");
 *   WMLScriptInterpreter interpreter(parseContext);
 *   WMLScriptFunctionParameters parameters;
 *   interpreter.linkLibrary(libLang);
 *   interpreter.linkLibrary(libFloat);
 *   interpreter.linkLibrary(libString);
 *   value = interpreter.call("test", &parameters);
 *
 * } catch(WMLScriptException& e) {
 *  
 *   delete value;
 *   delete parseContext;
 *  
 * }
 *
 * if(value) {
 *
 *   printf("returned: %s\n", value->toString());
 *
 * }
 * \endcode
 * 
 */
class WMLScriptInterpreter {
 public:
  WMLScriptInterpreter(WMLScriptParseContext *parseContext);
  ~WMLScriptInterpreter();
  
  WMLScriptValueObject *call(char *name, WMLScriptFunctionParameters *parameters);

  bool isAvailableLibrary(char *name) const;
  WMLScriptLibrary *getLibrary(char *name) const;  
  int linkLibrary(WMLScriptLibrary *library);
  int unlinkLibrary(WMLScriptLibrary *library);

 private:
  WMLScriptParseContext *parseContext;
  WMLScriptVarPool *varPool;
  WMLScriptLibrary **libraryList;
  int numberOfLibraries;
  int exitLine;
  int exitColumn;
  int abortLine;
  int abortColumn;

  WMLScriptValueObject *ex(nodeType* n);
  
  void operatorReturn(nodeType *n);
  void operatorBreak(nodeType *n);
  void operatorContinue(nodeType *n);

  WMLScriptValueObject *operatorLibraryScriptFunctionCall(nodeType *n);
  WMLScriptValueObject *operatorLocalScriptFunctionCall(nodeType *n);

  WMLScriptValueObject *operatorRemainder(nodeType *n);

  WMLScriptValueObject *operatorFor(nodeType *n);
  WMLScriptValueObject *operatorTypeOf(nodeType *n);
  WMLScriptValueObject *operatorIsValid(nodeType *n);
  WMLScriptValueObject *operatorLess(nodeType *n);
  WMLScriptValueObject *operatorLessEqual(nodeType *n);
  WMLScriptValueObject *operatorGreater(nodeType *n);
  WMLScriptValueObject *operatorGreaterEqual(nodeType *n);
  WMLScriptValueObject *operatorWhile(nodeType *n);
  WMLScriptValueObject *operatorIf(nodeType *n);
  WMLScriptValueObject *operatorLogicalAnd(nodeType *n);
  WMLScriptValueObject *operatorLogicalOr(nodeType *n);
  WMLScriptValueObject *operatorBitwiseAnd(nodeType *n);
  WMLScriptValueObject *operatorBitwiseXor(nodeType *n);
  WMLScriptValueObject *operatorBitwiseOr(nodeType *n);
  WMLScriptValueObject *operatorNotEquals(nodeType *n);
  WMLScriptValueObject *operatorEquals(nodeType *n);
  WMLScriptValueObject *operatorAddition(nodeType *n);
  WMLScriptValueObject *operatorSubtraction(nodeType *n);
  WMLScriptValueObject *operatorMultiplication(nodeType *n);
  WMLScriptValueObject *operatorDivision(nodeType *n);
  WMLScriptValueObject *operatorIntegerDivision(nodeType *n);
  WMLScriptValueObject *operatorUminus(nodeType *n);
  WMLScriptValueObject *operatorUplus(nodeType *n);
  WMLScriptValueObject *operatorUinvert(nodeType *n);
  WMLScriptValueObject *operatorUnot(nodeType *n);
  WMLScriptValueObject *operatorShiftLeft(nodeType *n);
  WMLScriptValueObject *operatorShiftRight(nodeType *n);
  WMLScriptValueObject *operatorShiftRightZeroFill(nodeType *n);

  void pushArguments(WMLScriptVarPool *vp, nodeType *n);
  int pushArgument(WMLScriptVarPool *vp, nodeType *n, int i);
  void pushArguments2(nodeType *n, WMLScriptFunctionParameters *parameters);

#ifdef WMLS_DEBUG
  void dprintf(char *s, ...);
#endif

};

#endif
