/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptLibraryFloat.cpp,v 1.6 2001/01/27 14:36:41 wouter Exp $
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "WMLScriptValueObject.h"
#include "WMLScriptFunctionParameters.h"
#include "WMLScriptLibraryFloat.h"
#include "WMLScriptExitedException.h"
#include "WMLScriptAbortedException.h"
#include "WMLScriptWrongArityException.h"
#include "WMLScriptUnknownLibraryFunctionException.h"

/**
 * Creates the library.
 */
WMLScriptLibraryFloat::WMLScriptLibraryFloat() {
}

/**
 * Destroys the library.
 */
WMLScriptLibraryFloat::~WMLScriptLibraryFloat() {
}

/**
 * Returns the function ID of a given function,
 * or -1 if the function does not exist.
 * @param name the function name.
 * @return the ID of the given function or -1 if the function
 * does not exist.
 */
int WMLScriptLibraryFloat::getFunctionID(const char *name) const {
  if(strcmp(name, "int") == 0) {
    return FUNC_INT;
  } else if(strcmp(name, "floor") == 0) {
    return FUNC_FLOOR;
  } else if(strcmp(name, "ceil") == 0) {
    return FUNC_CEIL;
  } else if(strcmp(name, "pow") == 0) {
    return FUNC_POW;
  } else if(strcmp(name, "round") == 0) {
    return FUNC_ROUND;
  } else if(strcmp(name, "sqrt") == 0) {
    return FUNC_SQRT;
  } else if(strcmp(name, "maxFloat") == 0) {
    return FUNC_MAXFLOAT;
  } else if(strcmp(name, "minFloat") == 0) {
    return FUNC_MINFLOAT;
  }

  return -1;
}

/**
 * Invokes the function with the given function ID with the
 * specified parameters.
 * @param id the function ID to invoke.
 * @param parameters the parameters to invoke the function with.
 * @throws WMLScriptWrongArityException
 * if the number of parameters does not match the expected
 * number of parameters.
 * @throws WMLScriptUnknownLibraryFunctionException
 * if the given function ID does not refer to a known function.
 * @return the return value of the library function.
 */
WMLScriptValueObject *WMLScriptLibraryFloat::call(int id, WMLScriptFunctionParameters *parameters) {
  switch(id) {
  case FUNC_INT:
    return funcInt(parameters);
  case FUNC_FLOOR:
    return funcFloor(parameters);
  case FUNC_CEIL:
    return funcCeil(parameters);
  case FUNC_POW:
    return funcPow(parameters);
  case FUNC_ROUND:
    return funcRound(parameters);
  case FUNC_SQRT:
    return funcSqrt(parameters);
  case FUNC_MAXFLOAT:
    return funcMaxFloat(parameters);
  case FUNC_MINFLOAT:
    return funcMinFloat(parameters);
  }
  throw WMLScriptUnknownLibraryFunctionException();
}

/**
 * Returns the name of the library.
 * @return pointer at library name.
 */
const char *WMLScriptLibraryFloat::getName() const {
  return "Float";
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcInt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcInt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) 

    returnValue->setValue(p1->getIntValue());
    
  else if(p1->getType() == dataFloat)

    returnValue->setValue((int) p1->getFloatValue());

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcFloor(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcFloor");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) 

    returnValue->setValue(p1->getIntValue());
    
  else if(p1->getType() == dataFloat)

    returnValue->setValue((int) floor(p1->getFloatValue()));

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcCeil(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcCeil");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) 

    returnValue->setValue(p1->getIntValue());
    
  else if(p1->getType() == dataFloat)

    returnValue->setValue((int) ceil(p1->getFloatValue()));

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcPow(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 2, "funcPow");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0),
    *p2 = parameters->valueAt(1);

  p1->convertTo(dataFloat);
  
  if(p1->getType() == dataFloat) {
    
    if(p1->getFloatValue() != 0) {
      
      if(p1->getFloatValue() < 0) {
	
	p2->convertToNumber();
	
	if(p2->getType() == dataInt)
	  returnValue->setValue(pow(p1->getFloatValue(), p2->getIntValue()));
	
      } else {
	
	p2->convertTo(dataFloat);
	returnValue->setValue(pow(p1->getFloatValue(), p2->getFloatValue()));
	
      } 

    } else {
	
      p2->convertTo(dataFloat);
      
      if(p2->getType() == dataFloat && p2->getFloatValue() >= 0)
	returnValue->setValue(pow(p1->getFloatValue(), p2->getFloatValue()));
      
    }	
    
  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcRound(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcRound");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) 

    returnValue->setValue(p1->getIntValue());
    
  else if(p1->getType() == dataFloat)

    returnValue->setValue((int) (p1->getFloatValue() + .5));

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcSqrt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcSqrt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataFloat);

  if(p1->getType() == dataFloat)

    returnValue->setValue(sqrt(p1->getFloatValue()));

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcMaxFloat(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcMaxFloat");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  returnValue->setValue(3.40282347E+38);

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryFloat::funcMinFloat(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcMinFloat");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  returnValue->setValue(1.17549435E-38);

  return returnValue;
}


