/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptLibraryLang.cpp,v 1.6 2001/01/27 14:36:41 wouter Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>

#include "WMLScriptValueObject.h"
#include "WMLScriptFunctionParameters.h"
#include "WMLScriptLibraryLang.h"
#include "WMLScriptExitedException.h"
#include "WMLScriptAbortedException.h"
#include "WMLScriptWrongArityException.h"
#include "WMLScriptUnknownLibraryFunctionException.h"

/**
 * Creates the library.
 */
WMLScriptLibraryLang::WMLScriptLibraryLang() {
}

/**
 * Destroys the library.
 */
WMLScriptLibraryLang::~WMLScriptLibraryLang() {
}

/**
 * Returns the function ID of a given function,
 * or -1 if the function does not exist.
 * @param name the function name.
 * @return the ID of the given function or -1 if the function
 * does not exist.
 */
int WMLScriptLibraryLang::getFunctionID(const char *name) const {
  if(strcmp(name, "abs") == 0) {
    return FUNC_ABS;
  } else if(strcmp(name, "min") == 0) {
    return FUNC_MIN;
  } else if(strcmp(name, "max") == 0) {
    return FUNC_MAX;
  } else if(strcmp(name, "parseInt") == 0) {
    return FUNC_PARSEINT;
  } else if(strcmp(name, "parseFloat") == 0) {
    return FUNC_PARSEFLOAT;
  } else if(strcmp(name, "isInt") == 0) {
    return FUNC_ISINT;
  } else if(strcmp(name, "isFloat") == 0) {
    return FUNC_ISFLOAT;
  } else if(strcmp(name, "maxInt") == 0) {
    return FUNC_MAXINT;
  } else if(strcmp(name, "minInt") == 0) {
    return FUNC_MININT;
  } else if(strcmp(name, "float") == 0) {
    return FUNC_FLOAT;
  } else if(strcmp(name, "exit") == 0) {
    return FUNC_EXIT;
  } else if(strcmp(name, "abort") == 0) {
    return FUNC_ABORT;
  } else if(strcmp(name, "random") == 0) {
    return FUNC_RANDOM;
  } else if(strcmp(name, "seed") == 0) {
    return FUNC_SEED;
  } else if(strcmp(name, "characterSet") == 0) {
    return FUNC_CHARACTERSET;
  }

  return -1;
}

/**
 * Invokes the function with the given function ID with the
 * specified parameters.
 * @param id the function ID to invoke.
 * @param parameters the parameters to invoke the function with.
 * @throws WMLScriptWrongArityException
 * if the number of parameters does not match the expected
 * number of parameters.
 * @throws WMLScriptUnknownLibraryFunctionException
 * if the given function ID does not refer to a known function.
 * @throws WMLScriptExitedException
 * if id == FUNC_EXIT
 * @throws WMLScriptAbortedException
 * if id == FUNC_ABORT
 * @return the return value of the library function.
 */
WMLScriptValueObject *WMLScriptLibraryLang::call(int id, WMLScriptFunctionParameters *parameters) {
  switch(id) {
  case FUNC_ABS:
    return funcAbs(parameters);
  case FUNC_MIN:
    return funcMin(parameters);
  case FUNC_MAX:
    return funcMax(parameters);
  case FUNC_PARSEINT:
    return funcParseInt(parameters);
  case FUNC_PARSEFLOAT:
    return funcParseFloat(parameters);
  case FUNC_ISINT:
    return funcIsInt(parameters);
  case FUNC_ISFLOAT:
    return funcIsFloat(parameters);
  case FUNC_MAXINT:
    return funcMaxInt(parameters);
  case FUNC_MININT:
    return funcMinInt(parameters);
  case FUNC_FLOAT:
    return funcFloat(parameters);
  case FUNC_EXIT:
    /* does not return, throws exception instead */
    funcExit(parameters);
    break;
  case FUNC_ABORT:
    /* does not return, throws exception instead */
    funcAbort(parameters);
    break;
  case FUNC_RANDOM:
    return funcRandom(parameters);
  case FUNC_SEED:
    return funcSeed(parameters);
  case FUNC_CHARACTERSET:
    return funcCharacterSet(parameters);
  }
  throw WMLScriptUnknownLibraryFunctionException();
}

const char *WMLScriptLibraryLang::getName() const {
  return "Lang";
}

WMLScriptValueObject *WMLScriptLibraryLang::funcAbs(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcAbs");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();
  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) 

    returnValue->setValue(abs(p1->getIntValue()));
    
  else if(p1->getType() == dataFloat)

    returnValue->setValue(fabs(p1->getFloatValue()));

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcMin(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 2, "funcMin");

  WMLScriptValueObject *returnValue,
    *p1 = new WMLScriptValueObject(parameters->valueAt(0)),
    *p2 = new WMLScriptValueObject(parameters->valueAt(1));

  p1->convertTo(dataFloat);
  p2->convertTo(dataFloat);

  if(p1->getType() == dataFloat && p2->getType() == dataFloat) {

    if(p1->getFloatValue() > p2->getFloatValue())
      
      returnValue = new WMLScriptValueObject(parameters->valueAt(1));

    else

      returnValue = new WMLScriptValueObject(parameters->valueAt(0));

  } else {

    returnValue = new WMLScriptValueObject();

  }

  delete p1;
  delete p2;

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcMax(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 2, "funcMax");

  WMLScriptValueObject *returnValue,
    *p1 = new WMLScriptValueObject(parameters->valueAt(0)),
    *p2 = new WMLScriptValueObject(parameters->valueAt(1));

  p1->convertTo(dataFloat);
  p2->convertTo(dataFloat);

  if(p1->getType() == dataFloat && p2->getType() == dataFloat) {

    if(p1->getFloatValue() < p2->getFloatValue())

      returnValue = new WMLScriptValueObject(parameters->valueAt(1));

    else

      returnValue = new WMLScriptValueObject(parameters->valueAt(0));

  } else {

    returnValue = new WMLScriptValueObject();

  }

  delete p1;
  delete p2;

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcParseInt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcParseInt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0);
  char *endptr;

  if(p1->getType() == dataString) {
    
    char *s = p1->getStringValue();
    int i = strtol(s, &endptr, 10);

    if(endptr != s)
      returnValue->setValue(i);
    
  }

  return returnValue;
}

/**
 * This function may not strictly conform to the specification.
 * The C library function strtod() is used to convert a string
 * to a number, but it's semantics are a bit different from
 * Lang.parseFloat(). <p>
 * 
 * Specifically, the strtod() function returns a value when
 * parsing was only partially successfull. Lang.parseFloat()
 * should return invalid however. In order for the tests to
 * succeed, this function does perform some additional checking
 * beyond relying solely on strtod(), but but to really solve the
 * problem, a proper DecimalFloatLiteral parser should be written.
 */
WMLScriptValueObject *WMLScriptLibraryLang::funcParseFloat(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcParseFloat");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0);
  char *endptr;

  if(p1->getType() == dataString) {
    
    char *s = p1->getStringValue();
    double d = strtod(s, &endptr);

    if(endptr != s) {
      /* at least some conversion was performed ->
	 do some additional checking. */

      if(!isdigit(*endptr)) {
	/* if parsing did not stop on a digit and there
	   is an exponent sign to the left, this means
	   the literal is not well-formed */
	
	while(endptr != s && *endptr != 'e' && *endptr != 'E' && 
	      !isdigit(*endptr)) 
	  endptr--;

	if(*endptr != 'e' && *endptr != 'E')
	  returnValue->setValue(d);
	  
      }

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcIsInt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcIsInt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0),
    *intValue;

  if(p1->getType() != dataInvalid) {

    p1->convertTo(dataString);
    
    WMLScriptFunctionParameters *parameters2 = 
      new WMLScriptFunctionParameters();
    parameters2->add(new WMLScriptValueObject(p1));
    
    intValue = funcParseInt(parameters2);
    
    if(intValue->getType() == dataInt) {
      
      returnValue->setValue(true);

    } else {

      returnValue->setValue(false);

    }

    delete intValue;
    delete parameters2;

  }

  return returnValue;
}
 
WMLScriptValueObject *WMLScriptLibraryLang::funcIsFloat(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcIsFloat");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0),
    *floatValue;

  if(p1->getType() != dataInvalid) {

    p1->convertTo(dataString);
    
    WMLScriptFunctionParameters *parameters2 = 
      new WMLScriptFunctionParameters();
    parameters2->add(new WMLScriptValueObject(p1));
    
    floatValue = funcParseFloat(parameters2);
    
    if(floatValue->getType() == dataFloat) {
      
      returnValue->setValue(true);

    } else {

      returnValue->setValue(false);

    }

    delete floatValue;
    delete parameters2;

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcMaxInt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcMaxInt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  returnValue->setValue(2147483647);

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcMinInt(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcMinInt");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  returnValue->setValue(-2147483647);

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcFloat(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcFloat");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  returnValue->setValue(true);

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcRandom(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcRandom");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) {

    if(p1->getIntValue() == 0) {

      returnValue->setValue(0);

    } else if(p1->getIntValue() > 0) {

      returnValue->setValue(1 + (int) (p1->getIntValue() * rand() / 
				       (RAND_MAX + 1.0)));

    }

    /* should handle floats by using Float.int */
  }
  
  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryLang::funcSeed(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcSeed");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject(),
    *p1 = parameters->valueAt(0);

  p1->convertToNumber();

  if(p1->getType() == dataInt) {

    srand(returnValue->getIntValue());
    returnValue->setValue("");

    /* should handle floats by using Float.int */
  }
  
  return returnValue;

}

/**
 * This function returns the IANA assigned MIBEnum value of the
 * iso-8859-1 character set, which is 4 (see [WAP-203]. This function 
 * conforms to static conformance requirement [WMLSSL-045].
 */
WMLScriptValueObject *WMLScriptLibraryLang::funcCharacterSet(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "funcCharacterSet");
  return new WMLScriptValueObject(4);
}

void WMLScriptLibraryLang::funcAbort(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcAbort");

  /* should (check whether|convert if not) parameter is string value */
  throw WMLScriptAbortedException(new WMLScriptValueObject(parameters->valueAt(0)));
}

void WMLScriptLibraryLang::funcExit(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "funcExit");

  throw WMLScriptExitedException(new WMLScriptValueObject(parameters->valueAt(0)));
}

