/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptLibraryURL.cpp,v 1.6 2001/01/27 14:36:41 wouter Exp $
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "../lib/URI.h"

#include "WMLScript.h"

/**
 * Creates the library.
 */
WMLScriptLibraryURL::WMLScriptLibraryURL() {
  baseURI = NULL;
  refererURI = NULL;
}

/**
 * Destroys the library.
 */
WMLScriptLibraryURL::~WMLScriptLibraryURL() {
  delete baseURI;
  delete refererURI;
}

/**
 * Returns the function ID of a given function,
 * or -1 if the function does not exist.
 * @param name the function name.
 * @return the ID of the given function or -1 if the function
 * does not exist.
 */
int WMLScriptLibraryURL::getFunctionID(const char *name) const {
  if(strcmp(name, "isValid") == 0) {
    return FUNC_ISVALID;
  } else if(strcmp(name, "getScheme") == 0) {
    return FUNC_GETSCHEME;
  } else if(strcmp(name, "getHost") == 0) {
    return FUNC_GETHOST;
  } else if(strcmp(name, "getPort") == 0) {
    return FUNC_GETPORT;
  } else if(strcmp(name, "getPath") == 0) {
    return FUNC_GETPATH;
  } else if(strcmp(name, "getParameters") == 0) {
    return FUNC_GETPARAMETERS;
  } else if(strcmp(name, "getQuery") == 0) {
    return FUNC_GETQUERY;
  } else if(strcmp(name, "getFragment") == 0) {
    return FUNC_GETFRAGMENT;
  } else if(strcmp(name, "getBase") == 0) {
    return FUNC_GETBASE;
  } else if(strcmp(name, "getReferer") == 0) {
    return FUNC_GETREFERER;
  } else if(strcmp(name, "resolve") == 0) {
    return FUNC_RESOLVE;
  } else if(strcmp(name, "escapeString") == 0) {
    return FUNC_ESCAPESTRING;
  } else if(strcmp(name, "unescapeString") == 0) {
    return FUNC_UNESCAPESTRING;
  } else if(strcmp(name, "loadString") == 0) {
    return FUNC_LOADSTRING;
  }


  return -1;
}

/**
 * Invokes the function with the given function ID with the
 * specified parameters.
 * @param id the function ID to invoke.
 * @param parameters the parameters to invoke the function with.
 * @throws WMLScriptWrongArityException
 * if the number of parameters does not match the expected
 * number of parameters.
 * @throws WMLScriptUnknownLibraryFunctionException
 * if the given function ID does not refer to a known function.
 * @return the return value of the library function.
 */
WMLScriptValueObject *WMLScriptLibraryURL::call(int id, WMLScriptFunctionParameters *parameters) {
  switch(id) {
  case FUNC_ISVALID:
    return funcIsValid(parameters);
  case FUNC_GETSCHEME:
    return funcGetScheme(parameters);
  case FUNC_GETHOST:
    return funcGetHost(parameters);
  case FUNC_GETPORT:
    return funcGetPort(parameters);
  case FUNC_GETPATH:
    return funcGetPath(parameters);
  case FUNC_GETPARAMETERS:
    return funcGetParameters(parameters);
  case FUNC_GETQUERY:
    return funcGetQuery(parameters);
  case FUNC_GETFRAGMENT:
    return funcGetFragment(parameters);
  case FUNC_GETBASE:
    return funcGetBase(parameters);
  case FUNC_GETREFERER:
    return funcGetReferer(parameters);
  case FUNC_RESOLVE:
    return funcResolve(parameters);
  case FUNC_ESCAPESTRING:
    return funcEscapeString(parameters);
  case FUNC_UNESCAPESTRING:
    return funcUnescapeString(parameters);
  case FUNC_LOADSTRING:
    return funcLoadString(parameters);
  }

  throw WMLScriptUnknownLibraryFunctionException();
}

/**
 * Returns the name of the library.
 * @return pointer at library name.
 */
const char *WMLScriptLibraryURL::getName() const {
  return "URL";
}

/**
 * Sets the URL to consider as the base URL. This affects the result
 * from the WMLScript function call URL.getBase(). 
 * @param uri the URL to set.  
 */
void WMLScriptLibraryURL::setBaseURI(URI *uri) {
  baseURI = uri;
}

/**
 * Returns the base URL previously set with setBaseURI(). If no
 * URL was previously set, returns NULL.
 * @return the base URL.
 */
const URI *WMLScriptLibraryURL::getBaseURI() const {
  return baseURI;
}

/**
 * Sets the URL to consider as the referer URL. This affects the
 * result from the WMLScript function URL.getBase().
 * @param uri the URL to set.  
 */
void WMLScriptLibraryURL::setRefererURI(URI *uri) {
  refererURI = uri;
}

/**
 * Returns the base URL previously set with setRefererURI(). If no
 * URL was previously set, returns NULL.
 * @return the base URL.
 */
const URI *WMLScriptLibraryURL::getRefererURI() const {
  return refererURI;
}



WMLScriptValueObject *WMLScriptLibraryURL::funcIsValid(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "isValid");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());
    returnValue->setValue(u.isValid());

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetScheme(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getScheme");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getScheme())
	returnValue->setValue(u.getScheme());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetHost(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getHost");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getHost())
	returnValue->setValue(u.getHost());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;
}
 
WMLScriptValueObject *WMLScriptLibraryURL::funcGetPort(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getPort");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getPort())
	returnValue->setValue(u.getPort());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetPath(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getPath");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getFragment())
	returnValue->setValue(u.getPathWithoutParameters());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;

}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetParameters(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getParameters");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getPath()) {

	char *first_slash = strrchr(u.getPath(), '/');

	if(first_slash) {

	  char *first_semicolon = strchr(first_slash, ';');

	  if(first_semicolon) 
	    returnValue->setValue(first_semicolon + 1);
	  else
	    returnValue->setValue("");

	}

      } else {

	returnValue->setValue("");

      }

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetQuery(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getQuery");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getQuery())
	returnValue->setValue(u.getQuery());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetFragment(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "getFragment");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    URI u(p1->getStringValue());

    if(u.isValid()) {

      if(u.getFragment())
	returnValue->setValue(u.getFragment());
      else
	returnValue->setValue("");

    }

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetBase(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "getBase");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  if(getBaseURI())
    returnValue->setValue(baseURI->toString());

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcGetReferer(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 0, "getReferer");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  if(getRefererURI())
    returnValue->setValue(refererURI->toString());

  return returnValue;
}
 
WMLScriptValueObject *WMLScriptLibraryURL::funcResolve(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 2, "resolve");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);
  WMLScriptValueObject *p2 = parameters->valueAt(1);

  p1->convertTo(dataString);
  p2->convertTo(dataString);

  if(p1->getType() == dataString && p2->getType() == dataString) {

    URI *u1 = new URI(p1->getStringValue());

    /* if the baseURL's path component is an empty string, then a
       single slash character is assumed as the path. */
    if(u1->getPath() == NULL)
      u1->setAbsolutePath("");

    URI *u2 = new URI(p2->getStringValue());
    URI *u3 = URI::resolve(u1, u2);
    
    if(u3) 
      returnValue->setValue(u3->toString());
    else
      returnValue->setValue();
    
    delete u1;
    delete u2;
    delete u3;
    
  }
  
  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcEscapeString(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "escapeString");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    char *s = URI::escape(p1->getStringValue());
    returnValue->setValue(s);
    free(s);

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcUnescapeString(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 1, "unescapeString");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  WMLScriptValueObject *p1 = parameters->valueAt(0);

  p1->convertTo(dataString);

  if(p1->getType() == dataString) {

    char *s = URI::unescape(p1->getStringValue());
    returnValue->setValue(s);
    free(s);

  }

  return returnValue;
}

WMLScriptValueObject *WMLScriptLibraryURL::funcLoadString(WMLScriptFunctionParameters *parameters) {
  paramCheck(parameters, 2, "loadString");

  WMLScriptValueObject *returnValue = new WMLScriptValueObject();

  return returnValue;
}

