/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptParseTree.cpp,v 1.5 2001/01/27 14:36:42 wouter Exp $
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include "WMLScriptValueObject.h"
#include "WMLScriptParseTreeNodes.h"
#include "VarPool.h"
#include "WMLScriptFunction.h"
#include "WMLScriptPragma.h"
#include "WMLScriptParseResult.h"
#include "WMLScriptParseTree.h"

/**
 * Creates a WMLScriptParseTree from a top node.
 * @param n the top node.
 */
WMLScriptParseTree::WMLScriptParseTree(nodeType *n) {
  parseTree = n;
}

/** 
 * Destroys the WMLScriptParseTree.
 */
WMLScriptParseTree::~WMLScriptParseTree() {
  freeNode(parseTree);
}

/**
 * Returns the top node of the parse tree.
 */
nodeType *WMLScriptParseTree::getTopNode() const {
  return parseTree;
}

/**
 * Deletes the parse tree.
 */
void WMLScriptParseTree::freeNode(nodeType *p) {
  int i;

  if (!p) return;
  if (p->type == typeOpr) {
    for (i = 0; i < p->opr.nops; i++)
      freeNode(p->opr.op[i]);
  }

  free (p);
}

/**
 * Returns the line number of a parse tree node.
 * @param n the parse tree node.
 * @return line number of parse tree node, or -1 if
 * the line number is unknown.
 */
int getLineNumber(const nodeType *n) {
  switch(n->type) {
  case typeId: return n->id.location.first_line;
  case typeCon: return n->con.location.first_line;
  case typeOpr: return n->opr.location.first_line;
  }
  assert(n->type == typeId || n->type == typeCon || n->type == typeOpr);
  return -1;
}

/**
 * Returns the column number of a parse tree node.
 * @param n the parse tree node.
 * @return column number of parse tree node, or -1 if
 * the column number is unknown.
 */
int getColumnNumber(const nodeType *n) {
  switch(n->type) {
  case typeId: return n->id.location.first_column;
  case typeCon: return n->con.location.first_column;
  case typeOpr: return n->opr.location.first_column;
  }
  assert(n->type == typeId || n->type == typeCon || n->type == typeOpr);
  return -1;
}

/**
 * Unsophisticated method of displaying a parse tree on stdout.
 * @param n pointer at parse tree.
 */
void WMLScriptParseTree::dump() {
  dump(parseTree, 0);
}

void WMLScriptParseTree::dump(nodeType *n, int depth) {
  int i;
  char spaces[101];
  if(depth > 50)
    depth = 50;
  for(i = 0; i < depth * 2; i++)
    spaces[i] = ' ';
  spaces[i] = '\0';
  printf("%s", spaces);
  switch(n->type) {
  case typeId:
    printf("id\n");
    break;
  case typeCon:
    printf("con \"");
    switch((dataType) n->con.type) {
    case dataBool:
      printf(n->con.value.value.boolValue ? "true" : "false");
      break;
    case dataInt:
      printf("%d", n->con.value.value.intValue);
      break;
    case dataFloat:
      printf("%f", n->con.value.value.floatValue);
      break;
    case dataString:
      printf("%s", n->con.value.value.stringValue);
      break;
    case dataInvalid:
      printf("invalid");
      break;
    }
    printf("\"\n");
    break;
  case typeOpr:
    printf("opr %d\n", n->opr.oper);
    for(i = 0; i < n->opr.nops; i++)
      dump(n->opr.op[i], depth+1);
    break;
  default:
    printf("??? %d\n", n->type);
    break;
  }
}



