/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptParser.cpp,v 1.9 2001/01/27 14:36:42 wouter Exp $
 */

#include <stdio.h>

#include "../lib/Garbage.h"

#include "WMLScript.h"

extern int yyparse(void *parseResult);

/**
 * Constructs a WMLScriptParser.
 */
WMLScriptParser::WMLScriptParser() {
}

/**
 * Destructs the WMLScriptParser.
 */
WMLScriptParser::~WMLScriptParser() {
}

/**
 * This function parses a string of WMLScript code
 * and returns the result as a WMLScriptParseContext.
 * @param s a pointer to a string of WMLScript code.
 * @throws WMLScriptParseException
 * When a parse error occurs. The exception is guaranteed to contain
 * a more-or-less informative error message which can be retrieved
 * using WMLScriptParseException::getMessage().
 * @throws WMLScriptOutOfMemoryException
 * @return pointer to WMLScriptParseContext
 */
WMLScriptParseContext *WMLScriptParser::parse(char *s) {
  WMLScriptParseResult parseResult;

  /* initialize parseResult, this structure is used by 
     yyparse() to communicate with us. */
  parseResult.numberOfFunctions = 0;
  parseResult.functionTable = NULL;
  parseResult.numberOfPragmas = 0;
  parseResult.pragmaTable = NULL;
  parseResult.conditionalGarbage.count = 0;
  parseResult.conditionalGarbage.table = NULL;
  parseResult.unconditionalGarbage.count = 0;
  parseResult.unconditionalGarbage.table = NULL;
  parseResult.lineNumber = 0;
  parseResult.columnNumber = 0;
  strcpy(parseResult.parseErrorMessage, "parse error");

  /* flex/bison stuff follows. */
  flex_scan_start();
  void *bufferState = flex_scan_string(s);
  int err = yyparse(&parseResult);
  flex_delete_buffer(bufferState);
  /* end of flex/bison stuff. */

  if(err) {
    /* on parse error, collect garbage and leave the 
       function by throwing an exception */

    /* first release the function table and associated data ... */
    if(parseResult.functionTable) {

      for(int i = 0; i < parseResult.numberOfFunctions; i++) {

	free(parseResult.functionTable[i].name);
	DestroyVarPool(parseResult.functionTable[i].varPool);

      }

      free(parseResult.functionTable);

    }

    GarbageRelease(&parseResult.conditionalGarbage);
    GarbageRelease(&parseResult.unconditionalGarbage);
    
    WMLScriptParseException parseException(parseResult.lineNumber, parseResult.columnNumber);

    parseException.setMessage(parseResult.parseErrorMessage);

    throw parseException;
    /* NOT REACHED */
  }

  /* if the parse was successfull then there are no dangling
     node pointers so we can release the garbage store */
  if(parseResult.conditionalGarbage.table) 
    free(parseResult.conditionalGarbage.table);

  GarbageRelease(&parseResult.unconditionalGarbage);

  WMLScriptParseContext *parseContext = 
    new WMLScriptParseContext(&parseResult);  

  return parseContext;

}
