/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptTester.cpp,v 1.16 2001/01/27 14:36:42 wouter Exp $
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "WMLScript.h"

#undef WMLS_DEBUG

/* Change the following #define's to enable/disable tests. */

/** Conformance tests. */
#define TEST_CONFORMANCE
/** Conformance: string escape tests. */
#define TEST_ESCAPES
/** Conformance: type conversion tests. */
#define TEST_TYPE
/** Conformance: function tests. */
#define TEST_FUNCTIONS
/** Conformance: variable tests. */
#define TEST_VARIABLES
/** Conformance: operator tests. */
#define TEST_OPERATORS
/** Conformance: control (loop) tests. */
#define TEST_CONTROL
/** Conformance: compound expressions tests. */
#define TEST_COMPOUND
/** Conformance: Lang library methods tests. */
#define TEST_LIBLANG
/** Conformance: Float library methods tests. */
#define TEST_LIBFLOAT
/** Conformance: String library methods tests. */
#define TEST_LIBSTRING
/** Conformance: URL library methods tests. */
#define TEST_LIBURL
/** Conformance: formatting tests. */
#define TEST_FORMATTING
/** Conformance: pragma tests. */
#define TEST_PRAGMAS

/** All of the following tests are disabled because some
    compilers seem to have a problem with the funky method 
    pointer magic used in them. */

/** All error handling tests. */
#undef TEST_ERROR
/** Error handling: misplaced break tests. */
#undef TEST_BREAK
/** Error handling: misplaced continue tests. */
#undef TEST_CONTINUE
/** Error handling: arity checking. */
#undef TEST_WRONGARITY
/** Error handling: declaration checking. */
#undef TEST_DECLARATION
/** Error handling: parse error checking. */
#undef TEST_PARSER
/** Error handling: function access checking. */
#undef TEST_ACCESS
/** Error handling: linking. */
#undef TEST_LINKING
/** Error handling: call parameters. */
#undef TEST_CALLPARAMS
/** Error handling: nested comments. */
#undef TEST_NESTEDCOMMENTS
/** Error handling: escape sequences. */
#undef TEST_WRONGESCAPE
/** Error handling: string termination. */
#undef TEST_STRINGTERMINATION

/**
 * Bison debugging flag.
 */
extern int yydebug;

/**
 * Constructs the test object.
 */
WMLScriptTester::WMLScriptTester() {
  /* enable bison debugging? */
  yydebug = 0; 
  totalTestsRun = 0;
  totalTestsFailed = 0;
  libLang = new WMLScriptLibraryLang();
  libFloat = new WMLScriptLibraryFloat();
  libString = new WMLScriptLibraryString();
  libURL = new WMLScriptLibraryURL();
  ((WMLScriptLibraryURL *) libURL)->setBaseURI(new URI("http://b.com/"));
  ((WMLScriptLibraryURL *) libURL)->setRefererURI(new URI("http://a.com/"));
}

/**
 * Destroys the test object.
 */
WMLScriptTester::~WMLScriptTester() {
  delete libString;
  delete libFloat;
  delete libLang;
  delete libURL;
}

/**
 * Runs all tests and returns the number of tests that failed.
 *
 * @return the number of tests that failed.
 */
int WMLScriptTester::test() {
  totalTestsRun = 0;
  totalTestsFailed = 0;

#ifdef TEST_CONFORMANCE
  testConformance();
#endif

#ifdef TEST_ERROR
  testErrorHandling();
#endif

#ifdef TEST_PRAGMAS
  testPragmas();
#endif

  printf("Test total:\t\t\ttests run: %3d, tests failed: %d\n", totalTestsRun, totalTestsFailed);
  return totalTestsFailed;
}

/**
 * This method runs many tests testing the conformance of 
 * the WMLScript parser and interpreter.
 *
 * @return the number of tests that failed.
 */
int WMLScriptTester::testConformance() {

#ifdef TEST_ESCAPES
  Test escapes[] = {
    { "extern function test() { return \"\\\\\"; };", new WMLScriptValueObject("\\") },
    { "extern function test() { return \"\\\\\\\\\"; };", new WMLScriptValueObject("\\\\") },
    { "extern function test() { return \"\\n\"; };", new WMLScriptValueObject("\n") },
    { "extern function test() { return \"\\'\"; };", new WMLScriptValueObject("'") },
    { "extern function test() { return \"\\\\\\\"\"; };", new WMLScriptValueObject("\\\"") },
    { NULL, NULL }
  };
#endif

#ifdef TEST_TYPE
  Test typeConversions[] = {
    { "extern function test() { return 12 div 4; };", new WMLScriptValueObject(3) }, 
    { "extern function test() { return \"7\" << 2; };", new WMLScriptValueObject(28) },
    { "extern function test() { return true && 3.4; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 1 && 0; };", new WMLScriptValueObject(false) },
    { "extern function test() { return \"A \\n\\' \\\"  \" || \"\"; };", new WMLScriptValueObject(true) },
    { "extern function test() { return !42; };", new WMLScriptValueObject(false) },
    { "extern function test() { return !invalid; };", new WMLScriptValueObject() },
    { "extern function test() { return 3 && invalid; };", new WMLScriptValueObject() },
    { "extern function test() { return \"7\" << 2; };", new WMLScriptValueObject(28) },
    { "extern function test() { return true << 2; };", new WMLScriptValueObject(4) },
    { "extern function test() { return 7.2 >> 3; };", new WMLScriptValueObject() },
    { "extern function test() { return 2.1 div 4; };", new WMLScriptValueObject() },
    { "extern function test() { return +10; };", new WMLScriptValueObject(10) },
    { "extern function test() { return -10.3; };", new WMLScriptValueObject(-10.3) },
    { "extern function test() { return -\"33\"; };", new WMLScriptValueObject(-33) },
    { "extern function test() { return +\"47.3\"; };", new WMLScriptValueObject(47.3) },
    { "extern function test() { return +true; };", new WMLScriptValueObject(1) },
    { "extern function test() { return -false; };", new WMLScriptValueObject(0) },
    { "extern function test() { return -\"ABC\"; };", new WMLScriptValueObject() },
    { "extern function test() { return -\"9e9999\"; };", new WMLScriptValueObject() },
    { "extern function test() { return 100/10.3; };", new WMLScriptValueObject(100/10.3) },
    { "extern function test() { return 33*44; };", new WMLScriptValueObject(1452) },
    { "extern function test() { return \"10\"*3; };", new WMLScriptValueObject(30) },
    { "extern function test() { return 3.4*\"4.3\"; };", new WMLScriptValueObject(14.62) },
    { "extern function test() { return \"10\"-\"2\"; };", new WMLScriptValueObject(8) },
    { "extern function test() { return \"2.3\"*\"3\"; };", new WMLScriptValueObject(2.3*3) },
    { "extern function test() { return 3.2*\"A\"; };", new WMLScriptValueObject() },
    { "extern function test() { return .9*\"9e999\"; };", new WMLScriptValueObject() },
    { "extern function test() { return invalid*1; };", new WMLScriptValueObject() },
    { "extern function test() { return 12+3; };", new WMLScriptValueObject(15) },
    { "extern function test() { return 32.4+65; };", new WMLScriptValueObject(97.4) },
    { "extern function test() { return \"12\"+5.4; };", new WMLScriptValueObject("125.400000") },
    { "extern function test() { return 43.2<77; };", new WMLScriptValueObject(true) },
    { "extern function test() { return \"Hey\"<56; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 2.7+\"4.2\"; };", new WMLScriptValueObject("2.7000004.2") },
    { "extern function test() { return 9.9+true; };", new WMLScriptValueObject(10.9) },
    { "extern function test() { return 3<false; };", new WMLScriptValueObject(false) },
    { "extern function test() { return \"A\"+invalid; };", new WMLScriptValueObject() },
    { "extern function test() { return isvalid \"123\"; };", new WMLScriptValueObject(true) },
    { "extern function test() { return isvalid (1/0); };", new WMLScriptValueObject(false) },
    { "extern function test() { return ((1/0) == invalid); };", new WMLScriptValueObject() },
    { "extern function test() { return (1/0) || foo(); };", new WMLScriptValueObject() },
    { "extern function test() { return true || (1/0); };", new WMLScriptValueObject(true) },
    { "extern function test() { return false || (1/0); };", new WMLScriptValueObject() },
    { "extern function test() { return 9.3E+3; };", new WMLScriptValueObject(9300.00) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_FUNCTIONS
  Test functionCalls[] = {
    { "extern function test() { return plus(1,2); }; function plus(a, b) { return a+b; };", new WMLScriptValueObject(3) },
    { "extern function test() { return plus(1, plus(1,2)); }; extern function plus(a, b) { return a+b; };", new WMLScriptValueObject(4) },
    { "extern function test() { return plus(plus(1,2), 1); }; extern function plus(a, b) { return a+b; };", new WMLScriptValueObject(4) },
    { "extern function test() { return plus(1, plus(2, plus(1, 2))); }; extern function plus(a, b) { return a+b; };", new WMLScriptValueObject(6) },
    { "extern function test() { return plus(1, plus(plus(2, 1), 2)); }; extern function plus(a, b) { return a+b; };", new WMLScriptValueObject(6) },
    { "extern function test() { return plus(plus(1, 2), plus(1,2)); }; extern function plus(a, b) { return a+b; };", new WMLScriptValueObject(6) },
    { "extern function test() { return triplus(1, plus(1, 2), plus(1,2)); }; extern function plus(a, b) { return a+b; }; extern function triplus(a, b, c) { return a+b+c; };", new WMLScriptValueObject(7) },
    { "extern function test() { return null(); }; extern function null() { var a=3; };", new WMLScriptValueObject("") },
    { NULL, NULL }
  };
#endif

#ifdef TEST_VARIABLES
  Test variableExpressions[] = {
    { "extern function test() { var a=010; return a; };", new WMLScriptValueObject(8) },
    { "extern function test() { var a=0xff; return a; };", new WMLScriptValueObject(255) },
    { "extern function test() { var a, b; a=1; b=3; return a+b; };", new WMLScriptValueObject(4) },
    { "extern function test() { var a = 1, b='a'; return a+b; };", new WMLScriptValueObject("1a") },
    { "extern function test() { var a=3, b=2; return a*b; }", new WMLScriptValueObject(6) },
    { "extern function test() { var a=3, b=2; return a*(b+2); }", new WMLScriptValueObject(12) },
    { "extern function test() { var a=3, b=2; return mul(a, b); }; function mul(a, b) { return a*b; }", new WMLScriptValueObject(6) },
    { NULL, NULL }
  };   
#endif

#ifdef TEST_OPERATORS
  Test operators[] = { 
    { "extern function test() { return; };", new WMLScriptValueObject("") },
    { "extern function test() { return 1 << 64; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 1 << 32; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 1 << 16; };", new WMLScriptValueObject(65536) },
    { "extern function test() { return 1 >>> 1; };", new WMLScriptValueObject(0) },
    { "extern function test() { return ~1; };", new WMLScriptValueObject(-2) },
    { "extern function test() { return !true; };", new WMLScriptValueObject(false) },
    { "extern function test() { return true && true; };", new WMLScriptValueObject(true) },
    { "extern function test() { return false && true; };", new WMLScriptValueObject(false) },
    { "extern function test() { return false || false; };", new WMLScriptValueObject(false) },
    { "extern function test() { return false || true; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 1 & 1; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 1 | 1; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 3 % 2; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 3 div 5; };", new WMLScriptValueObject(0) },
    { "extern function test() { var a = 1; return a++; };", new WMLScriptValueObject(1) },
    { "extern function test() { var a = 1; return ++a; };", new WMLScriptValueObject(2) },
    { "extern function test() { var a = 1; return a--; };", new WMLScriptValueObject(1) },
    { "extern function test() { var a = 1; return --a; };", new WMLScriptValueObject(0) },
    { "extern function test() { return +1; };", new WMLScriptValueObject(1) },
    { "extern function test() { return -1; };", new WMLScriptValueObject(-1) },
    { "extern function test() { return 1+1; };", new WMLScriptValueObject(2) },
    { "extern function test() { return 'a'+'b'; };", new WMLScriptValueObject("ab") },
    { "extern function test() { return 1+'b'; };", new WMLScriptValueObject("1b") },
    { "extern function test() { return 1-1; };", new WMLScriptValueObject(0) },
    { "extern function test() { return 1*1; };", new WMLScriptValueObject(1) },
    { "extern function test() { return 1/1; };", new WMLScriptValueObject(1.0) },
    { "extern function test() { return 3/4; };", new WMLScriptValueObject(0.75) },
    { "extern function test() { return 3.0/4; };", new WMLScriptValueObject(0.75) },
    { "extern function test() { return 1<2; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 'a'<'b'; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 1<=1; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 'a'<='a'; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 1>2; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 'a'>'b'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 1>=2; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 'a'>='b'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 1==1; };", new WMLScriptValueObject(true) },
    { "extern function test() { return '1'==1; };", new WMLScriptValueObject(true) },
    { "extern function test() { return '1'=='1'; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 1=='1'; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 2=='1'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return '2'=='1'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 2==1; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 1!='1'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return 1!='2'; };", new WMLScriptValueObject(true) },
    { "extern function test() { return 2!='2'; };", new WMLScriptValueObject(false) },
    { "extern function test() { return '2'!='2'; };", new WMLScriptValueObject(false) },
    { "extern function test() { var a=3; a*=2; return a; };", new WMLScriptValueObject(6) },
    { "extern function test() { var a=2; a-=2; return a; };", new WMLScriptValueObject(0) },
    { "extern function test() { var a=2; a/=2; return a; };", new WMLScriptValueObject(1.0) },
    { "extern function test() { var a=2; a+=2; return a; };", new WMLScriptValueObject(4) },
    { "extern function test() { var a=2; return typeof a; };", new WMLScriptValueObject(0) },
    { "extern function test() { var a=0.2; return typeof a; };", new WMLScriptValueObject(1) },
    { "extern function test() { var a='2'; return typeof a; };", new WMLScriptValueObject(2) },
    { "extern function test() { var a=true; return typeof a; };", new WMLScriptValueObject(3) },
    { "extern function test() { var a=invalid; return typeof a; };", new WMLScriptValueObject(4) },
    { "extern function test() { var a=invalid; return isvalid a; };", new WMLScriptValueObject(false) },
    { "extern function test() { var a=1; return isvalid a; };", new WMLScriptValueObject(true) },
    { "extern function test() { var a=1; return a == 1 ? 1 : 2; };", new WMLScriptValueObject(1) },
    { "extern function test() { var a=4; return a == 1 ? 1 : 2; };", new WMLScriptValueObject(2) },
    { "extern function test() { var a=4; return a; };", new WMLScriptValueObject(4) },
    { "extern function test() { var a=4; return a, a*2; };", new WMLScriptValueObject(8) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_CONTROL
  Test controlStructs[] = {
    { "extern function test() { var a=1; while(a < 10) { a++; }; return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { var a=1; while(a++ < 10); return a; };", new WMLScriptValueObject(11) },
    { "extern function test() { var a=1; while(++a < 10); return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { var a=1; while(true) { if(a++ > 10) break; } return a; };", new WMLScriptValueObject(12) },
    { "extern function test() { var a=1, b=0; while(true) { if(a++ < 10) b++; else break; } return b+a; };", new WMLScriptValueObject(20) },
    { "extern function test() { var a=0, b=0; while(true) { if(a++ > 10) break; if(a == 5) continue; b++; } return b; };", new WMLScriptValueObject(10) },
    { "extern function test() { for(var a=1; ; a++) { if(a == 10) break; } return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { var b = 0; for(var a=1; ; a++) { if(a == 10) break; if(a == 5) continue; b++; } return b; };", new WMLScriptValueObject(8) },
    { "extern function test() { for(var a=1; a<10; a++); return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { for(var a=1; a<10; ++a); return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { var a=1; for(a=1; a<10; ++a); return a; };", new WMLScriptValueObject(10) },
    { "extern function test() { var a, b, c = 0; for(a=0; a < 10; a++) { for(b=0; b < 10; b++) { c++; } } return c; };", new WMLScriptValueObject(100) },
    { "extern function test() { var a, b, c, d, e = 0; for(a=0; a < 10; a++) { for(b=0; b < 10; b++) { for(c=0; c < 10; c++) { for(d=0; d < 10; d++) { e++; } } } } return e; };", new WMLScriptValueObject(10000) },
    { "extern function test() { var a, b, c = 0; for(a=0; a < 10; a++) { b = 0; while(b++ < 10) { c++; } } return c; };", new WMLScriptValueObject(100) },
    { "extern function test() { var a, b, c = 0; for(a=0; a < 100; a++) { b = 0; while(b++ < 10) { if(c > 500) break; c++; } } return c; };", new WMLScriptValueObject(501) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_COMPOUND
  Test compoundExpressions[] = {
    { "extern function test() { return -(6+(2*4-(7*8))); };", new WMLScriptValueObject(42) },
    { "extern function test() { return 3.0/(12/(1+2)); };", new WMLScriptValueObject(0.75) },
    { "extern function test() { var a = 3.0/(12/(1+2)), b = 2.0/(12/(1+2)); return a/b; };", new WMLScriptValueObject(1.5) },
    { "extern function test() { var a = 3.0/(12^(1+2)), b = 2.0/(12/(1+2)); return a*b; };", new WMLScriptValueObject(0.1) },
    { "extern function test() { var a = (50 & 1) + -1, b = -1; return a*b; };", new WMLScriptValueObject(1) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_LIBLANG
  Test libLang[] = {
    { "extern function test() { return Lang.abs(-3); };", new WMLScriptValueObject(3) },
    { "extern function test() { var a = -3; var b = Lang.abs(a); var c = Lang.min(a, b); return c; }", new WMLScriptValueObject(-3) },
    { "extern function test() { return Lang.min(45, 76.3); };", new WMLScriptValueObject(45) },    
    { "extern function test() { return Lang.min(45, 45.0); };", new WMLScriptValueObject(45) },    
    { "extern function test() { var a = -3; var b = Lang.abs(a); var c = Lang.max(a, b); return c; }", new WMLScriptValueObject(3) },
    { "extern function test() { return Lang.max(45.5, 76); };", new WMLScriptValueObject(76) },    
    { "extern function test() { return Lang.max(45.0, 45); };", new WMLScriptValueObject(45.0) },    
    { "extern function test() { return Lang.parseInt('1234'); };", new WMLScriptValueObject(1234) },    
    { "extern function test() { return Lang.parseInt('100 m/s'); };", new WMLScriptValueObject(100) },    
    { "extern function test() { return Lang.parseFloat('123.7'); };", new WMLScriptValueObject(123.7) },    
    { "extern function test() { return Lang.parseFloat('  +7.34e2 Hz'); };", new WMLScriptValueObject(7.34e+2) },    
    { "extern function test() { return Lang.parseFloat(' 70e-2 F'); };", new WMLScriptValueObject(70.0e-2) },    
    { "extern function test() { return Lang.parseFloat('-.1 C'); };", new WMLScriptValueObject(-0.1) },    
    { "extern function test() { return Lang.parseFloat(' 100 '); };", new WMLScriptValueObject(100.0) },    
    { "extern function test() { return Lang.parseFloat('Number: 5.5'); };", new WMLScriptValueObject() },    
    { "extern function test() { return Lang.parseFloat('7.3e meters'); };", new WMLScriptValueObject() },    
    { "extern function test() { return Lang.parseFloat('7.3e- m/s'); };", new WMLScriptValueObject() },    
    { "extern function test() { return Lang.isInt(' -123'); };", new WMLScriptValueObject(true) },
    { "extern function test() { return Lang.isInt(' 123.33'); };", new WMLScriptValueObject(true) },
    { "extern function test() { return Lang.isInt('string'); };", new WMLScriptValueObject(false) },
    { "extern function test() { return Lang.isInt('#123'); };", new WMLScriptValueObject(false) },
    { "extern function test() { return Lang.isInt(invalid); };", new WMLScriptValueObject() },
    { "extern function test() { return Lang.isFloat(' -123'); };", new WMLScriptValueObject(true) },
    { "extern function test() { return Lang.isFloat(' 123.33'); };", new WMLScriptValueObject(true) },
    { "extern function test() { return Lang.isFloat('string'); };", new WMLScriptValueObject(false) },
    { "extern function test() { return Lang.isFloat('#123'); };", new WMLScriptValueObject(false) },
    { "extern function test() { return Lang.isFloat(invalid); };", new WMLScriptValueObject() },
    { "extern function test() { return Lang.maxInt(); };", new WMLScriptValueObject(2147483647) },
    { "extern function test() { return Lang.minInt(); };", new WMLScriptValueObject(-2147483647) },
    { "extern function test() { return Lang.float(); };", new WMLScriptValueObject(true) },
    
    { "extern function test() { return Lang.exit('exit'); };", new WMLScriptValueObject("exit") },
    { "extern function test() { return Lang.abort('abort'); };", new WMLScriptValueObject("abort") },
    { "extern function test() { return Lang.random(0); };", new WMLScriptValueObject(0) },
    { "extern function test() { return Lang.random(-1); };", new WMLScriptValueObject() },
    { "extern function test() { return Lang.random('string'); };", new WMLScriptValueObject() },
    { "extern function test() { return Lang.seed(123); };", new WMLScriptValueObject("") },
    { "extern function test() { return Lang.seed('seed'); };", new WMLScriptValueObject() },
    { "extern function test() { return Lang.characterSet(); };", new WMLScriptValueObject(4) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_LIBFLOAT
  Test libFloat[] = {
    { "extern function test() { return Float.int(3.14); };", new WMLScriptValueObject(3) },
    { "extern function test() { return Float.int(-2.8); };", new WMLScriptValueObject(-2) },
    { "extern function test() { return Float.int('invalid'); };", new WMLScriptValueObject() },
    { "extern function test() { return Float.floor(3.14); };", new WMLScriptValueObject(3) },
    { "extern function test() { return Float.floor(-2.8); };", new WMLScriptValueObject(-3) },
    { "extern function test() { return Float.ceil(3.14); };", new WMLScriptValueObject(4) },
    { "extern function test() { return Float.ceil(-2.8); };", new WMLScriptValueObject(-2) },
    { "extern function test() { return Float.pow(3, 2); };", new WMLScriptValueObject(9.0) },
    { "extern function test() { return Float.pow(0, -1); };", new WMLScriptValueObject() },
    { "extern function test() { return Float.pow(-1, 3.2); };", new WMLScriptValueObject() },
    { "extern function test() { return Float.round(3.5); };", new WMLScriptValueObject(4) },
    { "extern function test() { return Float.round(-3.5); };", new WMLScriptValueObject(-3) },
    { "extern function test() { return Float.round(0.5); };", new WMLScriptValueObject(1) },
    { "extern function test() { return Float.round(-0.5); };", new WMLScriptValueObject(0) },
    { "extern function test() { return Float.sqrt(4); };", new WMLScriptValueObject(2.0) },
    { "extern function test() { return Float.sqrt(5); };", new WMLScriptValueObject(sqrt(5)) },
    { "extern function test() { return Float.maxFloat(); };", new WMLScriptValueObject(3.40282347E+38) },
    { "extern function test() { return Float.minFloat(); };", new WMLScriptValueObject(1.17549435E-38) },
    { NULL, NULL }
  };
#endif

#ifdef TEST_LIBSTRING
  Test libString[] = {
    { "extern function test() { return String.length('ABC'); };", new WMLScriptValueObject(3) },
    { "extern function test() { return String.length(''); };", new WMLScriptValueObject(0) },
    { "extern function test() { return String.length(342); };", new WMLScriptValueObject(3) },
    { "extern function test() { return String.isEmpty('Hello'); };", new WMLScriptValueObject(false) },
    { "extern function test() { return String.isEmpty(''); };", new WMLScriptValueObject(true) },
    { "extern function test() { return String.isEmpty(true); };", new WMLScriptValueObject(false) },
    { "extern function test() { return String.charAt('My name is Joe', 0); };", new WMLScriptValueObject("M") },
    { "extern function test() { return String.charAt('My name is Joe', 100); };", new WMLScriptValueObject("") },
    { "extern function test() { return String.charAt(34, 0); };", new WMLScriptValueObject("3") },
    { "extern function test() { return String.charAt('My name is Joe', 'first'); };", new WMLScriptValueObject() },
    { "extern function test() { return String.subString('ABCD', 1, 2); };", new WMLScriptValueObject("BC") },
    { "extern function test() { return String.subString('ABCD', 2, 5); };", new WMLScriptValueObject("CD") },
    { "extern function test() { return String.subString(1234, 0, 2); };", new WMLScriptValueObject("12") },
    { "extern function test() { return String.find('abcde', 'cd'); };", new WMLScriptValueObject(2) },
    { "extern function test() { return String.find(34.2, 'de'); };", new WMLScriptValueObject(-1) },
    { "extern function test() { return String.find('abcde', 'qz'); };", new WMLScriptValueObject(-1) },
    { "extern function test() { return String.find(34, '3'); };", new WMLScriptValueObject(0) },
    { "extern function test() { return String.find('abcde', ''); };", new WMLScriptValueObject() },
    { "extern function test() { var newName = 'Don'; var oldName = 'Joe'; return String.replace('Hello Joe. What is up Joe?', oldName, newName); };", new WMLScriptValueObject("Hello Don. What is up Don?") },
    { "extern function test() { var newName = 'Don'; var oldName = 'Joe'; return String.replace('Hello Don. What is up Don?', newName, oldName); };", new WMLScriptValueObject("Hello Joe. What is up Joe?") },
    { "extern function test() { return String.elements('My name is Joe; Age 50;', ' '); };", new WMLScriptValueObject(6) },
    { "extern function test() { return String.elements('My name is Joe; Age 50;', ';'); };", new WMLScriptValueObject(3) },
    { "extern function test() { return String.elements('', ';'); };", new WMLScriptValueObject(1) },
    { "extern function test() { return String.elements('a', ';'); };", new WMLScriptValueObject(1) },
    { "extern function test() { return String.elements(';', ';'); };", new WMLScriptValueObject(2) },
    { "extern function test() { return String.elements(';;,;', ';,'); };", new WMLScriptValueObject(4) },
    { "extern function test() { return String.elements(';;,;', ''); };", new WMLScriptValueObject() },
    { "extern function test() { return String.elementAt('My name is Joe; Age 50;', 0, ' '); };", new WMLScriptValueObject("My") },
    { "extern function test() { return String.elementAt('My name is Joe; Age 50;', 14, ';'); };", new WMLScriptValueObject("") },
    { "extern function test() { return String.elementAt('My name is Joe; Age 50;', 1, ';'); };", new WMLScriptValueObject(" Age 50") },
    { "extern function test() { return String.elementAt('My name is Joe; Age 50;', 1, ''); };", new WMLScriptValueObject() },
    { "extern function test() { return String.removeAt('A A;  B C D', 1, ' '); };", new WMLScriptValueObject("A  B C D") },
    { "extern function test() { return String.removeAt('A A;  B C D', 0, ';'); };", new WMLScriptValueObject("  B C D") },
    { "extern function test() { return String.removeAt('A A;  B C D', 14, ';'); };", new WMLScriptValueObject("A A") },
    { "extern function test() { return String.removeAt('A A;  B C D', 14, ''); };", new WMLScriptValueObject() },
    { "extern function test() { return String.replaceAt('B C; E', 'A', 0, ' '); };", new WMLScriptValueObject("A C; E") },
    { "extern function test() { return String.replaceAt('B C; E', 'F', 5, ';'); };", new WMLScriptValueObject("B C;F") },
    { "extern function test() { return String.replaceAt('B C; E', 'F', 5, ''); };", new WMLScriptValueObject() },
    { "extern function test() { return String.insertAt('B C; E', 'A', 0, ' '); };", new WMLScriptValueObject("A B C; E") },
    { "extern function test() { return String.insertAt('B C; E', 'X', 3, ' '); };", new WMLScriptValueObject("B C; E X") },
    { "extern function test() { return String.insertAt('B C; E', 'D', 1, ';'); };", new WMLScriptValueObject("B C;D; E") },
    { "extern function test() { return String.insertAt('B C; E', 'F', 5, ';'); };", new WMLScriptValueObject("B C; E;F") },
    { "extern function test() { return String.squeeze('Hello'); };", new WMLScriptValueObject("Hello") },
    { "extern function test() { return String.squeeze(\" Bye    Jon . \\r\\n See you!  \"); };", new WMLScriptValueObject(" Bye Jon . See you! ") },
    { "extern function test() { return String.trim(\"Hello\"); };", new WMLScriptValueObject("Hello") },
    { "extern function test() { return String.trim(\" Bye    Jon .  See you!  \"); };", new WMLScriptValueObject("Bye    Jon .  See you!") },
    { "extern function test() { return String.compare('Hello', 'Hello'); };", new WMLScriptValueObject(0) },
    { "extern function test() { return String.compare('Bye', 'Jon'); };", new WMLScriptValueObject(-1) },
    { "extern function test() { return String.compare('Jon', 'Bye'); };", new WMLScriptValueObject(1) },
    { "extern function test() { return String.toString(12); };", new WMLScriptValueObject("12") },
    { "extern function test() { return String.toString(true); };", new WMLScriptValueObject("true") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('e: %6d', a); };", new WMLScriptValueObject("e:     45") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%6d', b); };", new WMLScriptValueObject("   -45") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%6.4d', a); };", new WMLScriptValueObject("  0045") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%6.4d', b); };", new WMLScriptValueObject(" -0045") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('Do it %s', c); };", new WMLScriptValueObject("Do it now") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%3f', d); };", new WMLScriptValueObject("1.234568") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%10.2f%%', d); };", new WMLScriptValueObject("      1.23%") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%3f %2f.', d); };", new WMLScriptValueObject("1.234568 .") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%.0d', 0); };", new WMLScriptValueObject("") },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%7d', 'Int'); };", new WMLScriptValueObject() },
    { "extern function test() { var a=45; var b=-45; var c='now'; var d=1.2345678; return String.format('%s', true); };", new WMLScriptValueObject("true") },
    { NULL, NULL }
  };
#endif

#ifdef TEST_LIBURL
  Test libURL[] = {
    { "extern function test() { return URL.isValid(\"http://w.hst.com/script#func()\"); };", new WMLScriptValueObject(true) },
    { "extern function test() { return URL.isValid(\"../common#test()\"); };", new WMLScriptValueObject(true) },
    { "extern function test() { return URL.isValid(\"experimental?://www.host.com/cont>\"); };", new WMLScriptValueObject(false) },
    { "extern function test() { return URL.getScheme(\"http://w.h.com/path#frag\"); };", new WMLScriptValueObject("http") },
    { "extern function test() { return URL.getScheme(\"w.h.com/path#frag\"); };", new WMLScriptValueObject("") },
    { "extern function test() { return URL.getHost(\"http://w.h.com/path#frag\"); };", new WMLScriptValueObject("w.h.com") },
    { "extern function test() { return URL.getHost(\"path#frag\"); };", new WMLScriptValueObject("") },
    { "extern function test() { return URL.getHost(\"zyx://me@ismo.k.oksa#fab\"); };", new WMLScriptValueObject("ismo.k.oksa") },
    { "extern function test() { return URL.getPort(\"http://w.h.com:80/path#frag\"); };", new WMLScriptValueObject("80") },
    { "extern function test() { return URL.getPort(\"http://w.h.com/path#frag\"); };", new WMLScriptValueObject("") },
    { "extern function test() { return URL.getPath(\"http://w.h.com/home/sub/comp#frag\"); };", new WMLScriptValueObject("/home/sub/comp") },
    { "extern function test() { return URL.getPath(\"../home/sub/comp#frag\"); };", new WMLScriptValueObject("../home/sub/comp") },
    { "extern function test() { return URL.getPath(\"http://w.a.p/a;x/b;y=1/c;fg#a\"); };", new WMLScriptValueObject("/a/b/c") },
    { "extern function test() { return URL.getPath(\"http://w.a.p/a;x/b;y=1/c#b\"); };", new WMLScriptValueObject("/a/b/c") },
    { "extern function test() { return URL.getParameters(\"http://w.h.com/script;3;2?x=1&y=3\"); };", new WMLScriptValueObject("3;2") },
    { "extern function test() { return URL.getParameters(\"../script;3;2?x=1&y=3\"); };", new WMLScriptValueObject("3;2") },
    { "extern function test() { return URL.getParameters(\"http://w.a.p/a;x/b;y=1/c;fg\"); };", new WMLScriptValueObject("fg") },
    { "extern function test() { return URL.getParameters(\"http://w.a.p/a;x/b;y=1/c\"); };", new WMLScriptValueObject("") },
    { "extern function test() { return URL.getQuery(\"http://w.h.com/home;3;2?x=1&y=3\"); };", new WMLScriptValueObject("x=1&y=3") },
    { "extern function test() { return URL.getFragment(\"http://w.h.com/cont#frag\"); };", new WMLScriptValueObject("frag") },
    { "extern function test() { return URL.getBase(); };", new WMLScriptValueObject("http://b.com/") },
    { "extern function test() { return URL.getReferer(); };", new WMLScriptValueObject("http://a.com/") },
    { "extern function test() { return URL.resolve(\"http://foo.com/\", \"foo.vcf\"); };", new WMLScriptValueObject("http://foo.com/foo.vcf") },
    { "extern function test() { return URL.resolve(\"http://foo.com\", \"c\"); };", new WMLScriptValueObject("http://foo.com/c") },
    { "extern function test() { return URL.resolve(\"http://foo.com\", \"/c\"); };", new WMLScriptValueObject("http://foo.com/c") },
    { "extern function test() { return URL.resolve(\"http://foo.com\", \"?q\"); };", new WMLScriptValueObject("http://foo.com/?q") },
    { "extern function test() { return URL.resolve(\"http://\", \"x\"); };", new WMLScriptValueObject("http:///x") },
    { "extern function test() { return URL.escapeString(\"http://w.h.com/dck?x=\\u007f#crd\"); };", new WMLScriptValueObject("http%3a%2f%2fw.h.com%2fdck%3fx%3d%7f%23crd") },
    { "extern function test() { return URL.escapeString(\"http://w.h.com/dck?x=\\\\u007f#crd\"); };", new WMLScriptValueObject("http%3a%2f%2fw.h.com%2fdck%3fx%3d%5cu007f%23crd") },
    { "extern function test() { return URL.unescapeString(\"http%3a%2f%2fw.h.com%2fdck%3fx%3d12%23crd\"); };", new WMLScriptValueObject("http://w.h.com/dck?x=12#crd") },
    { "extern function test() { return URL.loadString(\"www.altavista.com\", \"text/html\"); };", new WMLScriptValueObject("") },
    { NULL, NULL }
  };
#endif

#ifdef TEST_FORMATTING
  Test formatting[] = {
    { "/* this is a comment\nspanning multiple lines. */\r\nextern function test() { return 1; };", new WMLScriptValueObject(1) },
    { "// /* this is a comment\n// second line */\nextern function test() { return 1; };", new WMLScriptValueObject(1) },
    { "extern function test() { return \"Quote: \\\"\"; };", new WMLScriptValueObject("Quote: \"") },
    //    { "extern function test() { return 'Specials: \x00 \' \b''; };", new WMLScriptValueObject("Specials") },
    {NULL, NULL}
  };
#endif

#ifdef TEST_ESCAPES
  printf("Testing string escapes:\t\t");
  runConformanceTests((Test *) (escapes));
#endif

#ifdef TEST_TYPE
  printf("Testing type conversions:\t");
  runConformanceTests((Test *) (typeConversions));
#endif
  
#ifdef TEST_FUNCTIONS
  printf("Testing function calls:\t\t");
  runConformanceTests((Test *) (functionCalls));
#endif

#ifdef TEST_VARIABLES
  printf("Testing variable expressions:\t");
  runConformanceTests((Test *) (variableExpressions));
#endif

#ifdef TEST_OPERATORS
  printf("Testing operators:\t\t");
  runConformanceTests((Test *) (operators));
#endif

#ifdef TEST_CONTROL
  printf("Testing control structures:\t");
  runConformanceTests((Test *) (controlStructs));
#endif

#ifdef TEST_COMPOUND
  printf("Testing compound expressions:\t");
  runConformanceTests((Test *) (compoundExpressions));
#endif

#ifdef TEST_LIBLANG
  printf("Testing library Lang:\t\t");
  runConformanceTests((Test *) (libLang));
#endif

#ifdef TEST_LIBFLOAT
  printf("Testing library Float:\t\t");
  runConformanceTests((Test *) (libFloat));
#endif

#ifdef TEST_LIBSTRING
  printf("Testing library String:\t\t");
  runConformanceTests((Test *) (libString));
#endif

#ifdef TEST_LIBURL
  printf("Testing library URL:\t\t");
  runConformanceTests((Test *) (libURL));
#endif

#ifdef TEST_FORMATTING
  printf("Testing formatting\t\t");
  runConformanceTests((Test *) (formatting));
#endif

  return totalTestsFailed;
}

/**
 * This function tests pragma handling.
 * 
 * @return the number of tests failed.
 */
int WMLScriptTester::testPragmas() {

  printf("Testing pragmas ...\n");

  printf("\t\tuse url\t\t\t\t");
  testPragma("use url OtherScript \"http://www.host.com/app/script\";\nextern function test() { return; };", typeExternalCompilationUnit, typeName, "OtherScript", "http://www.host.com/app/script", NULL);

  printf("\t\tuse access, path + domain\t");
  testPragma("use access domain \"wapforum.org\" path \"/finance\";\nextern function test() { return; };", typeAccess, typeName, "wapforum.org", "/finance", NULL);

  printf("\t\tuse access, domain only\t\t");
  testPragma("use access domain \"wapforum.org\";\nextern function test() { return; };", typeAccess, typeName, "wapforum.org", NULL, NULL);

  printf("\t\tuse access, path only\t\t");
  testPragma("use access path \"/finance\";\nextern function test() { return; };", typeAccess, typeName, NULL, "/finance", NULL);

  printf("\t\tuse meta name\t\t\t");
  testPragma("use meta name \"Created\" \"18-March-1998\";\nextern function test() { return; };", typeMeta, typeName, "Created", "18-March-1998", NULL);

  printf("\t\tuse meta http equiv\t\t");
  testPragma("use meta http equiv \"Keywords\" \"Script,Language\";\nextern function test() { return; };", typeMeta, typeHTTPEquiv, "Keywords", "Script,Language", NULL);

  printf("\t\tuse meta user agent\t\t");
  testPragma("use meta user agent \"Type\" \"Test\";\nextern function test() { return; };", typeMeta, typeUserAgent, "Type", "Test", NULL);

  return totalTestsFailed;

}

/**
 * This function parses a script containing a pragma and
 * decides whether the pragma was successfully parsed.
 * @param script the script to parse.
 * @param expectType the expected type of the pragma.
 * @param expectMetaType the expected meta type of the pragma, if this
 * pragma is a meta pragma.
 * @param expectData1 first expected data.
 * @param expectData1 second expected data.
 * @param expectData1 third expected data.
 */
bool WMLScriptTester::testPragma(char *script, 
				 WMLScriptPragmaType expectType,
				 WMLScriptMetaPragmaType expectMetaType,
				 char *expectData1,
				 char *expectData2,
				 char *expectData3) {
  bool testOK = false;

  WMLScriptParseContext *parseContext = NULL;

  try {

    WMLScriptParser parser = WMLScriptParser();
    parseContext = parser.parse(script);
    WMLScriptPragmaPool *pragmaPool = parseContext->getPragmaPool();

    /* check for the right number of pragma's. */
    if(pragmaPool->getNumberOfPragmas(expectType) == 1) {

      WMLScriptPragma *pragma = pragmaPool->getPragma(expectType, 0);

      /* check for the right type of pragma's. */
      if(pragma->type == expectType) {

	switch(pragma->type) {

	case typeExternalCompilationUnit:
	  /* check for the right data in the pragma's. */
	  if(strcmp(pragma->ext.id, expectData1) == 0 &&
	     strcmp(pragma->ext.url, expectData2) == 0)
	    testOK = true;
	  break;
	  
	case typeAccess:
	  /* either path or domain may be NULL, but not both. */
	  if(expectData1 == NULL &&
	     pragma->access.domain == NULL) {
	    
	    assert(pragma->access.path != NULL);

	    if(strcmp(pragma->access.path, expectData2) == 0)
	      testOK = true;

	  } else if(expectData2 == NULL &&
	     pragma->access.path == NULL) {

	    assert(pragma->access.domain != NULL);

	    if(strcmp(pragma->access.domain, expectData1) == 0)
	      testOK = true;

	  } else {

	    assert(pragma->access.domain != NULL);
	    assert(pragma->access.path != NULL);

	    if(strcmp(pragma->access.domain, expectData1) == 0 &&
	       strcmp(pragma->access.path, expectData2) == 0) 
	      testOK = true;

	  }
	  break;

	case typeMeta:
	  /* check that meta type is correct. */
	  if(pragma->meta.metaType == expectMetaType) {

	    if(expectData3 == NULL) {

	      if(strcmp(pragma->meta.key, expectData1) == 0 &&
		 strcmp(pragma->meta.value, expectData2) == 0 &&
		 pragma->meta.scheme == NULL)
		testOK = true;
	      
	    } else {

	      assert(pragma->meta.scheme != NULL);

	      if(strcmp(pragma->meta.key, expectData1) == 0 &&
		 strcmp(pragma->meta.value, expectData2) == 0 &&
		 strcmp(pragma->meta.scheme, expectData3) == 0)
		testOK = true;

	    }

	  }

	  break;

	}

      }

    }

  } catch(WMLScriptException&) {

    printf("exception while parsing! ");

  }

  delete parseContext;

  totalTestsRun++;
  if(!testOK) totalTestsFailed++;

  printf("%s\n", (testOK ? "ok" : "failed"));

  return testOK;
  
}

/**
 * This function runs a number of methods that
 * test error handling.
 *
 * @return the number of tests failed.
 */
int WMLScriptTester::testErrorHandling() {
  ErrorTest tests[] = {
#ifdef TEST_BREAK
    { "\t\tmisplaced break\t\t\t", &WMLScriptTester::errorTestBreak },
#endif
#ifdef TEST_CONTINUE
    { "\t\tmisplaced continue\t\t", &WMLScriptTester::errorTestContinue },
#endif
#ifdef TEST_WRONGARITY
    { "\t\twrong arity library\t\t", &WMLScriptTester::errorTestWrongArityLibrary },
    { "\t\twrong arity function\t\t", &WMLScriptTester::errorTestWrongArityFunction },
    { "\t\twrong arity call\t\t", &WMLScriptTester::errorTestWrongArityCall },
#endif
#ifdef TEST_DECLARATION
    { "\t\tundeclared variable\t\t", &WMLScriptTester::errorTestVarDecl },
    { "\t\tundeclared variable (postfix)\t", &WMLScriptTester::errorTestVarDeclPost },
    { "\t\tundeclared function\t\t", &WMLScriptTester::errorTestFuncDecl },
    { "\t\tundeclared library\t\t", &WMLScriptTester::errorTestLibDecl },
    { "\t\tundeclared lib function\t\t", &WMLScriptTester::errorTestLibFuncDecl },
    { "\t\tundeclared call\t\t\t", &WMLScriptTester::errorTestCallDecl },
#endif
#ifdef TEST_PARSER
    { "\t\tparse error\t\t\t", &WMLScriptTester::errorTestParser },
#endif
#ifdef TEST_ACCESS
    { "\t\tfunction access violation\t", &WMLScriptTester::errorTestAccess },
#endif
#ifdef TEST_LINKING
    { "\t\tlibrary linking\t\t\t", &WMLScriptTester::errorTestLibLinking },
#endif
#ifdef TEST_CALLPARAMS
    { "\t\tcall parameters\t\t\t", &WMLScriptTester::errorTestCallParams },  
    { "\t\tcall return\t\t\t", &WMLScriptTester::errorTestCallReturn },  
#endif
#ifdef TEST_NESTEDCOMMENTS
    { "\t\tnested comments\t\t\t", &WMLScriptTester::errorTestNestedComments },  
#endif
#ifdef TEST_WRONGESCAPE
    { "\t\thex nul escapes\t\t\t", &WMLScriptTester::errorTestHexNulEscape },  
    { "\t\toctal nul escapes\t\t", &WMLScriptTester::errorTestOctalNulEscape },  
    { "\t\tUnicode nul escapes\t\t", &WMLScriptTester::errorTestUnicodeNulEscape },  
    { "\t\tUnicode multibyte escapes\t", &WMLScriptTester::errorTestUnicodeMultibyteEscape },  
#endif
#ifdef TEST_STRINGTERMINATION
    { "\t\tunterminated string\t\t", &WMLScriptTester::errorTestUnterminatedString },  
#endif
    { NULL, NULL }
  };

  printf("Testing error handling ...\n");

  int i = 0;
  while(tests[i].description) {
    
    printf("%s", tests[i].description);
    if(!(callMemberFunction(*this, tests[i].test)())) {

      totalTestsFailed++;
      printf("failed");

    } else {
      
      printf("ok");

    }
    totalTestsRun++;
    printf("\n");
    i++;

  }

  return totalTestsFailed;

}

bool WMLScriptTester::errorTestBreak() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { break; };");
  
  } catch (WMLScriptBreakException& e) {

    if(e.getColumnNumber() == 25 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestContinue() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { continue; };");
  
  } catch (WMLScriptContinueException& e) {

    if(e.getColumnNumber() == 25 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestWrongArityLibrary() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return Lang.abs(-1, -1); };");
  
  } catch (WMLScriptWrongArityException& e) {

    if(e.getColumnNumber() == 48 || e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong message, line, column: %s, %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestWrongArityFunction() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return plus(1); }; function plus(a, b) { return a+b; };");
  
  } catch (WMLScriptWrongArityException& e) {

    if(strcmp(e.getMessage(), "plus") == 0 &&
       e.getColumnNumber() == 32 || e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong message, line, column: %s, %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestWrongArityCall() {
  bool testOK = false;

  WMLScriptParseContext *parseContext = NULL;
  WMLScriptParser parser = WMLScriptParser();

  try {

    parseContext = parser.parse("extern function test(a) { return a; };");
    WMLScriptInterpreter interpreter(parseContext);
    WMLScriptFunctionParameters parameters;
    delete interpreter.call("test", &parameters);

  } catch(WMLScriptWrongArityException& e) {

    if(strcmp(e.getMessage(), "test") == 0 &&
       e.getLineNumber() == -1 && e.getColumnNumber() == -1)
      testOK = true;
    else
      printf("wrong message, line, column: %s, %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
    
  }

  delete parseContext;

  return testOK;

}

bool WMLScriptTester::errorTestVarDecl() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { a=1; }; ");
  
  } catch (WMLScriptUndeclaredIdentifierException& e) {

    if(e.getColumnNumber() == 25 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestVarDeclPost() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { a++; }; ");
  
  } catch (WMLScriptUndeclaredIdentifierException& e) {

    if(e.getColumnNumber() == 25 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestFuncDecl() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return plus(1); }; ");
  
  } catch (WMLScriptUndeclaredIdentifierException& e) {

    if(e.getColumnNumber() == 32 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestLibDecl() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return NONEXIST.a(1); }; ");
  
  } catch (WMLScriptUnknownLibraryException& e) {

    if(e.getColumnNumber() == 32 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestLibFuncDecl() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return Lang.a(1); }; ");
  
  } catch (WMLScriptUnknownLibraryFunctionException& e) {

    if(e.getColumnNumber() == 32 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());
    
  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestCallDecl() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test2() { return Lang.a(1); }; ");
  
  } catch (WMLScriptUndeclaredIdentifierException& e) {

    if(strcmp(e.getMessage(), "test") == 0)
      testOK = true;
    else
      printf("wrong message, %s\n", e.getMessage());
    
  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;

}

bool WMLScriptTester::errorTestParser() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { eturn Lang.a(1); }; ");
  
  } catch (WMLScriptParseException& e) {

    if(e.getColumnNumber() == 35 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());
    
  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestAccess() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("function test() { return Lang.a(1); }; ");
  
  } catch (WMLScriptAccessException&) {

    testOK = true;

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestLibLinking() {
  int fail = 0;

  WMLScriptParseContext *parseContext = NULL;

  WMLScriptParser parser = WMLScriptParser();

  try {

    parseContext = parser.parse("extern function test() { };");

  } catch(WMLScriptException&) {
    
    fail++;

  }

  WMLScriptInterpreter interpreter(parseContext);
  /** multiple links and unlinks to test linking/unlinking code */
  if(interpreter.linkLibrary(libLang) != 1)
    fail++;

  if(interpreter.linkLibrary(libFloat) != 2)
    fail++;

  if(interpreter.linkLibrary(libString) != 3)
    fail++;

  if(interpreter.unlinkLibrary(libLang) != 2)
    fail++;

  if(interpreter.unlinkLibrary(libFloat) != 1)
    fail++;

  if(interpreter.unlinkLibrary(libString) != 0)
    fail++;

  if(interpreter.linkLibrary(libLang) != 1)
    fail++;

  if(interpreter.linkLibrary(libFloat) != 2)
    fail++;

  if(interpreter.linkLibrary(libString) != 3)
    fail++;

  if(interpreter.unlinkLibrary(NULL) != -ENOTFOUND)
    fail++;

  delete parseContext;

  return (fail == 0 ? true : false);
}

bool WMLScriptTester::errorTestCallParams() {
  int fail = 0;

  WMLScriptParseContext *parseContext = NULL;
  WMLScriptValueObject *val = NULL;

  WMLScriptParser parser = WMLScriptParser();

  try {

    parseContext = parser.parse("extern function test(a, b) { return a * b; };");
    WMLScriptInterpreter interpreter(parseContext);
    WMLScriptFunctionParameters parameters;
    parameters.add(3);
    parameters.add(4);
    val = interpreter.call("test", &parameters);

    if(val == NULL || val->getType() != dataInt || val->getIntValue() != 12)
      fail++;
    
  } catch(WMLScriptException&) {
    
    fail++;

  }

  delete val;
  delete parseContext;

  return (fail == 0 ? true : false);
}

bool WMLScriptTester::errorTestCallReturn() {
  int fail = 0;

  WMLScriptParseContext *parseContext = NULL;
  WMLScriptValueObject *val = NULL;

  WMLScriptParser parser = WMLScriptParser();

  try {

    parseContext = parser.parse("extern function test(a, b) { };");
    WMLScriptInterpreter interpreter(parseContext);
    WMLScriptFunctionParameters parameters;
    parameters.add(3);
    parameters.add(4);
    val = interpreter.call("test", &parameters);

    if(val == NULL || val->getType() != dataString ||
       *(val->getStringValue()) != '\0')
      fail++;
    
  } catch(WMLScriptException&) {
    
    fail++;

  }

  delete val;
  delete parseContext;

  return (fail == 0 ? true : false);
}

bool WMLScriptTester::errorTestNestedComments() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("/* first comment /* second comment */ end of first comment */");
  
  } catch (WMLScriptParseException& e) {

    if(e.getColumnNumber() == 61 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("wrong line, column: %d, %d\n", e.getLineNumber(), e.getColumnNumber());

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestHexNulEscape() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return \"\\x00\"; };");
  
  } catch (WMLScriptParseException& e) {

    if(strcmp("unsupported NUL escape", e.getMessage()) == 0 && 
       e.getColumnNumber() == 37 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("error %s at line, column: %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());
    

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestOctalNulEscape() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return \"\\0\"; };");
  
  } catch (WMLScriptParseException& e) {

    if(strcmp("unsupported NUL escape", e.getMessage()) == 0 && 
       e.getColumnNumber() == 35 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("error %s at line, column: %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());
    
  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestUnicodeNulEscape() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return \"\\u0000\"; };");
  
  } catch (WMLScriptParseException& e) {

    if(strcmp("unsupported NUL escape", e.getMessage()) == 0 && 
       e.getColumnNumber() == 39 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("error %s at line, column: %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());    

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestUnicodeMultibyteEscape() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return \"\\uff00\"; };");
  
  } catch (WMLScriptParseException& e) {

    if(strcmp("unsupported multibyte Unicode escape", e.getMessage()) == 0 && 
       e.getColumnNumber() == 39 && e.getLineNumber() == 0)
      testOK = true;
    else
      printf("error %s at line, column: %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());    

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

bool WMLScriptTester::errorTestUnterminatedString() {
  bool testOK = false;
  WMLScriptValueObject *val = NULL;

  try {

    val = eval("extern function test() { return \"bkabkabla\n\"; };");
  
  } catch (WMLScriptParseException& e) {

    if(strcmp("unterminated string constant", e.getMessage()) == 0 && 
       e.getColumnNumber() == 0 && e.getLineNumber() == 1)
      testOK = true;
    else
      printf("error %s at line, column: %d, %d\n", e.getMessage(), e.getLineNumber(), e.getColumnNumber());    

  } catch(WMLScriptException&) {
  }

  delete val;

  return testOK;
}

void WMLScriptTester::runConformanceTests(Test *tests) {
  char *errors[] = {
    { "wrong result" },
    { "unknown exception" },
    { "misplaced return" },
    { "misplaced break" },
    { "misplaced continue" },
    { "corrupt parsetree" },
    { "parse error" },
    { "no result" }
  };

  int i = 0;
  WMLScriptValueObject *val = NULL;
  int fail = 0;
  int testError = 0;

  while(tests[i].expr) {

    testError = 0;

    try {

      val = NULL;
      val = eval(tests[i].expr);

      if(!val)

	testError = 8;

      else if(!val->equals(tests[i].expected)) 

	testError = 1;

    } catch(WMLScriptParseException& e) {

      printf("parse error at line %d, column %d\n", e.getLineNumber(), e.getColumnNumber());
      testError = 7;

    } catch(WMLScriptReturnException&) {

      testError = 3;

    } catch(WMLScriptBreakException& e) {

      printf("misplaced break at line %d, column %d\n", e.getLineNumber(), e.getColumnNumber());
      testError = 4;

    } catch(WMLScriptContinueException&) {

      testError = 5;

    } catch(WMLScriptCorruptParseTreeException&) {

      testError = 6;

    } catch(WMLScriptException&) {
      
      testError = 2;
      
    }

    if(testError > 0) {
      fail++;
      printf("\n*** test failed (%s): %s ", errors[testError - 1], tests[i].expr);

      if(val) {

	char *s1 = val->toString();
	char *s2 = tests[i].expected->toString();
	printf(" --> result: %s, expected: %s\n", s1, s2);
	free(s1);
	free(s2);

      } else {

	printf(" --> test did not return a result\n");

      }

    }

    /* val is NULL on exception or on */
    if(val) 
      delete val;

    delete tests[i].expected;

    i++;
  }

  printf("tests run: %3d, tests failed: %d\n", i, fail);
  totalTestsRun += i;
  totalTestsFailed += fail;
}

/**
 * Parses and evaluates a script.
 *
 * @throws WMLScriptException
 * (except WMLScriptAbortedException and WMLScriptExitedException).
 * @return the result of the WMLScript evaluation.
 */
WMLScriptValueObject *WMLScriptTester::eval(char *exp) {
  WMLScriptValueObject *value = NULL;
  WMLScriptParseContext *parseContext = NULL;

  try {

    WMLScriptParser parser = WMLScriptParser();
    parseContext = parser.parse(exp);
    WMLScriptInterpreter interpreter(parseContext);
    WMLScriptFunctionParameters parameters;
    interpreter.linkLibrary(libLang);
    interpreter.linkLibrary(libFloat);
    interpreter.linkLibrary(libString);
    interpreter.linkLibrary(libURL);
    value = interpreter.call("test", &parameters);

  } catch(WMLScriptAbortedException& e) {

    value = e.getValue();

  } catch(WMLScriptExitedException& e) {

    value = e.getValue();
    
  } catch(WMLScriptException&) {
    
    delete value;
    delete parseContext;

    throw;

  }

#ifdef WMLS_DEBUG
  char *s = value->toString();
  printf("WMLScriptTester::eval: test %s\n\t\t---> type: %d, value: %s\n", exp, value->getType(), s);
  free(s);
#endif

  delete parseContext;

  return value;
}

