/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptValueObject.cpp,v 1.15 2001/01/27 14:36:42 wouter Exp $
 */

#include <math.h>
#include <stdarg.h>
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#ifdef HAVE_IEEEFP_H /* SunOS 5.6 needs this */
#include <ieeefp.h>
#endif

#ifdef WIN32
#include <float.h>
#define finite(a) _finite(a)
#endif

#include "WMLScriptValueObject.h"
#include "WMLScriptOutOfMemoryException.h"

#undef WMLS_DEBUG

extern int errno;

/**
 * Creates a WMLScriptValueObject with the value invalid.
 * @throws WMLScriptOutOfMemoryException
 * when memory can not be acquired.
 */
WMLScriptValueObject::WMLScriptValueObject() {
  initialize();
  setValue();
}

/**
 * Creates a WMLScriptValueObject with the specified value.
 * @param v the boolean value.
 * @throws WMLScriptOutOfMemoryException
 * when memory can not be acquired.
 */
WMLScriptValueObject::WMLScriptValueObject(bool v) {
  initialize();
  setValue(v);
}

/**
 * Creates a WMLScriptValueObject with the specified value.
 * @param v the integer value.
 * @throws WMLScriptOutOfMemoryException
 * when memory can not be acquired.
 */
WMLScriptValueObject::WMLScriptValueObject(int v) {
  initialize();
  setValue(v);
}

/**
 * Creates a WMLScriptValueObject with the specified value.
 * NOTE: If the value is positive or negative infinity or NaN, then the
 * result is an invalid value.
 * @param v the floating point value.
 * @throws WMLScriptOutOfMemoryException
 * when memory can not be acquired.
 */
WMLScriptValueObject::WMLScriptValueObject(double v) {
  initialize();
  setValue(v);
}

/**
 * Creates a WMLScriptValueObject with the specified value.
 * @param v the string value.
 * @throws WMLScriptOutOfMemoryException
 * when memory can not be acquired.
 */
WMLScriptValueObject::WMLScriptValueObject(char *v) {
  initialize();
  setValue(v);
}

/**
 * Clones a WMLScriptValueObject from another WMLScriptValueObject.
 * @param orig the original WMLScriptValueObject.
 */
WMLScriptValueObject::WMLScriptValueObject(WMLScriptValueObject *orig) {
  copy(*orig);
}

/**
 * Destroys the WMLScriptValueObject.
 */
WMLScriptValueObject::~WMLScriptValueObject() {
  destroy();
}

/**
 * Assign a boolean to the value.
 * @param v boolean value.
 */
void WMLScriptValueObject::setValue(bool v) {
  value->type = dataBool;
  value->value.boolValue = v ? 1 : 0;
}

/**
 * Assign an integer to the value.
 * @param v integer value.
 */
void WMLScriptValueObject::setValue(int v) {
  value->type = dataInt;
  value->value.intValue = v;
}

/**
 * Assign a double precision floating point value to the value.  
 * NOTE: If the value is positive or negative infinity or NaN, then the
 * result is an invalid value.  
 * @param v double precision floating point value.  
 */
void WMLScriptValueObject::setValue(double v) {

  if(!finite(v)) {

    value->type = dataInvalid;
    
  } else {
  
    value->type = dataFloat;
    value->value.floatValue = v;
    
  }

}

/**
 * Assign a string to the value.
 * @param v pointer to a string to assign. A copy of
 * the string is stored.
 */
void WMLScriptValueObject::setValue(const char *v) {
  value = (WMLScriptValue *) realloc(value, sizeof(WMLScriptValue) + strlen(v));
  if(value == NULL)
    throw WMLScriptOutOfMemoryException();
  value->type = dataString;
  strcpy(value->value.stringValue, v);
}

/**
 * Assign invalid to the value.
 */
void WMLScriptValueObject::setValue() {
  value->type = dataInvalid;
}

/**
 * Returns the integer value of the value; only meaningful if 
 * getType() returns dataInt.
 * @return the integer value of the value.
 */
int WMLScriptValueObject::getIntValue() const {
  return value->value.intValue;
}

/**
 * Returns the double precision floating point value of the 
 * value; only meaningful if getType() returns dataFloat.
 * @return the double precision floating point value of the value.
 */
double WMLScriptValueObject::getFloatValue() const {
  return value->value.floatValue;
}

/**
 * Returns a pointer to the string value of the 
 * value; only meaningful if getType() returns dataString.
 * @return the boolean value of the value.
 */
char *WMLScriptValueObject::getStringValue() const {
  return value->value.stringValue;
}
 
/**
 * Returns the boolean value of the 
 * value; only meaningful if getType() returns dataBool.
 * @return the boolean value of the value.
 */
bool WMLScriptValueObject::getBoolValue() const {
  return(value->value.boolValue ? true : false);
}

/**
 * Returns the current type of the value.
 * @return the type of the value.
 */
dataType WMLScriptValueObject::getType() const {
  return value->type;
}

/**
 * Attempts to convert the value into a number,
 * that is, either an integer or a floating point
 * value. It will always prefer conversion to
 * an integer value, if possible. The result of 
 * the conversion may be invalid.
 */
void WMLScriptValueObject::convertToNumber() {
  if(value->type == dataString)
    stringToNumber();

  if(value->type == dataBool)
    convertTo(dataInt);
}

/**
 * Attempts to concert the value into the specified type.
 * The result of the conversion may be invalid.
 * @param type desired type to convert the value into.
 */
void WMLScriptValueObject::convertTo(dataType type) {
  WMLScriptValue *conv_val, *val = value;
  conv_val = (WMLScriptValue *) malloc(sizeof(WMLScriptValue));

  /* assume conversion is legal; if it isn't, this will be picked up
     by the conversion code and the type will be appropriately adjusted. */
  conv_val->type = type;

  switch(val->type) {
  case dataBool:
    /* conversion from boolean */
    switch(type) {
    case dataBool:
      conv_val->value.boolValue = val->value.boolValue;
      break;
    case dataInt:
      conv_val->value.intValue = (val->value.boolValue ? 1 : 0);
      break;
    case dataFloat:
      conv_val->value.floatValue = (val->value.boolValue ? 1 : 0);
      break;
    case dataString:     
      conv_val = 
	(WMLScriptValue *) realloc(conv_val, sizeof(WMLScriptValue) + 6);
      if(conv_val == NULL)
	throw WMLScriptOutOfMemoryException();
      strcpy(conv_val->value.stringValue, (val->value.boolValue ? "true" : "false"));
      break;
    case dataInvalid:
      conv_val->type = dataInvalid;
      break;
    }
    break;
  case dataInt:
    /* conversion from integer */
    switch(type) {
    case dataBool:
      conv_val->value.boolValue = (val->value.intValue ? 1 : 0);
      break;
    case dataInt:
      conv_val->value.intValue = val->value.intValue;
      break;
    case dataFloat:
      conv_val->value.floatValue = val->value.intValue;
      break;
    case dataString:
      conv_val = 
	(WMLScriptValue *) realloc(conv_val, sizeof(WMLScriptValue) + 
				   sizeof(val->value.intValue) * 4);
      if(conv_val == NULL)
	throw WMLScriptOutOfMemoryException();
      sprintf(conv_val->value.stringValue, "%d", val->value.intValue);
      break;
    case dataInvalid:
      conv_val->type = dataInvalid;
      break;
    }
    break;
  case dataFloat:
    /* conversion from float */
    switch(type) {
    case dataBool:
      conv_val->value.boolValue = (val->value.floatValue ? 1 : 0);
      break;
    case dataInt:
      /* conversion from float to integer is illegal */
      conv_val->type = dataInvalid;
      break;
    case dataFloat:
      conv_val->value.floatValue = val->value.floatValue;
      break;
    case dataString:
      conv_val = 
	(WMLScriptValue *) realloc(conv_val, 100); // probably excessive
      if(conv_val == NULL)
	throw WMLScriptOutOfMemoryException();
      sprintf(conv_val->value.stringValue, "%f", val->value.floatValue);
      break;
    case dataInvalid:
      conv_val->type = dataInvalid;
      break;
    }
    break;
  case dataString:
    /* conversion from string */
    switch(type) {
    case dataBool:
      conv_val->value.boolValue = (strlen(val->value.stringValue) ? 1 : 0);
      break;
    case dataInt:
      if(strlen(val->value.stringValue)) {
	errno = 0;
	char *endptr;
	conv_val->value.intValue = strtol(val->value.stringValue, &endptr, 0);
	if(*endptr != '\0' || errno)
	  conv_val->type = dataInvalid;
	
      } else {
	/* conversion from empty string to integer is illegal */
	conv_val->type = dataInvalid;
      }
      break;
    case dataFloat:
      if(strlen(val->value.stringValue)) {
	errno = 0;
	char *endptr;
	conv_val->value.floatValue = strtod(val->value.stringValue, &endptr);
	if(*endptr != '\0' || errno)
	  conv_val->type = dataInvalid;

      } else {
	/* conversion from empty string to float is illegal */
	conv_val->type = dataInvalid;
      }      
      break;
    case dataString:
      conv_val = 
	(WMLScriptValue *) realloc(conv_val, sizeof(WMLScriptValue) +
				   strlen(val->value.stringValue) + 1);
      if(conv_val == NULL)
	throw WMLScriptOutOfMemoryException();
      strcpy(conv_val->value.stringValue, val->value.stringValue);
      break;
    case dataInvalid:
      conv_val->type = dataInvalid;
      break;
    }
    break;
  case dataInvalid:
    conv_val->type = dataInvalid;
    break;
  }

  free(val);

  value = conv_val;

}

/**
 * Attempts to convert a string to a number, that is, either
 * an integer or floating point value. The result of the conversion
 * may be invalid.
 */
void WMLScriptValueObject::stringToNumber() {
  WMLScriptValue *val = value;
  WMLScriptValueObject *floatVal = NULL, *intVal = NULL;

  if(val->type == dataString) {
    floatVal = new WMLScriptValueObject(this);
    intVal = new WMLScriptValueObject(this);
    floatVal->convertTo(dataFloat);
    intVal->convertTo(dataInt);

    if(floatVal->getType() == dataFloat && intVal->getType() == dataInt &&
       floatVal->getFloatValue() == intVal->getIntValue()) {
      
      convertTo(dataInt);

    } else if(floatVal->getType() == dataFloat && 
	      intVal->getType() == dataInt &&
	      floatVal->getFloatValue() != intVal->getIntValue()) {

      convertTo(dataFloat);

    } else if(floatVal->getType() == dataFloat && 
	      intVal->getType() != dataInt) {

      convertTo(dataFloat);

    } else if(intVal->getType() == dataInt && 
	      floatVal->getType() != dataFloat) {

      convertTo(dataInt);

    } else {

      setValue();

    }

  }

  delete floatVal;
  delete intVal;

}

/**
 * Constructs a string representation of the value and returns
 * a pointer to it.
 * The caller is responsible for releasing the memory 
 * held by the string representation.
 * @return pointer at string representation.
 */
char *WMLScriptValueObject::toString() {
  char *s = (char*) malloc(251);
  WMLScriptValue *val = value;

  switch(val->type) {
  case dataBool:
    strcpy(s, (val->value.boolValue ? "true" : "false"));
    break;
  case dataInt:
    sprintf(s, "%d", val->value.intValue);
    break;
  case dataFloat:
    sprintf(s, "%f", val->value.floatValue);
    break;
  case dataString:
    if(val->value.stringValue[0]) {
      s = (char *) realloc(s, strlen(val->value.stringValue) + 1);
      strcpy(s, val->value.stringValue);
    } else {
      strcpy(s, "");
    }
    break;
  case dataInvalid:
    strcpy(s, "invalid");
    break;
  }

  return s;
}

/**
 * Attempts to convert the value and another value to the
 * same type. The conversion may yield the invalid type.
 * @param v the other value
 * @param minAllowedType the minimum allowed common type.
 * @param maxAllowedType the maximum allowed common type.
 */
void WMLScriptValueObject::convertToCommon(WMLScriptValueObject *v, dataType minAllowedType, dataType maxAllowedType) {

  if(getType() == dataString && maxAllowedType < dataString)
    convertToNumber();

  if(v->getType() == dataString && maxAllowedType < dataString)
    v->convertToNumber();

  if(getType() == dataInvalid || v->getType() == dataInvalid) {

    dprintf("convertToCommon: invalid\n");
    convertTo(dataInvalid);
    v->convertTo(dataInvalid);

  } else if(maxAllowedType >= dataString && (minAllowedType == dataString ||
					     (getType() == dataString || 
					      v->getType() == dataString))) {

    dprintf("convertToCommon: string\n");
    convertTo(dataString);
    v->convertTo(dataString);

  } else if(maxAllowedType >= dataFloat && (minAllowedType == dataFloat || 
					    (getType() == dataFloat || 
					     v->getType() == dataFloat))) {

    dprintf("convertToCommon: float\n");
    convertTo(dataFloat);
    v->convertTo(dataFloat);

  } else if(maxAllowedType >= dataInt && (minAllowedType == dataInt || 
					  (getType() == dataInt || 
					   v->getType() == dataInt))) {

    dprintf("convertToCommon: integer\n");
    convertTo(dataInt);
    v->convertTo(dataInt);

  } else if(maxAllowedType >= dataBool && (minAllowedType == dataBool || 
					   (getType() == dataBool || 
					    v->getType() == dataBool))) {

    dprintf("convertToCommon: boolean\n");
    convertTo(dataBool);
    v->convertTo(dataBool);

  }

}

/**
 * Compares this value to another value and evaluates whether
 * they are equal. Two values are equal if they have identical types
 * and identical values.
 * @param v value to compare with.
 * @return true if both values are equal.
 */
bool WMLScriptValueObject::equals(WMLScriptValueObject *v) const {
  if(getType() != v->getType()) 
    return false;

  if(getType() == dataInvalid)
    return true;

  if(v->getType() == dataBool)
    return(getBoolValue() == v->getBoolValue() ? true : false);
  if(v->getType() == dataInt)
    return(getIntValue() == v->getIntValue() ? true : false);
  if(v->getType() == dataFloat)
    return(getFloatValue() == v->getFloatValue() ? true : false);
  if(v->getType() == dataString)
    return((strcmp(getStringValue(), v->getStringValue()) == 0) ? true : false);

  return false;
}

/**
 * This method implements a copy constructor for the
 * WMLScriptValueObject.
 */
void WMLScriptValueObject::copy(const WMLScriptValueObject &orig) {
  size_t size = sizeof(WMLScriptValue);

  if(orig.getType() == dataString)
    size += strlen(orig.getStringValue());

  value = (WMLScriptValue *) malloc(size);
  if(value == NULL)
    throw WMLScriptOutOfMemoryException();

  value->type = orig.getType();

  switch(orig.getType()) {
  case dataBool:
    setValue(orig.getBoolValue());
    break;
  case dataInt:
    setValue(orig.getIntValue());
    break;
  case dataFloat:
    setValue(orig.getFloatValue());
    break;
  case dataString:
    setValue(orig.getStringValue());
    break;
  case dataInvalid:
    break;
  }

}

/**
 * This method destroys the object.
 */
void WMLScriptValueObject::destroy() {
  free(value);
}

/**
 * Copy constructor.
 */
WMLScriptValueObject::WMLScriptValueObject(WMLScriptValueObject const &other) {
  /* Unconditionally copy other's data. */
  copy(other);
}

/**
 * Overloaded assignment.
 */
WMLScriptValueObject const &WMLScriptValueObject::operator=(const WMLScriptValueObject& other) {
  if (this != &other) {
    /* Only take action if no auto-assignment. */
    destroy();
    copy(other);
  }
  
  /* Return reference to current object for chain-assignments */
  return (*this);
}

WMLScriptValueObject operator-(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);

  if(val1.getType() == dataBool)

    val1.convertTo(dataInt);

  else if(val1.getType() == dataString)

    val1.convertToNumber();

  if(val1.getType() != dataInvalid) {

    if(val1.getType() == dataInt)
      
      returnValue.setValue((int) -(val1.getIntValue()));
    
    else if(val1.getType() == dataFloat) 
      
      returnValue.setValue((double) -(val1.getFloatValue()));
    
  }

  return returnValue;

}

WMLScriptValueObject operator+(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);

  if(val1.getType() == dataBool)

    val1.convertTo(dataInt);

  else if(val1.getType() == dataString)

    val1.convertToNumber();

  if(val1.getType() != dataInvalid) {

    if(val1.getType() == dataInt)
      
      returnValue.setValue((int) +(val1.getIntValue()));
    
    else if(val1.getType() == dataFloat) 
      
      returnValue.setValue((double) +(val1.getFloatValue()));
    
  }

  return returnValue;

}

WMLScriptValueObject operator~(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  val1.convertTo(dataInt);

  if(val1.getType() == dataInt)

    returnValue.setValue((int) ~val1.getIntValue());

  return returnValue;
}

WMLScriptValueObject operator!(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  val1.convertTo(dataBool);

  if(val1.getType() == dataBool) 

    returnValue.setValue((bool) (val1.getBoolValue() ? false : true));

  return returnValue;
}

WMLScriptValueObject operator<(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) 
    
    returnValue.setValue((bool) (strcmp(val1.getStringValue(), 
					val2.getStringValue()) < 0 ? 
				 true : false));
  
  else if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((bool) (val1.getFloatValue() < 
				 val2.getFloatValue() ? 
				 true : false));

  else if(val1.getType() == dataInt && val2.getType() == dataInt)

    returnValue.setValue((bool) (val1.getIntValue() < 
				 val2.getIntValue() ? 
				 true : false));

  return returnValue;
}

WMLScriptValueObject operator<=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) 
    
    returnValue.setValue((bool) (strcmp(val1.getStringValue(), 
					val2.getStringValue()) <= 0 ? 
				 true : false));
  
  else if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((bool) (val1.getFloatValue() <=
				 val2.getFloatValue() ? 
				 true : false));
  
  else if(val1.getType() == dataInt && val2.getType() == dataInt)
    
    returnValue.setValue((bool) (val1.getIntValue() <= 
				 val2.getIntValue() ? 
				 true : false));



  return returnValue;
}

WMLScriptValueObject operator>(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) 
    
    returnValue.setValue((bool) (strcmp(val1.getStringValue(), 
					val2.getStringValue()) > 0 ? 
				 true : false));
  
  else if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((bool) (val1.getFloatValue() >
				 val2.getFloatValue() ? 
				 true : false));
  
  else if(val1.getType() == dataInt && val2.getType() == dataInt)
    
    returnValue.setValue((bool) (val1.getIntValue() >
				 val2.getIntValue() ? 
				 true : false));

  return returnValue;
}

WMLScriptValueObject operator>=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) 
    
    returnValue.setValue((bool) (strcmp(val1.getStringValue(), 
					val2.getStringValue()) >= 0 ? 
				 true : false));
  
  else if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((bool) (val1.getFloatValue() >=
				 val2.getFloatValue() ? 
				 true : false));

  else if(val1.getType() == dataInt && val2.getType() == dataInt)
    
    returnValue.setValue((bool) (val1.getIntValue() >=
				 val2.getIntValue() ? 
				 true : false));

  return returnValue;

}

WMLScriptValueObject operator!=(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue = (in1 == in2);

  if(returnValue.getType() == dataBool)

    returnValue.setValue((bool) (!returnValue.getBoolValue()));

  return returnValue;
}

WMLScriptValueObject operator==(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) 
    
    returnValue.setValue((bool) (strcmp(val1.getStringValue(), 
					val2.getStringValue()) == 0 ? 
				 true : false));
  
  else if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((bool) (val1.getFloatValue() == 
				 val2.getFloatValue() ? 
				 true : false));
  
  else if(val1.getType() == dataInt && val2.getType() == dataInt)
    
    returnValue.setValue((bool) (val1.getIntValue() == 
				 val2.getIntValue() ? 
				 true : false));
  

  return returnValue;

}

WMLScriptValueObject operator<<(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) 

    returnValue.setValue((int) (val1.getIntValue() << val2.getIntValue()));
  
  return returnValue;
}

WMLScriptValueObject operator>>(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) 

    returnValue.setValue((int) (val1.getIntValue() >> val2.getIntValue()));
  
  return returnValue;

}

WMLScriptValueObject operator%(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) {

    returnValue.setValue((int) (val1.getIntValue() % val2.getIntValue()));

  }
  
  return returnValue;
  
}

WMLScriptValueObject operator|(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) {

    returnValue.setValue((int) (val1.getIntValue() | val2.getIntValue()));

  }

  return returnValue;
}

WMLScriptValueObject operator&(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) {

    returnValue.setValue((int) (val1.getIntValue() &val2.getIntValue()));

  }

  return returnValue;
}

WMLScriptValueObject operator^(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) {

    returnValue.setValue((int) (val1.getIntValue() ^val2.getIntValue()));

  }

  return returnValue;
}

WMLScriptValueObject operator*(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataFloat);

  if(val1.getType() == dataFloat && val2.getType() == dataFloat) {

    returnValue.setValue((double) (val1.getFloatValue() * 
				   val2.getFloatValue()));

  } else if(val1.getType() == dataInt && val2.getType() == dataInt) {

    returnValue.setValue((int) (val1.getIntValue() * 
				val2.getIntValue()));
    
  }

  return returnValue;
}

WMLScriptValueObject operator/(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataFloat);
  
  if(val1.getType() == dataFloat && val2.getType() == dataFloat) {
    /* division by zero is invalid */
    if(val2.getFloatValue() != 0)

      returnValue.setValue((double) (val1.getFloatValue() / 
				     val2.getFloatValue()));

  } else if(val1.getType() == dataInt && val2.getType() == dataInt) {
    /* division by zero is invalid */
    if(val2.getIntValue() != 0)

      /* division operator always returns float value */
      returnValue.setValue((double) ((double) val1.getIntValue() / 
				     (double) val2.getIntValue()));

  }

  return returnValue;
}

WMLScriptValueObject operator-(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataFloat);

  if(val1.getType() == dataFloat && val2.getType() == dataFloat)
    
    returnValue.setValue((double) val1.getFloatValue() - 
			 val2.getFloatValue());
  
  else if(val1.getType() == dataInt && val2.getType() == dataInt) 
    
    returnValue.setValue((int) val1.getIntValue() - val2.getIntValue());

  return returnValue;
}

WMLScriptValueObject operator+(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataString);

  if(val1.getType() == dataString && val2.getType() == dataString) {
	
    /* string concatenation */
    char *buf = (char *) malloc(strlen(val1.getStringValue()) + 
				strlen(val2.getStringValue()) + 1);

    if(buf == NULL)
      throw WMLScriptOutOfMemoryException();

    strcpy(buf, val1.getStringValue());
    strcat(buf, val2.getStringValue());
    returnValue.setValue((char *) buf);

    free(buf);
    
  } else if(val1.getType() == dataFloat && val2.getType() == dataFloat) 
    
    returnValue.setValue((double) (val1.getFloatValue() + 
				   val2.getFloatValue()));
  
  else if(val1.getType() == dataInt && val2.getType() == dataInt) 
    
    returnValue.setValue((int) val1.getIntValue() + val2.getIntValue());
    
  return returnValue;

}

WMLScriptValueObject isValid(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;
  returnValue.setValue((bool) (in1.getType() != dataInvalid ? true : false));
  return returnValue;
}

WMLScriptValueObject typeOf(const WMLScriptValueObject& in1) {
  WMLScriptValueObject returnValue;

  switch(in1.getType()) {
  case dataInt:
    returnValue.setValue(0);
    break;
  case dataFloat:
    returnValue.setValue(1);
    break;
  case dataString:
    returnValue.setValue(2);
    break;
  case dataInvalid:
    returnValue.setValue(4);
    break;
  case dataBool:
    returnValue.setValue(3);
    break;
  }
  
  return returnValue;
  
}

WMLScriptValueObject div(const WMLScriptValueObject& in1, const WMLScriptValueObject& in2) {
  WMLScriptValueObject returnValue;
  WMLScriptValueObject val1(in1);
  WMLScriptValueObject val2(in2); 
  val1.convertToCommon(&val2, dataInt, dataInt);

  if(val1.getType() == dataInt && val2.getType() == dataInt) {
    /* division by zero is invalid */
    if(val2.getIntValue() != 0)
      
      returnValue.setValue((int) (val1.getIntValue() / 
				  val2.getIntValue()));

  }
  
  return returnValue;

}

void WMLScriptValueObject::dprintf(char *s, ...) {
  va_list ap;
  va_start(ap, s);
#ifdef WMLS_DEBUG
  vprintf(s, ap);
#endif
  va_end(ap);
}

/**
 * Object initialization (acquires memory).
 */
void WMLScriptValueObject::initialize() {
  value = (WMLScriptValue *) malloc(sizeof(WMLScriptValue));
  if(value == NULL)
    throw WMLScriptOutOfMemoryException();
}

