/* 
 * Copyright (c) 2000-2001, catchy.net. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *      - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 * 
 *      - Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 * 
 *      - All advertising materials mentioning features or use of this software
 *      must display the following acknowledgement: This product includes
 *      software developed by catchy.net and its contributors.
 * 
 *      - Neither name of catchy.net nor the names of its contributors may
 *      be used to endorse or promote products derived from this software
 *      without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY CATCHY.NET AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL CATCHY.NET OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * $Id: WMLScriptVarPool.cpp,v 1.4 2001/01/27 14:36:42 wouter Exp $
 */

#include <stdio.h>

#include "VarPool.h"
#include "WMLScriptVarPool.h"

/**
 * Given an existing VarPool, this constructor creates a new
 * WMLScriptVarPool. The existing VarPool is then no longer the 
 * responsibility of the caller, that is, subsequent manipulation of
 * the existing VarPool should only happen through WMLScriptVarPool
 * methods.
 * 
 * @param vp the VarPool to inherit.  
 */
WMLScriptVarPool::WMLScriptVarPool(VarPool *vp) {
  varPool = vp;
}

/**
 * Constructs a variable pool object from an already existing
 * WMLScriptVarPool. The existing WMLScriptVarPool remains
 * the responsibility of the caller.
 * 
 * @param orig the WMLScriptVarPool to copy.
 */
WMLScriptVarPool::WMLScriptVarPool(WMLScriptVarPool *orig) {
  int i;

  varPool = CreateVarPool();

  for(i = 0; i < orig->size(); i++) {
    createVar(orig->varAt(i));
    setVarValue(orig->varAt(i), orig->getVarValue(orig->varAt(i)));
  }

}

/**
 * Creates a new, empty variable pool.
 */
WMLScriptVarPool::WMLScriptVarPool() {
  varPool = CreateVarPool();
}

/**
 * Destroys the variable pool and any associated values.
 */
WMLScriptVarPool::~WMLScriptVarPool() {
  VarPool *vp = varPool;

  if(varPool) {
    /* first delete any values contained in the variable pool */
    while(vp) {
      delete (WMLScriptValueObject *) vp->value;
      vp = vp->next;
    }

    /* then finally destroy the pool itself */
    DestroyVarPool(varPool);
  }

}

/**
 * Returns the number of variables in the pool.
 *
 * @return the number of variables in the pool.
 */
int WMLScriptVarPool::size() {
  int i = 0;
  VarPool *vp = varPool;

  while(vp) {
    if(vp->var) i++;
    vp = vp->next;
  }

  return i;
    
}

/**
 * Creates a new, uninitialized variable in the pool. See also the
 * CreateVar method, to which this method defers control.
 * 
 * @return error code (see CreateVar).
 */
int WMLScriptVarPool::createVar(char *var) {
  return CreateVar(varPool, var);
}

/**
 * Returns a pointer to the name of the variable at the specified
 * index.
 * 
 * @param index the index of the variable.
 * @return pointer to the name of the variable or NULL if the index
 * is out of bounds.
 */
char *WMLScriptVarPool::varAt(int index) {
  int i = 0;
  VarPool *vp = varPool;

  while(vp) {
    if(i == index) return vp->var;
    i++;
    vp = vp->next;
  }

  return NULL;
  
}

/**
 * Replaces the value of an existing variable with a new value.
 * Contrary to SetVarValue, this method does delete previously
 * stored values. Furthermore, any stored value becomes the 
 * responsibility of this class, that is, the value is automatically
 * dstroyed when this object is destroyed.
 *
 * @param var pointer to the name of the variable.
 * @param val pointer to the new value.
 * @return error code (see SetVarValue).
 */
int WMLScriptVarPool::setVarValue(char *var, WMLScriptValueObject *val) {
  delete (WMLScriptValueObject *) getVarValue(var);
  return SetVarValue(varPool, var, (void *) val);
}

/**
 * Returns the value for the specified variable.
 *
 * @param var pointer to the name of the variable.
 * @return pointer to the value (or NULL on error, see GetVarValue).
 */
WMLScriptValueObject *WMLScriptVarPool::getVarValue(char *var) {
  return (WMLScriptValueObject *) GetVarValue(varPool, var);
}

/**
 * Queries whether a variable exists in the pool.
 *
 * @return error code (see ExistsVar).
 */
int WMLScriptVarPool::existsVar(char *var) {
  return ExistsVar(varPool, var);
}

/**
 * Dumps the contents of the variable pool to stdout.
 */
void WMLScriptVarPool::dump() {
  VarPool *vp = varPool;
  while(vp) {
    if(vp->var) {
      WMLScriptValueObject *value = (WMLScriptValueObject *) vp->value;
      if(value) {
	char *s = value->toString();
	printf("%s\t\t-> type: %d value: %s\n", vp->var, value->getType(), s);
	free(s);
      } else {
	printf("%s\t\t-> (null)\n", vp->var);
      }

    }
    vp = vp->next;
  }
}


